/*
** Copyright (c) 1999, 2000 Oracle Corporation, All rights reserved.
**
** Class:  ConnectionPool
** Author: Doug McMahon
** Date:   May 15, 2000
**
** Description:
**
**  Manages a global pool of JDBC database connections.  Attempts to
**  reuse connections maintained in an internal array, otherwise
**  creates new connections as necessary.  The main methods clients
**  should use are getInstance, getConnection, and releaseConnection.
**
** Methods:
**
**  getInstance         Returns handle to the singleton connection pool object
**  makeDatabaseString  Create a JDBC/SQL*Net driver/database connect string 
**  addConnection       Insert a connection into the pool
**  newConnection       Create a connection and insert it into the pool
**  removeConnection    Remove a locked connection from the pool
**  releaseConnection   Return a connection back to the pool, or close it
**  getConnection       Get a connection from pool, create if necessary
**  getSize             Return the current number of connections in the pool
**  getMaximum          Get the current maximum pool size
**  setMaximum          Set the current maximum pool size
**  clearPool           Remove and close all unused connections in the pool
**  closePool           Free unused connections, set maximum to zero
**  lockConnection      Lock a specific connection, return true if successful
**  findConnection      Determine if a specific connection is in the pool
*/
import java.util.*;
import java.sql.*;

public class ConnectionPool
{
    /*
    ** Global variable common to all threads, holding the singleton instance
    */
    private static ConnectionPool s_GlobalInstance = null;

    /*
    ** Instance variables; since there will be only one instance
    ** of this class, they're effectively global, too.  It would
    ** probably be a better design to have a single Vector holding
    ** structures, but this involves an additional class and may
    ** not actually perform any better because of the way many
    ** of the searches work.
    */
    private Vector    m_Conns;   // Holds the list of connections
    private Vector    m_Names;   // Holds the list of connection names
    private boolean[] m_Flags;   // Holds the list of lock flags
    private int       m_PoolMax; // Limit on size of pool

    /*
    ** Constants
    */
    final static int VECTOR_INIT_SIZE = 10; // Initial size of pool
    final static int VECTOR_INCR_SIZE = 10; // Pool size increment
    final static int VECTOR_MAX_SIZE = 100; // Maximum pool size

    /*
    ** Static method creates/returns a handle to the one and only
    ** instance of this object; this is what programmers will call
    ** to get a handle to the pool.
    */
    public static synchronized ConnectionPool getInstance()
    {
        if (s_GlobalInstance == null)
            s_GlobalInstance = new ConnectionPool();
        return(s_GlobalInstance);
    }

    /*
    ** Constructor: because it's private, it can only be called from
    ** within this class, guaranteeing a single instance.  All this
    ** function does is initialize the pool vectors.
    */
    private ConnectionPool()
    {
        m_PoolMax = 1; // Default is one connection in the pool
        m_Conns = new Vector(VECTOR_INIT_SIZE, VECTOR_INCR_SIZE);
        m_Names = new Vector(VECTOR_INIT_SIZE, VECTOR_INCR_SIZE);
        m_Flags = new boolean[VECTOR_MAX_SIZE];
    }

    /*
    ** Create Oracle JDBC connect string given driver and database alias
    ** Valid driver values are "thin" (the default) and "oci8".
    ** Valid alias values are a NET8 V2 alias (only for OCI8) or a
    ** thin driver connect string of the form <host>:<port>:<sid>
    ** (if this format is passed for anything other than the thin driver,
    ** this routine converts it to the long format).
    */
    public static String makeDatabaseString(String p_Driver, String p_Alias)
    {
        String drvName = (p_Driver == null) ? "thin" : p_Driver.toLowerCase();
        String dbName  = (p_Alias  == null) ? ""     : p_Alias;

        if (!drvName.equals("thin"))
        {
            String host;
            String port;
            String sid;
            int    i;
            int    j;

            i = dbName.indexOf(':', 0);
            if (i >= 0)
            {
                host = dbName.substring(0, i);
                j = dbName.indexOf(':', ++i);
                sid = dbName.substring((j < 0) ? i : j + 1);
                port = (j > i) ? dbName.substring(i, j) : "1521";
                dbName = "(description=(address=(host="+host+
                         ")(protocol=tcp)(port="+port+
                         "))(connect_data=(sid="+sid+")))";
            }
        }

        return("jdbc:oracle:"+drvName+":@"+dbName);
    }

    /*
    ** Create a key string to label a connection
    ** Consists of the username + separator + database driver/connection string
    ** Uses a space as the separator, which seems safe enough since
    ** normally spaces can't occur in the driver/connection string.
    ** The username is first because it's more selective during searches.
    */
    private static String makeNameString(String p_Username,
                                         String p_Database)
    {
        return(p_Username+" "+p_Database);
    }

    /*
    ** Add a connection to the pool; the connection is created in its locked
    ** state.  To unlock it and make it available, use releaseConnection.
    ** This method supports clients that want to add their own externally-
    ** created connections to the pool.
    */
    public void addConnection(Connection p_Conn, String p_Username,
                              String p_Database)
    {
        String nameString;

        nameString = makeNameString(p_Username, p_Database);

        synchronized(this)
        {
            int i = m_Conns.size();
            m_Conns.addElement(p_Conn);
            m_Names.addElement(nameString);
            m_Flags[i] = true;
        }
    }

    /*
    ** Create a new connection and add it to the pool
    */
    public Connection newConnection(String p_Username, String p_Password,
                                    String p_Database)
               throws SQLException
    {
        Connection conn;
        String     oradrv = "oracle.jdbc.driver.OracleDriver";

        try
        {
            Class.forName(oradrv);
        }
        catch(ClassNotFoundException e)
        {
            SQLException esql = new SQLException(oradrv+":"+e.getMessage());
            throw(esql);
        }

        System.out.println("DBG> connect to "+p_Database); // ###
        conn = DriverManager.getConnection(p_Database, p_Username, p_Password);

        addConnection(conn, p_Username, p_Database);

        System.out.println("DBG> making new connection"); // ###
        return(conn);
    }

    /*
    ** Remove the element at the specified index location
    ** Only call this method from within a synchronized block!
    */
    private void removeIndex(int p_Loc, int p_Total)
    {
        int i;

        m_Conns.removeElementAt(p_Loc);
        m_Names.removeElementAt(p_Loc);

        for (i = p_Loc + 1; i < p_Total; ++i)
            m_Flags[i - 1] = m_Flags[i];
    }

    /*
    ** Remove a connection from the pool
    ** Checks to make sure the connection is locked; never remove a
    ** connection without locking it first.  Removal supports clients
    ** that might want to permanently withdraw a connection from the
    ** pool (the reverse of addConnection).
    */
    public synchronized void removeConnection(Connection p_Conn)
    {
        int i = m_Conns.indexOf(p_Conn);

        if (i >= 0)
            if (m_Flags[i])
                removeIndex(i, m_Conns.size());
    }

    /*
    ** Release a connection to the pool
    ** If the pool is over the maximum, the connection is removed and closed,
    ** otherwise it's just marked as unlocked.
    */
    public synchronized void releaseConnection(Connection p_Conn)
                      throws SQLException
    {
        int i = m_Conns.indexOf(p_Conn);

        if (i >= 0)
        {
            if (m_Conns.size() > m_PoolMax)
            {
                removeConnection(p_Conn);
                p_Conn.close();
            }
            else
                m_Flags[i] = false;
        }
    }

    /*
    ** Attempt to find a matching connection in the pool;
    ** if not, add one and return it.
    */
    public Connection getConnection(String p_Username, String p_Password,
                                    String p_Database)
               throws SQLException
    {
        String nameString;

        nameString = makeNameString(p_Username, p_Database);

        synchronized(this)
        {
            int        num = m_Names.size();
            int        i = num;
            int        unused = num;
            Connection conn;

            /*
            ** First, attempt to find a matching active connection
            ** Since connections are added to the end of the list,
            ** searching backwards to reuse them is a rough form of LRU.
            */
            while (i > 0)
            {
                if (!m_Flags[--i])
                {
                    String s = (String)m_Names.elementAt(i);
                    if (nameString.equals(s))
                    {
                        /*
                        ** Found a match; remove it and add it back to
                        ** the vector so that it will be pushed to the
                        ** end of the list (LRU connections migrate to
                        ** the front of the list).
                        */
                        conn = (Connection)m_Conns.elementAt(i);
                        if (conn.isClosed())
                        {
                            /*
                            ** This connection has "gone bad" for some
                            ** reason; remove it and drop through to
                            ** creating a new one.
                            */
			    System.out.println("DBG> bad connection"); // ###
                            removeIndex(i, num);
                            return(newConnection(p_Username, p_Password,
                                                 p_Database));
                        }
                        if (i != (num - 1))
                        {
                            removeIndex(i, num);
                            m_Conns.addElement(conn);
                            m_Names.addElement(s);
                            m_Flags[num - 1] = true;
                        }
                        System.out.println("DBG> reused connection"); // ###
                        return(conn);
                    }
                    unused = i;
                }
            }

            /*
            ** No perfect, unlocked match was found; unfortunately there
            ** is no way to simply re-bind a connection to a new user
            ** account so it's pointless to try to reuse one.  Instead,
            ** if at maximum pool size, remove the first unused connection.
            */
            if ((num >= m_PoolMax) && (unused < num))
            {
                conn = (Connection)m_Conns.elementAt(unused);
                m_Flags[unused] = true;
                removeConnection(conn);
                conn.close();
            }
        }

        return(newConnection(p_Username, p_Password, p_Database));
    }

    /*
    ** Returns the current maximum pool size
    */
    public synchronized int getSize()
    {
        return(m_Conns.size());
    }

    /*
    ** Returns the current maximum pool size
    */
    public synchronized int getMaximum()
    {
        return(m_PoolMax);
    }

    /*
    ** Sets the maximum pool size; size must be at least 0
    */
    public synchronized void setMaximum(int p_MaxSize)
    {
        if ((p_MaxSize >= 0) && (p_MaxSize < VECTOR_MAX_SIZE))
            m_PoolMax = p_MaxSize;
    }

    /*
    ** Remove all connections from the pool that aren't currently in use
    */
    public synchronized void clearPool()
                      throws SQLException
    {
        int    n = m_Conns.size();
        int    i;

        /*
        ** Loop forward to avoid quadratic behavior in removeConnection
        ** and to avoid revisiting locked connections when indexes change
        */
        for (i = 0; i < n; ++i)
            if (!m_Flags[i])
            {
                Connection conn = (Connection)m_Conns.elementAt(i);
                removeConnection(conn);
                conn.close();
                --n;
            }
    }

    /*
    ** Close the pool by setting the maximum to zero and removing
    ** any unused connections; as the remaining connections are released,
    ** they will be closed because the maximum is by definition exceeded.
    */
    public void closePool()
         throws SQLException
    {
        setMaximum(0);
        clearPool();
    }

    /*
    ** Search for a specific connection in the pool and lock it
    ** if possible.  Returns true if successful, false otherwise.
    ** To unlock the connection, use releaseConnection.
    */
    public synchronized boolean lockConnection(Connection p_Conn)
    {
        int i = m_Conns.indexOf(p_Conn);

        if (i >= 0)
            if (!m_Flags[i])
            {
                m_Flags[i] = true;
                return(true);
            }

        return(false);
    }

    /*
    ** Search for a specific connection in the pool, return true
    ** if the connection is still in the pool, false otherwise.
    ** If a lockConnection operation returns false, and you need
    ** to know if the failure was due to the connection not being
    ** in the pool anymore, call this function to find out.
    */
    public synchronized boolean findConnection(Connection p_Conn)
    {
        if (m_Conns.indexOf(p_Conn) < 0) return(false);
        return(true);
    }
}
