/*
** Copyright (c) 1999, 2000 Oracle Corporation, All rights reserved.
**
** Class:  GatewayManager
** Author: Doug McMahon
** Date:   August 12, 2000
**
** Description:
**
**  Handles traffic to/from the OWA PL/SQL gateway.  Clients should
**  set the connection and request context, then send the CGI information.
**  Then clients can call PL/SQL to generate the OWA response.  Finally,
**  clients call getHeader and/or getContentBlock to read back the
**  information.
**
** Methods:
**
**  getResetStatement   Get cursor for the reset package statement
**  getCGIStatement     Get cursor for CGI-passing statement
**  getPageStatement    Get cursor for page-reading statement
**  makeErrorMessage    Create SQL exception with error message
**  setConnection       Sets JDBC connection to use for OWA
**  getConnection       Returns the internal JDBC connection
**  setCGI              Sets the CGI environment for the OWA
**  sendCGI             Sends the CGI info to the OWA
**  getAltDataBlock     Get a block of data from the OWA alternative package
**  getDataBlock        Get a block of data from the OWA
**  getHeader           Get a vector containing the header elements from OWA
**  getContentBlock     Get the next block of content from OWA
**  runProcedure        Run an OWA PL/SQL procedure
**  getMimeType         Get content type returned with LOB
**  getBinarySource     Get stream for reading from BLOB
**  getCharacterSource  Get stream for reading from CLOB
**  getBinarySink       Get stream for writting to BLOB
**  getCharacterSink    Get stream for writting to CLOB
*/
import java.io.*;
import java.util.*;

import java.sql.*;
import oracle.sql.*;
import oracle.jdbc.driver.*;

public class GatewayManager
{
    /*
    ** Global to all objects
    */
    private static String s_ResetSql =
            "begin DBMS_SESSION.RESET_PACKAGE; end;";

    private static String s_CGISql =
            "declare\n"                                         +
            "  p_cnt number; p_ns owa_array; p_vs owa_array;\n" +
            "  nmarr OWA.vc_arr; vlarr OWA.vc_arr;\n"           +
            "begin\n"                                           +
            "  p_cnt := ?; p_ns := ?; p_vs := ?;\n"             +
            "  for i in 1..p_cnt loop\n"                        +
            "    nmarr(i) := p_ns(i); vlarr(i) := p_vs(i);\n"   +
            "  end loop;\n"                                     +
            "  OWA.INIT_CGI_ENV(p_cnt,nmarr,vlarr);\n"          +
            "end;";

    private static String s_GetPageSql =
            "declare nlns number; p_vl owa_array;\n" +
            "        lines htp.htbuf_arr;\n"         +
            "begin\n"                                +
            "  nlns := ?; p_vl := owa_array(1);\n" +
            "  OWA.GET_PAGE(lines, nlns);\n"         +
            "  if (nlns < 1) then p_vl(1) := '';\n"  +
            "  else p_vl.extend(nlns); end if;\n"    +
            "  for i in 1..nlns loop\n"              +
            "    p_vl(i) := lines(i);\n"             +
            "  end loop;\n"                          +
            "  ? := p_vl; ? := nlns;\n"              +
            "end;";

    /*
    ** Instance variables
    */
    private Connection        m_Conn;             // DB connection for gateway
    private CallableStatement m_ResetStmt;        // RESET_PACKAGE statement
    private CallableStatement m_CGIStmt;          // INIT_CGI_ENV statement
    private CallableStatement m_GetPageStmt;      // GET_PAGE statement
    private String[]          m_CGInames;         // CGI environment names
    private String[]          m_CGIvalues;        // CGI environment values
    private int               m_NumEnv = 0;       // No. of environment values
    private int               m_State;            // OWA state transitions
    private Vector            m_SavedHeader;      // Saves last header
    private String            m_SavedCharContent; // "leftover" char content
    private byte[]            m_SavedRawContent;  // "leftover" raw content
    private String            m_MimeType;         // Saves mime type return
    private BLOB              m_Blob;             // Saves BLOB return
    private CLOB              m_Clob;             // Saves CLOB return
    private String            m_AltPkgName;       // Alternate OWA package name
    private String            m_LastStmt;         // Last SQL statement run

    /*
    ** Constants
    */
    final static int MAX_OWA_ARRAY_SIZE    = 512; // Max size of an OWA_ARRAY
    final static int MAX_OWA_ELEMENT_WIDTH = 256; // Max width of OWA_ARRAY

    final static int VECTOR_INIT_SIZE  = 10;      // Initial size of pool
    final static int VECTOR_INCR_SIZE  = 10;      // Pool size increment

    final static int STATE_NONE     = 0;
    final static int STATE_READY    = 1;
    final static int STATE_FETCHING = 2;
    final static int STATE_FINISHED = 3;

    final static String ARRAY_TYPE = "OWA_ARRAY";

    public final static int RUN_MODE_NORMAL      = 0;
    public final static int RUN_MODE_POSITIONAL  = 1;
    public final static int RUN_MODE_FLEXIBLE    = 2;
    public final static int RUN_MODE_LOBREAD     = 3;
    public final static int RUN_SECURITY_NONE    = 0;
    public final static int RUN_SECURITY_INIT    = 4;
    public final static int RUN_SECURITY_CUSTOM  = 12;
    public final static int RUN_SECURITY_PACKAGE = 16;

    public final static String DOCUMENT_KEY = "document_name";

    /*
    ** Get/reuse one of the standard statements
    ** These all assume the connection handle is valid!
    */
    private OracleCallableStatement getResetStatement()
                             throws SQLException
    {
        m_LastStmt = s_ResetSql;
        if (m_ResetStmt == null)
            m_ResetStmt = m_Conn.prepareCall(m_LastStmt);
        return((OracleCallableStatement)m_ResetStmt);
    }

    private OracleCallableStatement getCGIStatement()
                             throws SQLException
    {
        if (m_AltPkgName == null)
            m_LastStmt = s_CGISql;
        else
        {
            String s = "";
            int    start_loc = 0;
            int    end_loc;

            while (true)
            {
                end_loc = s_CGISql.indexOf("OWA.", start_loc);
                if (end_loc < start_loc)
                {
                    s = s + s_CGISql.substring(start_loc);
                    break;
                }
                s = s + s_CGISql.substring(start_loc, end_loc) + m_AltPkgName;
                start_loc = end_loc + 3;
            }
            m_LastStmt = s;
        }

        if (m_CGIStmt == null)
            m_CGIStmt = m_Conn.prepareCall(m_LastStmt);
        return((OracleCallableStatement)m_CGIStmt);
    }

    private OracleCallableStatement getPageStatement()
                             throws SQLException
    {
        if (m_AltPkgName == null)
            m_LastStmt = s_GetPageSql;
        else
            m_LastStmt = "begin\n  "+m_AltPkgName+
                         ".GET_PAGE(:cbuffer, :rbuffer, :moredata);\nend;";
        if (m_GetPageStmt == null)
            m_GetPageStmt = m_Conn.prepareCall(m_LastStmt);
        return((OracleCallableStatement)m_GetPageStmt);
    }

    /*
    ** Create a SQL exception with both the error message
    ** and the last SQL statement being executed.
    */
    private SQLException makeErrorMessage(String p_Msg)
    {
        return(new SQLException("\n"+p_Msg+"\n"+m_LastStmt));
    }

    /*
    ** Set the JDBC connection to the OWA, reset flags
    ** If the connection doesn't match the previous one, close and
    ** invalidate the old statement handles so they won't be reused.
    **
    ** Note: call this with null as the argument to force closure
    ** of the statement handles before allowing the GatewayManager
    ** object to leave scope, otherwise you'll lose cursor handles!
    */
    public void setConnection(Connection p_Conn)
    {
        if (p_Conn != m_Conn)
        {
            try
            {
                if (m_ResetStmt != null)   m_ResetStmt.close();
                if (m_CGIStmt != null)     m_CGIStmt.close();
                if (m_GetPageStmt != null) m_GetPageStmt.close();
            }
            catch(SQLException e)
            {
                System.out.println("OWA Warning: "+e.getMessage());
            }
            m_ResetStmt = null;
            m_CGIStmt = null;
            m_GetPageStmt = null;
            m_Conn = p_Conn;
        }
        m_State = STATE_NONE;
        m_SavedCharContent = null;
        m_SavedRawContent = null;
        m_MimeType = null;
    }

    /*
    ** Set the JDBC connection to the OWA, reset flags
    ** If the connection doesn't match the previous one, close and
    ** invalidate the old statement handles so they won't be reused.
    **
    ** Note: call this with null as the argument to force closure
    ** of the statement handles before allowing the GatewayManager
    ** object to leave scope, otherwise you'll lose cursor handles!
    */
    public void setConnection(Connection p_Conn, String p_PkgName)
    {
        if (m_AltPkgName == null)
            setConnection(null);
        else if (!(m_AltPkgName.equalsIgnoreCase(p_PkgName)))
            setConnection(null);
        m_AltPkgName = p_PkgName;
        setConnection(p_Conn);
    }

    /*
    ** Constructor from database connection and alternate package name
    */
    public GatewayManager(Connection p_Conn, String p_PkgName)
    {
        setConnection(p_Conn, p_PkgName);
    }

    /*
    ** Constructor from database connection
    */
    public GatewayManager(Connection p_Conn)
    {
        setConnection(p_Conn);
    }

    /*
    ** Default constructor
    */
    public GatewayManager()
    {
        setConnection(null);
    }

    /*
    ** Get the JDBC connection
    */
    public Connection getConnection()
    {
        return(m_Conn);
    }

    /*
    ** Set the CGI environment arrays directly
    ** If any invalid argument is passed, sets a single default
    */
    public void setCGI(String[] p_Names, String[] p_Values, int p_Nenv)
    {
        if ((p_Names != null) && (p_Values != null) && (p_Nenv > 0))
        {
            m_CGInames = p_Names;
            m_CGIvalues = p_Values;
            m_NumEnv = p_Nenv;
        }
        else if (m_NumEnv < 1)
        {
            m_CGInames = new String[1];
            m_CGIvalues = new String[1];
            m_CGInames[0]  = "GATEWAY_INTERFACE";
            m_CGIvalues[0] = "CGI/1.1";
            m_NumEnv = 1;
        }
    }

    /*
    ** Convert servlet request information to Apache CGI format and
    ** transfer to the OWA via PL/SQL.
    */
    public void sendCGI()
         throws SQLException
    {
        String[] nams;
        String[] vals;
        int      nenv = m_NumEnv;

        if (m_Conn == null)
        {
            SQLException esql = new SQLException("Error: no connection");
            throw(esql);
        }

        if (nenv == 0)
        {
            setCGI(null, null, 0);
            nenv = 1;
        }
        nams = m_CGInames;
        vals = m_CGIvalues;

        m_State = STATE_NONE;
        m_SavedCharContent = null;
        m_SavedRawContent = null;
        m_MimeType = null;

        try
        {
            OracleCallableStatement cs;
            ArrayDescriptor         vcarray;
            ARRAY                   oarray1;
            ARRAY                   oarray2;

            if (m_AltPkgName == null)
            {
                cs = getResetStatement();
                cs.executeUpdate();
            }

            /*
            ** This next operation requires the following object type be
            ** pre-defined in the target schema:
            **
            **   create or replace type OWA_ARRAY as table of VARCHAR2(32767);
            **   /
            **
            ** ### NEED TO MAKE THIS A STANDARD PART OF THE OWA INSTALL?
            */
            vcarray = ArrayDescriptor.createDescriptor(ARRAY_TYPE, m_Conn);
            oarray1 = new ARRAY(vcarray, m_Conn, nams);
            oarray2 = new ARRAY(vcarray, m_Conn, vals);

            cs = getCGIStatement();
            cs.setInt(1, nenv);
            cs.setArray(2, oarray1);
            cs.setArray(3, oarray2);
            cs.executeUpdate();

            m_State = STATE_READY;
        }
        catch(SQLException e)
        {
            m_Conn.clearWarnings();
            throw(makeErrorMessage(e.getMessage()));
        }
    }

    /*
    ** Fetch a block of data from PL/SQL using the alternate calling
    ** mode that returns a pair of 32K buffers.
    **
    ** ### ADD SUPPORT FOR ALTERNATE CALLING STRUCTURES:
    ** ###   (String Header, Blob BinaryContent, Clob CharacterContent)
    ** ###   (String Cbuffer, Raw Rbuffer)
    */
    private String getAltDataBlock(OracleCallableStatement cs)
         throws SQLException
    {
        boolean moreflag = false;
        String  CharContent;
        byte[]  RawContent;
        int     clen = 0;
        int     rlen = 0;

        try
        {
            cs.setString(1, "\n");
            cs.registerOutParameter(1, OracleTypes.VARCHAR);
            cs.registerOutParameter(2, OracleTypes.VARBINARY);
            cs.registerOutParameter(3, OracleTypes.VARCHAR);
            cs.executeUpdate();

            CharContent = cs.getString(1);
            RawContent = cs.getBytes(2);
            if (CharContent != null)
            {
                clen = CharContent.length();
                if (clen <= 0) CharContent = null;
            }
            if (RawContent != null)
            {
                rlen = java.lang.reflect.Array.getLength(RawContent);
                if (rlen <= 0) RawContent = null;
            }
            m_SavedRawContent = RawContent;

            if ((clen > 0) || (rlen > 0))
            {
                String MoreFlag = cs.getString(3);
                if (MoreFlag != null)
                    if (MoreFlag.length() > 0)
                        moreflag = true;
            }
        }
        catch(SQLException e)
        {
            m_Conn.clearWarnings();
            throw(makeErrorMessage(e.getMessage()));
        }

        m_State = (moreflag) ? STATE_FETCHING : STATE_FINISHED;

        return(CharContent);
    }

    /*
    ** Fetch a block of data from the OWA and return it as a string;
    ** be sure the OWA is in the proper state first!
    */
    private String getDataBlock()
         throws SQLException
    {
        OracleCallableStatement cs = getPageStatement();
        ARRAY                   oarray;
        int                     nlines = MAX_OWA_ARRAY_SIZE;
        int                     i;
        String[]                lines;
        StringBuffer            ContentBuffer;

        if (m_AltPkgName != null) return(getAltDataBlock(cs));

        try
        {
            cs.setInt(1, nlines);
            cs.registerOutParameter(2, OracleTypes.ARRAY, ARRAY_TYPE);
            cs.registerOutParameter(3, OracleTypes.NUMBER);
            cs.executeUpdate();

            oarray = (ARRAY)cs.getObject(2);
            lines = (String[])oarray.getArray();
            nlines = cs.getInt(3);
        }
        catch(SQLException e)
        {
            m_Conn.clearWarnings();
            throw(makeErrorMessage(e.getMessage()));
        }

        m_State = STATE_FETCHING; // Signal that fetch is initiated
        if (nlines < MAX_OWA_ARRAY_SIZE)
        {
            m_State = STATE_FINISHED; // Signal end-of-fetch
            if (nlines < 1) return(null);
        }

        /*
        ** Concatenate all the lines into a single buffer and return it
        */
        ContentBuffer = new StringBuffer(MAX_OWA_ARRAY_SIZE *
                                         MAX_OWA_ELEMENT_WIDTH);

        for (i = 0; i < nlines; ++i)
            ContentBuffer.append(lines[i]);

        return(ContentBuffer.toString());
    }

    /*
    ** Initiate a page transfer from the OWA.  It's assumed that you
    ** already passed the CGI environment and called the action procedure
    ** to fill the OWA with return content.
    **
    ** Returns a vector of header elements (one element per header item,
    ** otherwise unparsed).  Note that it's possible to get a null return.
    **
    ** Any remaining content is saved until you call getContentBlock().
    */
    public Vector getHeader()
         throws SQLException
    {
        String  HeadString;
        int     hlen;
        int     cur_loc;
        int     next_nl;
        int     colon;
        boolean head_found;
        Vector  v;

        if (m_Conn == null)
        {
            SQLException esql = new SQLException("Error: no connection");
            throw(esql);
        }

        /*
        ** If CGI hasn't been sent, this shouldn't be run, so return null
        */
        if (m_State != STATE_READY) return(null);

        m_SavedHeader = null;
        m_SavedCharContent = null;
        m_SavedRawContent = null;

        HeadString = getDataBlock();

        /*
        ** Scan the buffer for header elements by looking for
        ** newline breaks.  Transfer elements into the vector in
        ** anticipation of returning it.  Stop at the first blank line.
        */
        v = new Vector(VECTOR_INIT_SIZE, VECTOR_INCR_SIZE);
        hlen = (HeadString == null) ? 0 : HeadString.length();
        next_nl = 0;
        cur_loc = 0;
        head_found = false;
        while (cur_loc < hlen)
        {
            String s;

            next_nl = HeadString.indexOf('\n', cur_loc);
            if (next_nl <= cur_loc) // blank line or no newline
            {
                if (next_nl == cur_loc) ++cur_loc;
                break;
            }

            s = HeadString.substring(cur_loc, next_nl);
            v.addElement(s);

            cur_loc = next_nl + 1;
            colon = s.indexOf(':', 0);
            if (colon > 0)
            {
                if ((s.regionMatches(true,0,"Content-Type:",0,colon))   ||
                    (s.regionMatches(true,0,"Content-Length:",0,colon)) ||
                    (s.regionMatches(true,0,"Location:",0,colon))       ||
                    (s.regionMatches(true,0,"Refresh:",0,colon))        ||
                    (s.regionMatches(true,0,"WWW-Authenticate:",0,colon)))
                    head_found = true;
            }
        }

        if (head_found)
        {
            /*
            ** If a blank line has been found and one of the trigger words
            ** above was sensed, then we've found the header; save the
            ** remainder of the string to return on the getContentBlock
            ** operation and return a vector with the header elements.
            */
            if (cur_loc < hlen)
                m_SavedCharContent = HeadString.substring(cur_loc);
            m_SavedHeader = v;
        }
        else
        {
            /*
            ** Otherwise there doesn't appear to be a header so save
            ** the entire string for the getContentBlock operation and
            ** return a null vector.
            */
            if (hlen > 0) m_SavedCharContent = HeadString;
        }
        if (m_SavedRawContent != null) m_SavedCharContent = null;

        return(m_SavedHeader);
    }

    /*
    ** Returns a block of content from the OWA as a string.
    ** Returns null if there is no more content.  If the operation is
    ** ready but hasn't started, reads and discards the header first.
    */
    public String getContentBlock()
         throws SQLException
    {
        if (m_Conn == null)
        {
            SQLException esql = new SQLException("Error: no connection");
            throw(esql);
        }

        /*
        ** If we haven't sent the CGI yet, this can't run
        */
        if (m_State == STATE_NONE) return(null);

        /*
        ** If in the ready state, we haven't fetched the header yet,
        ** so fetch and discard it.
        */
        if (m_State == STATE_READY)
        {
            Vector v = getHeader();
        }

        /*
        ** If there's any content left over from getHeader, return it.
        */
        if ((m_SavedCharContent != null) || (m_SavedRawContent != null))
        {
            String s = m_SavedCharContent;
            m_SavedCharContent = null;
            return(s);
        }

        /*
        ** If we haven't seen the last content block, get more data from OWA
        */
        if (m_State != STATE_FINISHED)
            return(getDataBlock());

        m_State = STATE_NONE; // No further data
        return(null);
    }

    /*
    ** Get any RAW content that is available
    */
    public byte[] getRawContentBlock()
         throws SQLException
    {
        byte[] rbuf = m_SavedRawContent;
        m_SavedRawContent = null;
        return(rbuf);
    }

    /*
    ** Build and run a PL/SQL gateway action procedure.
    **
    ** Arguments are:
    **  Name of package.procedure to be run
    **  Bit flag signalling the argument-passing and security modes
    **  A Vector of URL argument name strings
    **  A Vector of URL argument value string arrays
    **
    ** The vectors are assumed to be ordered.  It's assumed that the
    ** name/value pairs appear in the order in which they were found
    ** in the HTTP request, except where a name is duplicated.  For
    ** duplicated names, these should appear grouped with the first
    ** such argument found and in the order seen.
    **
    ** Example: If the HTTP input arguments were:
    **            ("A=Cat", "B=1", "C=x", "A=Dog", "C=y")
    **          Then the call to this procedure should pass:
    **            ("A","A","B","C","C"), ("Cat","Dog","1","x","y")
    **
    ** ### NEED TO ADD SUPPORT FOR BFILE CONTENT RETURNS
    */
    public void runProcedure(String p_ProcName, int p_Mode,
                             Vector p_Names, Vector p_Values,
                             String p_MimeType)
         throws SQLException
    {
        String                  sql; // ### PERHAPS SHOULD USE A STRING BUFFER?
        Object[]                arg_vals;
        String[]                arg_nams;
        String[]                arg_arr;
        int                     i, j;
        int                     tot_args;
        int                     num_args;
        int                     pkg_idx;
        boolean                 auth_flag;
        boolean                 lob_flag;
        OracleCallableStatement cs;
        ArrayDescriptor         vcarray;
        ARRAY                   oarray;

        if (m_Conn == null)
        {
            SQLException esql = new SQLException("Error: no connection");
            throw(esql);
        }

        /*
        ** If the CGI environment hasn't yet been sent, do so
        */
        if (m_State != STATE_READY) sendCGI();

        num_args = 0;
        if ((p_Names != null) && (p_Values != null))
        {
            num_args = p_Names.size();
            tot_args = p_Values.size();
            if (num_args > tot_args) num_args = tot_args;
        }

        lob_flag = ((p_Mode & RUN_MODE_LOBREAD) == RUN_MODE_LOBREAD);

        if ((p_Mode & RUN_MODE_FLEXIBLE) == 0)
        {
            /*
            ** For normal and positional modes, count the number of
            ** distinct arguments and then create string arrays or
            ** strings for each one.  Put them into the arg_nams and
            ** arg_vals vectors.
            */
            arg_nams = new String[num_args];
            arg_vals = new Object[num_args];
            for (i = 0; i < num_args; ++i)
            {
                String[] val = (String [])p_Values.elementAt(i);
                int      valsz = java.lang.reflect.Array.getLength(val);

                if (valsz > 1) arg_vals[i] = val;
                else           arg_vals[i] = val[0];

                arg_nams[i] = (String)p_Names.elementAt(i);
            }
        }
        else
        {
            /*
            ** For flexible arguments, build a pair of arrays
            ** First count the grand total of all arguments
            ** Then transfer the arguments to new arrays
            */
            j = 0;
            tot_args = 0;
            for (i = 0; i < num_args; ++i)
            {
                String[] val = (String [])p_Values.elementAt(i);
                tot_args += java.lang.reflect.Array.getLength(val);
                if (lob_flag)
                    if ((String)p_Names.elementAt(i) == DOCUMENT_KEY)
                        j = 1;
            }

            arg_nams = new String[tot_args];
            arg_arr  = new String[tot_args];

            if (num_args > 0)
            {
                for (i = 0; i < num_args; ++i)
                {
                    String   key = (String)p_Names.elementAt(i);
                    String[] val = (String [])p_Values.elementAt(i);
                    int      valsz = java.lang.reflect.Array.getLength(val);
                    int      n;

                    if ((lob_flag) && (key == DOCUMENT_KEY))
                    {
                        /* For LOB read, the document name must be first */
                        arg_nams[0] = key;
                        arg_arr[0]  = val[0];
                        continue;
                    }

                    for (n = 0; n < valsz; ++n, ++j)
                    {
                        arg_nams[j] = key;
                        arg_arr[j]  = val[n];
                    }
                }
            }

            num_args = 4;

            arg_vals = new Object[num_args];
            arg_vals[0] = new Integer(tot_args);
            arg_vals[1] = arg_nams;
            arg_vals[2] = arg_arr;

            arg_arr = new String[tot_args];
            for (i = 0; i < tot_args; ++i) arg_arr[i] = "";
            arg_vals[3] = arg_arr;

            arg_nams = new String[num_args];
            arg_nams[0] = "num_entries";
            arg_nams[1] = "name_array";
            arg_nams[2] = "value_array";
            arg_nams[3] = "reserved";
        }

        arg_arr = new String[0];

        /*
        ** Build a PL/SQL anonymous block of the form
        **
        **   begin
        **     if <authfunction> then
        **       <procedure>(<parameters>);
        **       commit;
        **     end if;
        **   end;
        */
        sql = "begin\n";

        /*
        ** Build the authorization call, if any
        */
        auth_flag = false;
        pkg_idx = p_ProcName.indexOf('.');
        if (((p_Mode & RUN_SECURITY_PACKAGE) != 0) && (pkg_idx > 0))
        {
            sql += ("  if "+p_ProcName.substring(0, pkg_idx));
            auth_flag = true;
        }
        else if ((p_Mode & RUN_SECURITY_CUSTOM) == RUN_SECURITY_CUSTOM)
        {
            if (m_AltPkgName == null)
                sql += "  if OWA_CUSTOM";
            else
                sql += "  if "+m_AltPkgName;
            auth_flag = true;
        }
        else if ((p_Mode & RUN_SECURITY_INIT) != 0)
        {
            if (m_AltPkgName == null)
                sql += "  if OWA_INIT";
            else
                sql += "  if "+m_AltPkgName;
            auth_flag = true;
        }
        if (auth_flag) sql += ".AUTHORIZE then\n  ";

        /*
        ** Now build the actual procedure call, with any bound arguments
        */
        sql += ("  "+p_ProcName);

        if (num_args > 0)
        {
            sql += "(";
            for (i = 0; i < num_args; ++i)
            {
                if (((p_Mode & RUN_MODE_POSITIONAL) == 0) &&
                    ((p_Mode & RUN_MODE_FLEXIBLE) == 0))
                    sql += (arg_nams[i]+" => ");
                sql += "?";
                if ((i + 1) < num_args) sql += ", ";
            }
            if (lob_flag) sql += ", ?, ?, ?"; // Add output arguments
            sql += ")";
        }
        sql += ";\n  ";

        /*
        ** Now build the terminating code
        */
        if (p_MimeType == null)
        {
            if (auth_flag) sql += "  ";
            sql += "commit;\n";
        }
        if (auth_flag) sql += "  end if;\n";
        sql += "end;";

        /*
        ** Now there is a statement of the form
        ** The argument order will be:
        **   array argument 1 size Int, array argument 1, ...
        **   non-array argument 1, non-array argument 2, ...
        */
        m_LastStmt = sql;

        try
        {
            cs = (OracleCallableStatement)m_Conn.prepareCall(sql);

            if (num_args > 0)
            {
                int bind_pos = 0;

                vcarray = ArrayDescriptor.createDescriptor(ARRAY_TYPE, m_Conn);

                for (i = 0; i < num_args; ++i)
                    if (arg_vals[i].getClass() == arg_arr.getClass())
                    {
                      arg_arr = (String [])arg_vals[i];
                      oarray = new ARRAY(vcarray, m_Conn, arg_arr);
                      cs.setArray(++bind_pos, oarray);
                    }
                    else if (arg_vals[i].getClass() == "".getClass())
                    {
                      cs.setString(++bind_pos, (String)arg_vals[i]);
                    }
                    else if (arg_vals[i].getClass() != arg_arr.getClass())
                    {
                      cs.setInt(++bind_pos, ((Integer)arg_vals[i]).intValue());
                    }

                if (lob_flag)
                {
                    ++bind_pos;
                    if (p_MimeType != null) cs.setString(bind_pos, p_MimeType);
                    cs.registerOutParameter(bind_pos, OracleTypes.VARCHAR);
                    cs.registerOutParameter(++bind_pos, OracleTypes.BLOB);
                    cs.registerOutParameter(++bind_pos, OracleTypes.CLOB);
                }
            }

            cs.executeUpdate();

            m_MimeType = null;
            if (lob_flag)
            {
                /*
                ** Now read back the LOB output arguments
                */
                m_Blob = (BLOB)cs.getObject(6);
                if (cs.wasNull()) m_Blob = null;
                m_Clob = (CLOB)cs.getObject(7);
                if (cs.wasNull()) m_Clob = null;
                m_MimeType = cs.getString(5);
                if (cs.wasNull())
                    m_MimeType = null;
                else if (m_MimeType != null)
                    if (m_MimeType.length() == 0) m_MimeType = null;

                if (m_Blob != null)
                {
                    m_Clob = null;
                    if (m_MimeType == null)
                        m_MimeType = "application/octet-stream";
                }
                else if (m_Clob != null)
                {
                    m_Blob = null;
                    if (m_MimeType == null)
                        m_MimeType = "text/plain";
                }
                else
                    m_MimeType = null;
                }

            cs.close();
        }
        catch(SQLException e)
        {
            m_Conn.clearWarnings();
            throw(makeErrorMessage(e.getMessage()));
        }
    }

    /*
    ** Run OWA PL/SQL procedure with no mime-type in/out parameter
    */
    public void runProcedure(String p_ProcName, int p_Mode,
                             Vector p_Names, Vector p_Values)
         throws SQLException
    {
        runProcedure(p_ProcName, p_Mode, p_Names, p_Values, null);
    }

    /*
    ** Convert a hash-table argument structure into a vector of names
    ** and a vector of values
    */
    public void runProcedure(String p_ProcName, int p_Mode, Hashtable p_Args)
         throws SQLException
    {
        Vector names = new Vector(VECTOR_INIT_SIZE, VECTOR_INCR_SIZE);
        Vector values = new Vector(VECTOR_INIT_SIZE, VECTOR_INCR_SIZE);

        if (p_Args != null)
        {
            Enumeration evals = p_Args.elements();
            Enumeration ekeys = p_Args.keys();
            while (ekeys.hasMoreElements())
            {
                names.addElement(ekeys.nextElement());
                values.addElement(evals.nextElement());
            }
        }
        runProcedure(p_ProcName, p_Mode, names, values, null);
    }

    /*
    ** Returns the mime type from the LOB operation,
    ** or null if no return stream is available
    */
    public String getMimeType()
    {
        return(m_MimeType);
    }

    /*
    ** Returns an InputStream handle for reading the BLOB,
    ** or null if not available
    */
    public InputStream getBinarySource()
    {
        try
        {
            if (m_Blob != null) return(m_Blob.getBinaryStream());
        }
        catch(SQLException e)
        {
            System.out.println("OWA Error: "+e.getMessage());
        }
        return(null);
    }

    /*
    ** Returns a Reader handle for reading the CLOB,
    ** or null if not available
    */
    public Reader getCharacterSource()
    {
        try
        {
            if (m_Clob != null) return(m_Clob.getCharacterStream());
        }
        catch(SQLException e)
        {
            System.out.println("OWA Error: "+e.getMessage());
        }
        return(null);
    }

    /*
    ** Returns an OutputStream handle for writing the BLOB,
    ** or null if not available
    */
    public OutputStream getBinarySink()
    {
        try
        {
            if (m_Blob != null) return(m_Blob.getBinaryOutputStream());
        }
        catch(SQLException e)
        {
            System.out.println("OWA Error: "+e.getMessage());
        }
        return(null);
    }

    /*
    ** Returns a Writer handle for writing the CLOB,
    ** or null if not available
    */
    public Writer getCharacterSink()
    {
        try
        {
            if (m_Clob != null) return(m_Clob.getCharacterOutputStream());
        }
        catch(SQLException e)
        {
            System.out.println("OWA Error: "+e.getMessage());
        }
        return(null);
    }
}
