/*
** Copyright (c) 1999, 2000 Oracle Corporation, All rights reserved.
**
** Class:  Owa
** Author: Doug McMahon
** Date:   August 12, 2000
**
** Description:
**
**  OWA gateway servlet
**
** Methods:
**
**  makeHeaderTable     Set up table of standard Servlet header element names
**  makeInitCap         Convert string to lower case with initial caps
**  countChars          Count complete characters in multi-byte
**  findBoundary        Find a boundary or partial boundary value in a buffer
**  putResponse         Put OWA header values into Servlet response
**  setCGI              Set the CGI-environment equivalent for the OWA
**  setProcedureName    Set package/procedure name and argument mode for OWA
**  streamFile          Handle a streaming file download out of the OWA
**  showErrorPage       Show a generic error page
**  renderPage          Get and render a normal OWA page
**  uploadFiles         Handle a multi-part file upload to the OWA
**  showPage            Execute procedure and show the result
**  parseArguments      Parse HTTP arguments into OWA-compatible format
**  loadParameters      Set servlet parameters from servlet context
**  doGet               Handle GET request
**  doPost              Handle POST request
**  doPut               Handle PUT request
**  init                Initialize servlet parameters
**
**
** Configuration Parameters:
**
**  Parameter        Example              Description
**  --------------------------------------------------------------------------
**  Required:
**    database       localhost:1521:ORCL  Connection info <host>:<port>:<SID>
**    dbusername     scott                Database username
**    dbpassword     tiger                Database password
**  Optional:
**    document_path  docs                 URL prefix for document reads
**    document_proc  mypackage.readfile   Document read procedure
**    default_page   mypackage.showpage   Default page
**    pool_size      10                   Connection pool size
**    auth_mode      PACKAGE              Specify PL/SQL authorization mode
**    alternate_owa  MY_OWA               Optional OWA alternative
**    dad_charset    iso-8859-1           Character set (default "iso-8859-1")
**
**  Example web.xml configuration file settings:
**
**    <servlet>
**      <servlet-name>owa</servlet-name>
**      <servlet-class>Owa</servlet-class>
**      <init-param>
**        <param-name>database</param-name>
**        <param-value>localhost:1521:ORCL</param-value>
**      </init-param>
**      <init-param>
**        <param-name>dbusername</param-name>
**        <param-value>scott</param-value>
**      </init-param>
**      <init-param>
**        <param-name>dbpassword</param-name>
**        <param-value>tiger</param-value>
**      </init-param>
**      <init-param>
**        <param-name>dbpassword</param-name>
**        <param-value>owa</param-value>
**      </init-param>
**    </servlet>
**    <servlet-mapping>
**      <servlet-name>owa</servlet-name>
**        <url-pattern>/owa/*</url-pattern>
**    </servlet-mapping>
*/
import java.io.*;
import java.text.*;
import java.util.*;
import java.lang.*;

import java.sql.*;
import oracle.sql.*;
import oracle.jdbc.driver.*;

import javax.servlet.*;
import javax.servlet.http.*;

/*
** Print Equivalent of CGI environment
*/

public class Owa extends HttpServlet
{
    /*
    ** Global to all objects
    */
    private static Hashtable    s_CGIheaders;   // Translations to CGI

    /*
    ** Constants
    */
    final static int MAX_OWA_ARRAY_SIZE = 256;   // Max size of an OWA_ARRAY
    final static int HEADER_TABLE_SIZE  = 20;    // Number of header elements
    final static int LOB_BUFFER_SIZE    = 65536; // Max size of LOB buffer

    final static int VECTOR_INIT_SIZE  = 10;      // Initial size of pool
    final static int VECTOR_INCR_SIZE  = 10;      // Pool size increment

    final static int CHARSET_SINGLEBYTE = 0; // All single-byte sets
    final static int CHARSET_UNICODE    = 1; // UTF-8
    final static int CHARSET_ASIAN4     = 2; // 4-byte Asian sets
    final static int CHARSET_JEUC3      = 3; // JEUC-based sets
    final static int CHARSET_SJIS2      = 4; // SJIS-based sets
    final static int CHARSET_ASIAN2     = 5; // Asian sets w/out 0x80/0xff
    final static int CHARSET_MULTIBYTE  = 6; // All other 2-byte sets

    /*
    ** Configuration parameters
    **
    ** ### Add support for variable LOB arguments and BFILEs
    */
    private String m_db_string  = "localhost:1521:ORCL";
    private String m_db_user    = null;
    private String m_db_pass    = null;
    private String m_doc_path   = "";
    private String m_doc_proc   = null;
    private String m_def_page   = null;
    private String m_alt_owa    = null;
    private String m_dad_cs     = "iso-8859-1";
    private int    m_auth_mode  = GatewayManager.RUN_SECURITY_NONE;
    private int    m_pool_size  = 10;
    private int    m_cs_mode    = CHARSET_SINGLEBYTE;

    /*
    ** Makes a lookup table to convert header elements from
    ** servlet names to Apache CGI names
    */
    private static synchronized void makeHeaderTable()
    {
        if (s_CGIheaders == null)
        {
            s_CGIheaders = new Hashtable(HEADER_TABLE_SIZE);
            s_CGIheaders.put("Host",            "HTTP_HOST");
            s_CGIheaders.put("Accept",          "HTTP_ACCEPT");
            s_CGIheaders.put("Accept-Encoding", "HTTP_ACCEPT_ENCODING");
            s_CGIheaders.put("Accept-Charset",  "HTTP_ACCEPT_CHARSET");
            s_CGIheaders.put("Accept-Language", "HTTP_ACCEPT_LANGUAGE");
            s_CGIheaders.put("Connection",      "HTTP_CONNECTION");
            s_CGIheaders.put("User-Agent",      "HTTP_USER_AGENT");
            s_CGIheaders.put("Referer",         "HTTP_REFERER");
//          s_CGIheaders.put("Cookie",          "HTTP_COOKIE");
            s_CGIheaders.put("From",            "HTTP_FROM");
            s_CGIheaders.put("Server",          "SERVER_SOFTWARE");
        }
    }

    /*
    ** Convert a header element string to init-cap format to ensure
    ** hash table matches, etc.
    */
    private static String makeInitCap(String p_Name)
    {
        int          i;
        StringBuffer icap = new StringBuffer(p_Name);
        int          nlen = icap.length();
        boolean      cap_flag = true;

        for (i = 0; i < nlen; ++i)
        {
            char c = icap.charAt(i);

            if (Character.isLetterOrDigit(c))
            {
                if (Character.isLowerCase(c))
                {
                    if (cap_flag)
                        icap.setCharAt(i, Character.toUpperCase(c));
                }
                else if (Character.isUpperCase(c))
                {
                    if (!cap_flag)
                        icap.setCharAt(i, Character.toLowerCase(c));
                }
                cap_flag = false;
            }
            else
            {
                cap_flag = true;
            }
        }
        return(icap.toString());
    }

    /*
    ** Count number of complete characters in a buffer.
    ** Returns the number of bytes consumed from the buffer.
    ** If needed, the character count is left in m_NumChars.
    */
    private long countChars(byte[] p_Buffer, int p_BufLen)
    {
        int amount;
        int i, j, k, n;

        if (m_cs_mode == CHARSET_SINGLEBYTE)
            return((long)p_BufLen | ((long)p_BufLen << 32));

        amount = 0;
        i = p_BufLen;
        j = 0;

        if (m_cs_mode == CHARSET_UNICODE)
        {
            /*
            ** Algorithm for all variable-width Unicode variants
            */
            while (j < i)
            {
                k = p_Buffer[j] & 0xFF;
                if (k < 0x80)
                {
                    ++j;
                    ++amount;
                }
                else if (k < 0xC0) /* Error: naked trailing byte */
                {
                    ++j;
                    ++amount;
                }
                else
                {
                    if (k < 0xE0)      n = 2;
                    else if (k < 0xF0) n = 3;
                    else if (k < 0xF8) n = 4;
                    else if (k < 0xFC) n = 5;
                    else if (k < 0xFE) n = 6;
                    else               n = 1; /* Error: unknown byte */

                    if ((i - j) >= n)
                    {
                        j += n;
                        ++amount;
                    }
                    else
                        i = j; /* Trailing fragment not consumed */
                }
            }
            /* Fall through the rest of the routine to the bottom */
        }

        /*
        ** The default implementation assumes single-byte below 0x80,
        ** all others are leaders of double-byte sequences.  There are
        ** special cases coded for:
        **  1) the 2/4-byte Chinese sets, which use 0x8e as a 4-byte leader
        **  2) the 2/3-byte JEUC sets, which use 0x8f as a 3-byte leader
        **  3) the 2-byte Asian sets, which don't use 0x80 or 0xff
        **  4) the Shift-JIS sets, which have some 1-byte values above 0x7f
        */
        while (j < i)
        {
            k = p_Buffer[j] & 0xFF;
            if (k < 0x80)
            {
                ++j;
                ++amount;
            }
            else
            {
                /*
                ** Seems to apply to:  JA16VMS, KO16KSC5601
                **
                ** ### Old experments show it also appears to work for:
                **     GB2312, JA16EUC
                */
                n = 2;
                /*
                ** This algorithm is used for the 4-byte sets.  Again there
                ** is a discrepancy between the documentation and experiment.
                ** Documentation:
                **
                **   Range      ZHT32EUC   ZHT32TRIS  ZHT32SOPS  CNS1164386
                **   ------------------------------------------------------
                **   0x80-0x8d  undefined  undefined  2?         2?
                **   0x8e       4-byte     4-byte     4-byte     4-byte
                **   0x8f-0xa0  undefined  undefined  2?         2?
                **   0xa1-0xfd  2-byte     undefined  2?         2?
                **   0xfe-0xff  undefined  undefined  2?         2?
                **
                ** The implementation maps all undefined values to 2.
                */
                if ((k == 0x8E) && (m_cs_mode == CHARSET_ASIAN4))
                    n = 4;
                /*
                ** Once again, there are some discrepancies between experiment
                ** and documentation:
                **
                **   Range      Experiment   Documentation   Implementation
                **   ------------------------------------------------------
                **   0x80-0x8e  2-byte lead  undefined       2-byte
                **   0x8f       3-byte lead  3-byte lead     3-byte
                **   0xa0       2-byte lead  undefined       2-byte
                **   0xa1-0xfe  2-byte lead  2-byte lead     2-byte
                **   0xff       2-byte lead  undefined       2-byte
                */
                else if ((k == 0x8F) && (m_cs_mode == CHARSET_JEUC3))
                    n = 3;
                /*
                ** For GB2312, this is an approximation
                ** (because 80-a0 and f8-ff are undefined).
                **
                ** An identical algorithm also works for:
                **   ZHT16MSWIN950, KO16MSWIN949, ZHT16HKCSC
                **
                ** Plus, the algorithm can be used for:
                **   KO16KSCCS (although 80-83 and fa-ff are undefined)
                */
                else if (((k == 0x80) || (k == 0xFF)) &&
                         (m_cs_mode == CHARSET_ASIAN2))
                    n = 1;
                /*
                ** All forms of Shift-JIS.  Unfortunately there appear to be
                ** some discrepancies between documentation and experiment:
                **
                **   Range      Experiment   Documentation   Implementation
                **   ------------------------------------------------------
                **   0x80       single-byte  undefined       1-byte
                **   0x81-0x9f  2-byte lead  2-byte lead     2-byte
                **   0xa0       single-byte  undefined       1-byte
                **   0xa1-0xdf  single-byte  single-byte     1-byte
                **   0xe0-0xfc  2-byte lead  2-byte lead     2-byte
                **   0xfe-0xff  2-byte lead  undefined       2-byte
                */
                else if (((k == 0x80) || ((k >= 0xA0) && (k < 0xE0))) &&
                         (m_cs_mode == CHARSET_SJIS2))
                    n = 1;

                if ((i - j) >= n)
                {
                    j += n;
                    ++amount;
                }
                else
                    i = j; /* Trailing fragment not consumed */
            }
        }

        return((long)i | ((long)amount << 32));
    }

    /*
    ** Scan a byte buffer for a boundary or part of a boundary
    */
    private static int findBoundary(int p_Start,
                                    byte[] p_Buffer, int p_BufLen,
                                    byte[] p_Pattern, int p_PatLen)
    {
        int midx;
        int plen;
        int mlen;

        plen = p_PatLen;
        mlen = p_BufLen - p_Start;

        if ((plen > 0) && (p_Buffer != null) && (p_Pattern != null))
            for (midx = p_Start; mlen > 0; --mlen, ++midx)
            {
                int i;
                if (plen > mlen) plen = mlen;
                for (i = 0; i < plen; ++i)
                    if (p_Buffer[i + midx] != p_Pattern[i])
                        break;
                if (i == plen) return(midx);
            }

        return(-1);
    }

    /*
    ** Transfer response header elements to the servlet response object.
    ** Return a status that will be either SC_OK, SC_MOVED_TEMPORARILY,
    ** or SC_UNAUTHORIZED.
    */
    private int putResponse(HttpServletResponse p_Response, Vector p_Hdr)
         throws IOException
    {
        int    status = p_Response.SC_OK;
        int    nhead  = (p_Hdr == null) ? 0 : p_Hdr.size();
        String rurl   = null;
        String rhdr   = null;
        int    ncookies = 0;

        while (nhead > 0)
        {
            String s = (String)p_Hdr.elementAt(--nhead);
            int    i = s.indexOf(':');
            if (i > 0)
            {
                String name;
                String value;
                name = s.substring(0, i++);
                while (s.charAt(i) == ' ') ++i;
                value = s.substring(i);

                if (name.equalsIgnoreCase("Location") ||
                    name.equalsIgnoreCase("Refresh"))
                {
                    rhdr = name;
                    // Deprecated - replace with encodeRedirectURL()
                    rurl = p_Response.encodeRedirectUrl(value);
                    status = p_Response.SC_MOVED_TEMPORARILY;
                }
                else if (name.equalsIgnoreCase("Content-Type"))
                {
                    p_Response.setContentType(value);
                }
                else if (name.equalsIgnoreCase("Content-Length"))
                {
                    try
                    {
                        int ival = Integer.parseInt(value);
                        p_Response.setContentLength(ival);
                    }
                    catch(NumberFormatException e)
                    {
                        System.out.println("DBG: "+e.getMessage()); // ###
                    }
                }
                else if (name.equalsIgnoreCase("Set-Cookie"))
                {
                    /*
                    ** Parse the cookie string for a series of cookies
                    ** Assumes the format is Name1=Value1;Name2=Value2;
                    */
                    StringTokenizer tok;
                    tok = new StringTokenizer(value, "=;", false);
                    while (tok.countTokens() >= 2)
                    {
                        String cookie_name = tok.nextToken();
                        String cookie_value = tok.nextToken();
                        Cookie cookie = new Cookie(cookie_name, cookie_value);
                        p_Response.addCookie(cookie);
                        ++ncookies;
                    }
                }
                else
                {
                    if (name.equalsIgnoreCase("WWW-Authenticate"))
                        status = p_Response.SC_UNAUTHORIZED;
                    p_Response.setHeader(name, value);
                }
            }
        }

        p_Response.setStatus(status);

        if (rurl != null)
        {
            /*
            ** ### Can't use sendRedirect() because then TomCat
            ** ### doesn't send the cookie back.
            */
            if (ncookies > 0) p_Response.setHeader(rhdr, rurl);
            else              p_Response.sendRedirect(rurl);
        }
        else if (status == p_Response.SC_UNAUTHORIZED)
            p_Response.sendError(status);

        return(status);
    }

    /*
    ** Set the CGI environment arrays from a servlet request context,
    ** then send them to the OWA
    */
    private void setCGI(HttpServletRequest p_Request, GatewayManager p_Gateway)
    {
        Enumeration e;
        Cookie[]    cookies;
        int         sz;
        int         nenv = 0;
        String[]    nams;
        String[]    vals;

        nams = new String[MAX_OWA_ARRAY_SIZE];
        vals = new String[MAX_OWA_ARRAY_SIZE];

        /*
        ** Must guarantee at least one value in the array(s).
        ** Also ensure the minimum bind width will be at least 2 chars
        ** to avoid an OCI bug.
        */
        nams[nenv]   = "GATEWAY_INTERFACE";
        vals[nenv++] = "CGI/1.1";

        /*
        ** Best-guess translations of standard servlet
        ** interfaces to Apache-compatible CGI names
        */
        nams[nenv]   = "SERVER_NAME";
        vals[nenv++] = p_Request.getServerName();
        nams[nenv]   = "SERVER_PORT";
        vals[nenv++] = String.valueOf(p_Request.getServerPort());
        nams[nenv]   = "SERVER_PROTOCOL";
        vals[nenv++] = p_Request.getProtocol();
        nams[nenv]   = "REMOTE_ADDR";
        vals[nenv++] = p_Request.getRemoteAddr();
        nams[nenv]   = "REMOTE_USER";
        vals[nenv++] = p_Request.getRemoteUser();
        nams[nenv]   = "REMOTE_HOST";
        vals[nenv++] = p_Request.getRemoteHost();
        nams[nenv]   = "REQUEST_METHOD";
        vals[nenv++] = p_Request.getMethod();
        nams[nenv]   = "SCRIPT_NAME";
        vals[nenv++] = p_Request.getServletPath();
        nams[nenv]   = "QUERY_STRING";
        vals[nenv++] = p_Request.getQueryString();
        nams[nenv]   = "CONTENT_LENGTH";
        vals[nenv++] = String.valueOf(p_Request.getContentLength());
        nams[nenv]   = "CONTENT_TYPE";
        vals[nenv++] = p_Request.getContentType();
        nams[nenv]   = "AUTH_TYPE";
        vals[nenv++] = p_Request.getAuthType();
        nams[nenv]   = "REQUEST_URI";
        vals[nenv++] = p_Request.getRequestURI();
        nams[nenv]   = "PATH_INFO";
        vals[nenv++] = p_Request.getPathInfo();
        nams[nenv]   = "PATH_TRANSLATED";
        vals[nenv++] = p_Request.getPathTranslated();
        nams[nenv]   = "SCRIPT_FILENAME";
        vals[nenv++] = p_Request.getPathTranslated();
        nams[nenv]   = "REQUEST_IANA_CHARSET";
        vals[nenv++] = p_Request.getCharacterEncoding();

        /*
        ** Get all the cookies and rebuild them as a series of
        ** HTTP_COOKIE values
        ** ### SHOULD WE USE THE VALUE OF getHeader("Cookie") INSTEAD?
        */
        cookies = p_Request.getCookies();
        if (cookies != null)
        {
            String allCookies = "";
            nams[nenv] = "HTTP_COOKIE";

            sz = java.lang.reflect.Array.getLength(cookies);
            while (sz > 0)
            {
                --sz;
                allCookies = allCookies +
                             cookies[sz].getName()+"="+cookies[sz].getValue();
                if (sz > 0) allCookies = allCookies+";";
            }
            vals[nenv] = allCookies;
            ++nenv;
        }
        
        /*
        ** No translations yet known for:
        **
        **   SERVER_ADDR
        **   SERVER_ADMIN
        **   SERVER_SIGNATURE
        **
        **   REMOTE_IDENT
        **   REMOTE_PORT
        **
        **   PATH
        **   DOCUMENT_ROOT     (e.g. the file system path to "htdocs")
        **   DOCUMENT_URI
        **   DOCUMENT_NAME
        */

        /*
        ** Looks for hash-table matches
        ** ### WHAT ABOUT THE value for Cookie?
        */
        makeHeaderTable();
        e = p_Request.getHeaderNames();
        while (e.hasMoreElements())
        {
            String headerName = (String)e.nextElement();
            String matchName  = makeInitCap(headerName);
            String cgiName    = (String)s_CGIheaders.get(matchName);

            if (cgiName != null)
            {
                String cgiValue = p_Request.getHeader(headerName);
                if (cgiValue != null)
                {
                    nams[nenv] = cgiName;
                    vals[nenv] = cgiValue;
                    if (++nenv == MAX_OWA_ARRAY_SIZE) break;
                }
            }
        }

        p_Gateway.setCGI(nams, vals, nenv);
    }

    /*
    ** Parse the "script name" to determine the package/procedure
    ** name and the argument passing mode.
    */
    private int setProcedureName(HttpServletRequest p_Request,
                                 String[] p_Result)
    {
        int     proc_loc;
        int     mode;
        boolean docflag = false;
        String  prefix = "/"+m_doc_path;
        int     plen = prefix.length();
        String  path = p_Request.getPathInfo();

        if (path == null) path = "";

        /*
        ** If the first portion of path (up to the first "/", if any)
        ** matches the doc_path prefix (with a leading "/" added),
        ** then this is a document read (but only if document_proc
        ** is set).
        */
        if ((m_doc_proc != null) &&
            (path.regionMatches(false, 0, prefix, 0, plen)))
        {
            if (path.length() == plen)         docflag = true;
            else if (path.charAt(plen) == '/') docflag = true;
        }

        /*
        ** If the path is null or just "/", set to the default page
        ** or, if no default page, use the document procedure.
        */
        if (path.length() <= 1)
        {
            if (m_def_page != null)
                path = m_def_page;
            else
            {
                docflag = (m_doc_proc != null);
                path = "/";
            }
        }

        /* Check for document read */
        if (docflag)
        {
            mode = GatewayManager.RUN_MODE_LOBREAD;
            p_Result[0] = path;
        }
        else if (path.length() > 1)
        {
            proc_loc = 0;
            if (path.charAt(proc_loc) == '/') ++proc_loc;
            switch (path.charAt(proc_loc))
            {
            case '!': // Flexible arguments
                ++proc_loc;
                mode = GatewayManager.RUN_MODE_FLEXIBLE;
                break;
            case '^': // Positional arguments
                ++proc_loc;
                mode = GatewayManager.RUN_MODE_POSITIONAL;
                break;
            case '~': // Reserved
                ++proc_loc;
                /* FALLTHROUGH */
            default:
                mode = GatewayManager.RUN_MODE_NORMAL;
                break;
            }

            if (proc_loc > 0) path = path.substring(proc_loc);

            p_Result[0] = path;
        }
        else
        {
            mode = GatewayManager.RUN_MODE_NORMAL;
            p_Result[0] = null;
        }
        return(mode);
    }

    /*
    ** Handle a file download
    */
    private void streamFile(HttpServletResponse p_Response,
                            String p_MimeType, GatewayManager p_Gateway)
          throws IOException
    {
        InputStream bstream = p_Gateway.getBinarySource();
        Reader      cstream = p_Gateway.getCharacterSource();
        int         nread;

        p_Response.setContentType(p_MimeType);
        p_Response.setStatus(p_Response.SC_OK);

        if (bstream != null)
        {
            byte[] buf            = new byte[LOB_BUFFER_SIZE];
            ServletOutputStream s = p_Response.getOutputStream();
            nread = LOB_BUFFER_SIZE;
            while (nread == LOB_BUFFER_SIZE)
            {
                nread = bstream.read(buf);
                if (nread > 0) s.write(buf, 0, nread);
            }
        }
        else
        {
            char[]      buf = new char[LOB_BUFFER_SIZE];
            PrintWriter out = p_Response.getWriter();
            nread = LOB_BUFFER_SIZE;
            while (nread == LOB_BUFFER_SIZE)
            {
                 nread = cstream.read(buf, 0, LOB_BUFFER_SIZE);
                 if (nread > 0) out.write(buf, 0, nread);
            }
        }
    }

    /*
    ** Generic routine to show an error page with an embedded string
    */
    private void showErrorPage(HttpServletResponse p_Response, String p_ErrStr)
         throws IOException
    {
        p_Response.setContentType("text/html");
        PrintWriter out = p_Response.getWriter();
        out.println("<html><head><title>OWA Error Page</title></head>");
        out.println("<body bgcolor=\"#ffffff\" text=\"#c00000\">");
        out.println("<b><pre>"+p_ErrStr+"</pre></b>");
        out.println("</body></html>");
    }

    /*
    ** Get and render a page from the OWA
    */
    private void renderPage(HttpServletResponse p_Response,
                            GatewayManager p_Gateway)
         throws IOException, SQLException
    {
        PrintWriter         out = null;
        ServletOutputStream st = null;

        p_Response.setContentType("text/html"); // default

        int status = putResponse(p_Response, p_Gateway.getHeader());

        while (true)
        {
            String s = p_Gateway.getContentBlock();
            byte[] b = p_Gateway.getRawContentBlock();
            if ((s != null) && (st == null))
            {
                if (out == null) out = p_Response.getWriter();
                out.print(s);
            }
            else if ((b != null) && (out == null))
            {
                if (st == null) st = p_Response.getOutputStream();
                st.write(b, 0, java.lang.reflect.Array.getLength(b));
            }
            else
                break;
        }
    }

    /*
    ** Handle a streaming content upload
    */
    private void uploadFiles(HttpServletRequest p_Request,
                             HttpServletResponse p_Response,
                             ServletInputStream p_InStream, String p_Boundary)
          throws IOException, SQLException
    {
        byte[]       buf = new byte[LOB_BUFFER_SIZE * 2];
        byte[]       boundary;
        int          mode;
        String       procname;
        Vector       names  = new Vector(VECTOR_INIT_SIZE, VECTOR_INCR_SIZE);
        Vector       values = new Vector(VECTOR_INIT_SIZE, VECTOR_INCR_SIZE);
        Vector       fsizes = new Vector(VECTOR_INIT_SIZE, VECTOR_INCR_SIZE);
        OutputStream bstream = null;
        Writer       cstream = null;
        int          blen;
        int          clen = p_Request.getContentLength();
        int          bloc = -1;
        int          offset;
        int          nbytes;
        int          total = 0;
        String       s;
        String[]     sarr;

        boundary = p_Boundary.getBytes(m_dad_cs);
        blen = java.lang.reflect.Array.getLength(boundary);

        /*
        ** Construct procedure directly, no authorization modes.  Use
        ** the LOBREAD mode, since the argument passing identical.
        */
        s = p_Request.getPathInfo();
        procname = (s.charAt(0) == '/') ? s.substring(1) : s;
        mode = GatewayManager.RUN_MODE_LOBREAD;

        /*
        ** Set up the CGI
        */
        ConnectionPool cp = ConnectionPool.getInstance();
        cp.setMaximum(m_pool_size);
        String db = cp.makeDatabaseString("thin", m_db_string);
        Connection conn = cp.getConnection(m_db_user, m_db_pass, db);
        GatewayManager gwy = new GatewayManager(conn, m_alt_owa);
        conn.setAutoCommit(false); // ### USE EVERYWHERE!

        setCGI(p_Request, gwy);
        gwy.sendCGI();

        /*
        ** First argument is always the document path
        ** Second argument reserved for file name
        */
        sarr = new String[1];
        sarr[0] = m_doc_path;
        names.addElement("document_path");
        values.addElement(sarr);
        sarr = new String[1];
        sarr[0] = "";
        names.addElement("");
        values.addElement(sarr);

        /*
        ** Now loop, reading and parsing the content stream
        */
        System.out.println("DBG> Content length "+clen+" bytes"); // ###
        buf[0] = (byte)'\r';
        buf[1] = (byte)'\n';
        offset = 2;
        while ((clen > 0) || (offset > 0))
        {
            if ((offset < LOB_BUFFER_SIZE) && (clen > 0))
            {
                nbytes = LOB_BUFFER_SIZE;
                if (nbytes > clen) nbytes = clen;
                nbytes = p_InStream.read(buf, offset, nbytes);
                if (nbytes < 0)
                {
                    nbytes = 0;
                    clen = 0;
                }
                offset += nbytes;
                clen -= nbytes;
		System.out.println("DBG> Read "+nbytes+" into buffer"); // ###
            }
            if ((offset < blen) && (clen > 0)) continue;

            /*
            ** ### NOTE: THIS CODE, AND THAT OF modowa.c, IS INEFFICIENT
            ** ### FOR SEVERAL REASONS:
            ** ### - REPEATEDLY RE-CHECKS THE INITIAL PART OF THE BUFFER
            ** ###   EVEN THOUGH THE continue IS JUST CAUSING REPEAT MATCHES.
            ** ### - SHIFTS AT THE BOTTOM OF THE BUFFER ARE LARGER THAN
            ** ###   NECESSARY BECAUSE OF THE REQUIREMENT THAT THE BUFFER
            ** ###   CONTAIN AT LEAST 32K OF DATA
            */
            if (bloc != 0)
                bloc = findBoundary(0, buf, offset, boundary, blen);

            if ((offset < LOB_BUFFER_SIZE) && (clen > 0) && (bloc == 0))
                continue;

            nbytes = (bloc < 0) ? offset : bloc;
            if (nbytes > 0)
            {
                int amount = nbytes;

                /*
                ** Write data to binary or character LOB,
                ** or ignore the data if no sink available
                */
                if (bstream != null)
                {
                    bstream.write(buf, 0, nbytes);
                }
                else if (cstream != null)
                {
                    long count;
                    count = countChars(buf, nbytes);
                    amount = (int)(count & 0xFFFFFFFF);
                    /* Leave fractional character for next iteration */
                    if ((bloc >= 0) && (amount != nbytes)) nbytes = amount;
                    cstream.write(new String(buf, 0, nbytes, m_dad_cs));
                    amount = (int)(count >> 32);
                }
                total += amount;
		System.out.println("DBG> Wrote "+amount+" to LOB"); // ###
            }

            /*
            ** If a boundary's been found, perform a commit
            */
            if ((bloc >= 0) && ((offset - bloc) >= blen))
            {
		System.out.println("DBG> Closing LOB"); // ###
                if (bstream != null)
                {
                    bstream.close();
                    conn.commit();
                }
                else if (cstream != null)
                {
                    cstream.close();
                    conn.commit();
                }
                bstream = null;
                cstream = null;
                fsizes.addElement(String.valueOf(total));
            }

            if (bloc == 0)
            {
                /*
                ** Boundary has been found, aligned at beginning of buffer
                */
                String fieldname = null;
                String filename  = null;
                String mimetype  = "";
                int    cur_loc;
                int    lin_loc;
                int    end_loc;

                /*
                ** If immediately followed by --, this is the terminator
                */
                cur_loc = blen;
                if ((buf[cur_loc] == (byte)'-') &&
                    (buf[cur_loc + 1] == (byte)'-'))
                    break;
                cur_loc += 2;

                /*
                ** Find the \n at the end of the Content-Disposition line
                */
                for (lin_loc = cur_loc; lin_loc < offset; ++lin_loc)
                    if (buf[lin_loc] == (byte)'\n') break;
                if (lin_loc == offset) break; // ### ERROR
                end_loc = lin_loc;
                buf[end_loc - 1] = (byte)';';

                /*
                ** Skip "Content-Disposition: form-data;"
                */
                cur_loc += 31;
                if (cur_loc >= end_loc) break; // ### ERROR
                s = new String(buf, cur_loc, end_loc - cur_loc, m_dad_cs);

                /*
                ** String is now one of two forms:
                ** 1.  name="ordinary_field_name";
                ** 2.  name="filefield_name"; filename="filename.ext";
                */
                cur_loc = 0;
                while (s.charAt(cur_loc) == ' ') ++cur_loc;
                if (!s.regionMatches(true, cur_loc, "name=\"", 0, 6))
                    break; // ### ERROR
                cur_loc += 6;
                end_loc = s.indexOf('"', cur_loc);
                if (end_loc < 0) break; // ### ERROR
                fieldname = s.substring(cur_loc, end_loc);

                cur_loc = end_loc + 1;
                if (s.charAt(cur_loc++) != ';') break; // ### ERROR
                if (s.length() > cur_loc)
                {
                    filename = "";
                    while (s.charAt(cur_loc) == ' ') ++cur_loc;
                    if (!s.regionMatches(true, cur_loc, "filename=\"", 0, 10))
                        break; // ### ERROR
                    cur_loc += 10;
                    end_loc = s.indexOf('"', cur_loc);
                    if (end_loc > cur_loc)
                        filename = s.substring(cur_loc, end_loc);
                }

                /*
                ** Optionally, the next line is "Content-Type: <mime_type>"
                */
                ++lin_loc;
                for (cur_loc = lin_loc; lin_loc < offset; ++lin_loc)
                    if (buf[lin_loc] == (byte)'\n') break;
                if (lin_loc == offset) break; // ### ERROR
                end_loc = lin_loc - 1;
                if ((end_loc - cur_loc) > 14)
                {
                    cur_loc += 14;
                    mimetype = new String(buf, cur_loc,
                                          end_loc - cur_loc, m_dad_cs);
                    cur_loc = ++lin_loc;
                    ++lin_loc;
                }

                while (buf[lin_loc] != (byte)'\n')
                {
                    while ((lin_loc < offset) && (buf[lin_loc] != (byte)'\n'))
                        ++lin_loc;
                    if (lin_loc == offset) break; // ### ERROR
                    lin_loc += 2;
                }
                if (buf[lin_loc] != (byte)'\n') break; // ### ERROR

                ++lin_loc;

                if (filename == null)
                {
                    /*
                    ** Scan for ending boundary
                    */
                    end_loc = findBoundary(lin_loc,
                                           buf, offset, boundary, blen);
                    if (end_loc < 0) break; // ### ERROR

                    String val = new String(buf, lin_loc,
                                            end_loc - lin_loc, m_dad_cs);

                    sarr = new String[1];
                    sarr[0] = val;
                    names.addElement(fieldname);
                    values.addElement(sarr);
                    System.out.println("DBG> Upload File"); // ###
                    System.out.println("DBG>  Field "+fieldname+" => "+val);
                    nbytes = end_loc;
                }
                else
                {
                    sarr = new String[1];
                    sarr[0] = filename;
                    names.setElementAt(fieldname, 1);
                    values.setElementAt(sarr, 1);
                    System.out.println("DBG> Upload File");              // ###
                    System.out.println("DBG>   Fieldname = "+fieldname); // ###
                    System.out.println("DBG>   Filename = "+filename);   // ###
                    System.out.println("DBG>   Mimetype = "+mimetype);   // ###
                    nbytes = lin_loc; 

                    /*
                    ** Run the PL/SQL procedure
                    */
                    gwy.runProcedure(procname, mode,
                                     names, values, mimetype);
                    bstream = gwy.getBinarySink();
                    cstream = gwy.getCharacterSink();
                    total = 0;
                }
            }
            else if (bloc > 0)
            {
                System.out.println("DBG> Boundary found at "+bloc); // ###
            }

            offset -= nbytes;
            if (offset > 0)
                System.arraycopy(buf, nbytes, buf, 0, offset);
	    System.out.println("DBG> Buffer has "+offset+" bytes"); // ###
            bloc = -1;
        }

        /*
        ** Run the termination operation
        */
        sarr = new String[1];
        sarr[0] = "";
        names.setElementAt("", 1);
        values.setElementAt(sarr, 1);
        for (total = 0; total < fsizes.size(); ++total)
        {
            sarr = new String[1];
            sarr[0] = (String)fsizes.elementAt(total);
            names.addElement("content_length");
            values.addElement(sarr);
        }
        gwy.runProcedure(procname, mode, names, values, "");

        /*
        ** Render the response page and then close the connection
        */
        renderPage(p_Response, gwy);
        gwy.setConnection(null);
        cp.releaseConnection(conn);
    }

    /*
    ** Execute an OWA procedure and show the response
    ** 1. Set and send the CGI environment
    ** 2. Call the procedure
    ** 3. Perform streaming file download (if this is a file operation)
    ** 3. Get and display an OWA page (if this is a normal operation)
    */
    private void showPage(HttpServletRequest p_Request,
                          HttpServletResponse p_Response,
                          int p_Mode, String p_Procedure,
                          Vector p_ArgNames, Vector p_ArgValues)
         throws IOException
    {
        if (p_Procedure == null)
        {
            showErrorPage(p_Response, "Could not determine procedure name");
            return;
        }

        try
        {
            ConnectionPool cp = ConnectionPool.getInstance();
            cp.setMaximum(m_pool_size);
            String db = cp.makeDatabaseString("thin", m_db_string);
            Connection conn = cp.getConnection(m_db_user, m_db_pass, db);
            GatewayManager gwy = new GatewayManager(conn, m_alt_owa);
            conn.setAutoCommit(false); // ### USE EVERYWHERE!

            setCGI(p_Request, gwy);
            gwy.sendCGI();
            gwy.runProcedure(p_Procedure, p_Mode, p_ArgNames, p_ArgValues);

            String mtype = gwy.getMimeType();

            if (mtype != null)
                streamFile(p_Response, mtype, gwy);
            else
                renderPage(p_Response, gwy);

            gwy.setConnection(null);
            cp.releaseConnection(conn);
        }
        catch(SQLException e)
        {
            showErrorPage(p_Response, "SQL failure: "+e.getMessage());
        }
    }

    /*
    ** Parse query string into argument vectors
    ** This routine is needed because unfortunately the default
    ** parsers provided by HttpUtils return Hashtables, which aren't
    ** ordered and therefore won't support positional and (possibly)
    ** flexible argument modes of the OWA.
    **
    ** First use HttpUtils to parse the string and unescape the URL.
    ** Then pass through the string looking for names and constructing
    ** a vector of names.  For each name found, look in the Hashtable
    ** for a matching key, and if found, add the parsed string array
    ** to the vector of values.
    **
    ** ### NEED TO FIX "Button.x"/"Button.y" PROBLEM WITH IMAGE ITEMS.
    */
    private void parseArguments(String p_Query, int p_Mode, String p_Document,
                                Vector p_ArgNames, Vector p_ArgValues)
    {
        Hashtable args = HttpUtils.parseQueryString(p_Query);

        if (p_Document != null)
        {
            String[] sarr = new String[1];
            sarr[0] = p_Document;
            p_ArgNames.addElement(GatewayManager.DOCUMENT_KEY);
            p_ArgValues.addElement(sarr);
        }

        if (args != null)
        {
            int cur_loc = 0;
            while (cur_loc >= 0)
            {
                String s;
                int    eq_loc = p_Query.indexOf('=', cur_loc);

                if (eq_loc < 0)
                {
                    s = p_Query.substring(cur_loc);
                    cur_loc = eq_loc;
                }
                else
                {
                    s = p_Query.substring(cur_loc, eq_loc);
                    cur_loc = p_Query.indexOf('&', eq_loc);
                    if (cur_loc > 0) ++cur_loc;
                }
                s = s.toUpperCase();

                Enumeration ekeys = args.keys();
                Enumeration evals = args.elements();

                while (ekeys.hasMoreElements())
                {
                    String val[] = (String [])evals.nextElement();
                    String key   = (String)ekeys.nextElement();

                    if (s.equalsIgnoreCase(key))
                    {
                        /*
                        ** If not using flexible argument mode,
                        ** strip substructure from argument names
                        */
                        String sname = s;
                        if (p_Mode != GatewayManager.RUN_MODE_FLEXIBLE)
                        {
                            int slen = s.indexOf('.');
                            if (slen > 0) sname = s.substring(0, slen);
                        }
                        int i = p_ArgNames.indexOf(sname);

                        if (i < 0)
                        {
                            p_ArgValues.addElement(val);
                            p_ArgNames.addElement(sname);
                            break;
                        }
                        else
                        {
                            String sarr[];
                            sarr = (String [])p_ArgValues.elementAt(i);
                            int len1 = java.lang.reflect.Array.getLength(sarr);
                            int len2 = java.lang.reflect.Array.getLength(val);
                            String snew[] = new String[len1 + len2];
                            int j;
                            for (j = 0; j < len1; ++j) snew[j] = sarr[j];
                            for (j = 0; j < len2; ++j) snew[j + len1] = val[j];
                            p_ArgValues.setElementAt(snew, i);
                        }
                    }
                }
            }
        }
    }

    /*
    ** Read initialization parameters from server configuration.
    */
    private void loadParameters(ServletConfig cfg)
    {
        String s;
        int    i;

        /*
        ** Load the database parameters, which are essential;
        ** generate an error page if these aren't available.
        */
        s = cfg.getInitParameter("database");
        if (s != null) m_db_string = s;
        s = cfg.getInitParameter("dbusername");
        if (s != null) m_db_user = s;
        s = cfg.getInitParameter("dbpassword");
        if (s != null) m_db_pass = s;

        s = cfg.getInitParameter("document_path");
        if (s != null) m_doc_path = s;
        s = cfg.getInitParameter("document_proc");
        if (s != null) m_doc_proc = s;
        if (m_doc_proc == null) m_doc_path = "";

        s = cfg.getInitParameter("default_page");
        if (s != null) m_def_page = s;

        s = cfg.getInitParameter("pool_size");
        if (s != null)
        {
            i = Integer.parseInt(s);
            if (i >= 0) m_pool_size = 1;
        }

        s = cfg.getInitParameter("alternate_owa");
        if (s != null) m_alt_owa = s;

        s = cfg.getInitParameter("auth_mode");
        if (s != null)
        {
            if (s.equalsIgnoreCase("OWA_CUSTOM"))
                m_auth_mode = GatewayManager.RUN_SECURITY_CUSTOM;
            else if (s.equalsIgnoreCase("OWA_INIT"))
                m_auth_mode = GatewayManager.RUN_SECURITY_INIT;
            else if (s.equalsIgnoreCase("PACKAGE"))
                m_auth_mode = (GatewayManager.RUN_SECURITY_CUSTOM |
                               GatewayManager.RUN_SECURITY_PACKAGE);
        }

        s = cfg.getInitParameter("dad_charset");
        if (s != null)
        {
            /*
            ** ### NEED A COMPLETE LIST OF SUPPORTED MULTIBYTE CHARACTER SETS
            */
            if (s.equalsIgnoreCase("iso-8859-1"))
            {
                m_dad_cs = s;
                m_cs_mode = CHARSET_SINGLEBYTE;
            }
            else if (s.equalsIgnoreCase("utf-8"))
            {
                m_dad_cs = s;
                m_cs_mode = CHARSET_UNICODE;
            }
        }
    }

    /*
    ** Handle a GET request
    **
    ** Parse the query string, set the PL/SQL procedure name and argument
    ** mode, and execute showPage().
    */
    public void doGet(HttpServletRequest  p_Request,
                      HttpServletResponse p_Response)
         throws IOException, ServletException
    {
        String qry    = p_Request.getQueryString();
        int    mode;
	String procname;
	String docname = null;
        Vector names   = new Vector(VECTOR_INIT_SIZE, VECTOR_INCR_SIZE);
        Vector values  = new Vector(VECTOR_INIT_SIZE, VECTOR_INCR_SIZE);

        if ((m_db_user == null) || (m_db_pass == null))
        {
            showErrorPage(p_Response,
                          "Database information must be specified");
            return;
        }

        if (qry == null) qry = "";

        String[] sresult = new String[1];
        mode = setProcedureName(p_Request, sresult);
        if (mode == GatewayManager.RUN_MODE_LOBREAD)
        {
            docname = sresult[0];
            procname = m_doc_proc;
        }
        else
            procname = sresult[0];
        parseArguments(qry, mode, docname, names, values);
        // ### REPLACED HttpUtils.parseQueryString(qry);
        mode |= m_auth_mode;
        showPage(p_Request, p_Response, mode, procname, names, values);
    }

    /*
    ** Handle a POST request
    **
    ** If this is a content upload, call uploadFiles()
    ** Otherwise, get the arguments into a string and handle like a GET
    */
    public void doPost(HttpServletRequest  p_Request,
                       HttpServletResponse p_Response)
         throws IOException, ServletException
    {
        String ctype           = p_Request.getContentType();
        ServletInputStream sin = p_Request.getInputStream();
        int    mode;
	String procname;
	String docname = null;
        Vector names   = new Vector(VECTOR_INIT_SIZE, VECTOR_INCR_SIZE);
        Vector values  = new Vector(VECTOR_INIT_SIZE, VECTOR_INCR_SIZE);

        if ((m_db_user == null) || (m_db_pass == null))
        {
            showErrorPage(p_Response,
                          "Database information must be specified");
            return;
        }

        if (ctype.regionMatches(true, 0, "multipart/form-data;", 0, 20))
        {
            int bloc, bend;
            bloc = ctype.indexOf("boundary=");
            if (bloc >= 0)
            {
                bend = ctype.indexOf(' ', bloc);

                bloc += 9;

                if (bend < 0) ctype = "\r\n--" + ctype.substring(bloc);
                else          ctype = "\r\n--" + ctype.substring(bloc, bend);

                try
                {
                    uploadFiles(p_Request, p_Response, sin, ctype);
                }
                catch(SQLException e)
                {
                    showErrorPage(p_Response, "SQL failure: "+e.getMessage());
                }

                return;
            }
        }

        int    clen = p_Request.getContentLength();
        int    blen = 0;
        byte[] buf  = new byte[clen];

        while (clen > blen) blen += sin.read(buf, blen, clen - blen);

        String qry = new String(buf, m_dad_cs);

        String[] sresult = new String[1];
        mode = setProcedureName(p_Request, sresult);
        if (mode == GatewayManager.RUN_MODE_LOBREAD)
        {
            docname = sresult[0];
            procname = m_doc_proc;
        }
        else
            procname = sresult[0];
        parseArguments(qry, mode, docname, names, values);
        // ### REPLACED HttpUtils.parsePostData(clen,sin);
        mode |= m_auth_mode;
        showPage(p_Request, p_Response, mode, procname, names, values);
    }

    /*
    ** Handle a PUT request
    **
    ** For now, just returns an error.
    */
    public void doPut(HttpServletRequest  p_Request,
                      HttpServletResponse p_Response)
         throws IOException, ServletException
    {
        p_Response.sendError(p_Response.SC_BAD_REQUEST);
    }

    /*
    ** Servlet configuration information
    */
    public void init(ServletConfig config)
         throws ServletException
    {
        super.init(config);
        loadParameters(config); // Also available from getServletConfig()
        System.out.println("DBG> OWA Servlet Parameters");
        System.out.println("DBG>  Database = "+m_db_string);
        if (m_db_user != null)
            System.out.println("DBG>  Username = "+m_db_user);
        if (m_db_pass != null)
            System.out.println("DBG>  Password = "+m_db_pass);
        System.out.println("DBG>  Connection Pool = "+m_pool_size);
        System.out.println("DBG>  Authentication Mode = "+m_auth_mode);
        if (m_def_page != null)
            System.out.println("DBG>  Default Procedure = "+m_def_page);
        if (m_doc_proc != null)
            System.out.println("DBG>  Document Procedure = "+m_doc_proc);
        System.out.println("DBG>  Document Path = "+m_doc_path);
        if (m_alt_owa != null)
            System.out.println("DBG>  Alternate OWA = "+m_alt_owa);
    }

    /*
    ** Servlet configuration information
    */
    public String getServletInfo()
    {
        return("Owa PL/SQL Gateway Servlet");
    }
}
