/*
** Copyright (c) 1999-2004 Oracle Corporation, All rights reserved.
**
** File:         modowa.h
** Description:  Interfaces between modowa files.
** Authors:      Doug McMahon      dmcmahon@us.oracle.com
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
** History
**
** 08/17/2000   D. McMahon      Created
** 05/31/2001   D. McMahon      Began revisioning with 1.3.9/2.3.9
** 06/16/2001   D. McMahon      Added support for RAW input/output bindings
** 09/01/2001   D. McMahon      Added array and scalar rounding
** 09/07/2001   D. McMahon      Added TIMING diagnostic
** 02/11/2002   D. McMahon      Add OwaDocFile
** 03/27/2002   D. McMahon      Add morq_stream and rework morq_read
** 04/12/2002   D. McMahon      Fix definition of TIMING diagnostic
** 06/12/2002   D. McMahon      Add nls_check_utf8 and nls_conv_utf8
** 08/08/2002   D. McMahon      Add owa_dav_request and related functions
** 10/10/2002   D. McMahon      Add ERROR diagnostic flag
** 10/10/2002   D. McMahon      Add prototype for socket_test
** 11/14/2002   D. McMahon      Add sql_bind_vcs and SQLT_VCS/SQLT_VBI
** 03/17/2003   D. McMahon      Add NOMERGE to OwaAlternate
** 05/26/2003   D. McMahon      Add random number generator prototypes
** 06/06/2003   D. McMahon      Support special error page for SQL errors
** 01/19/2004   D. McMahon      Add sessioning support
*/

#ifndef MODOWA_H
#define MODOWA_H

#ifdef WIN32

#ifdef WITH_APACHE
# define WITHOUT_WINDOWS /* ### windows.h INCOMPATIBLE WITH APACHE HEADERS */
#endif

#ifndef WITHOUT_WINDOWS
/*
** NT version should rely exclusively on Win32 APIs, no calls to the
** C run-time library so that modowa has no dependency on installing it.
*/
#include <windows.h>
#endif /* !WITHOUT_WINDOWS */

#define NO_MARK_FOR_DESTRUCT

#else /* !WIN32 == UNIX */

/*
** Standard Unix headers
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>
#include <ctype.h>

#ifdef SUN_OS5
#define NO_MARK_FOR_DESTRUCT
#endif

#endif /* UNIX */

#ifndef OLD_RESET
#define RESET_AFTER_EXEC
#endif

#ifndef WITH_APACHE
#ifdef APACHE_LINKAGE

#define HTTP_CONTINUE                      100
#define HTTP_OK                            200
#define HTTP_CREATED                       201
#define HTTP_NO_CONTENT                    204
#define HTTP_MULTI_STATUS                  207
#define HTTP_MOVED_TEMPORARILY             302
#define HTTP_TEMPORARY_REDIRECT            307
#define HTTP_BAD_REQUEST                   400
#define HTTP_UNAUTHORIZED                  401
#define HTTP_NOT_FOUND                     404
#define HTTP_INTERNAL_SERVER_ERROR         500
#define HTTP_NOT_IMPLEMENTED               501

#define OK  0  /* Apache signal that module has handled request stage */

typedef struct request_rec request_rec;

#endif /* APACHE_LINKAGE */
#else
# ifndef APACHE_LINKAGE
#  define APACHE_LINKAGE
# endif
#endif /* !WITH_APACHE */

#ifdef ORACLE7
# ifndef OLD_ORACLE
#  define OLD_ORACLE
# endif
#endif

#ifdef WITH_OCI
/*
** Oracle headers
*/
#ifdef ORACLE7
#include <s.h>
#include <upidef.h>
#include <oerdef.h>
#else
#include <oratypes.h>
#include <oci.h>
#endif

#endif /* WITH_OCI */

/*
** Mappings to match Apache method codes
*/
#define DAV_METHOD_GET       0
#define DAV_METHOD_PUT       1
#define DAV_METHOD_POST      2
#define DAV_METHOD_DELETE    3
#define DAV_METHOD_CONNECT   4
#define DAV_METHOD_OPTIONS   5
#define DAV_METHOD_TRACE     6
#define DAV_METHOD_PATCH     7
#define DAV_METHOD_PROPFIND  8
#define DAV_METHOD_PROPPATCH 9
#define DAV_METHOD_MKCOL     10
#define DAV_METHOD_COPY      11
#define DAV_METHOD_MOVE      12
#define DAV_METHOD_LOCK      13
#define DAV_METHOD_UNLOCK    14
#define DAV_METHOD_INVALID   15

/*
** Diagnostic flags
*/
#define DIAG_COMMAND    0x0001
#define DIAG_ARGS       0x0002
#define DIAG_CGIENV     0x0004
#define DIAG_POOL       0x0008
#define DIAG_HEADER     0x0010
#define DIAG_COOKIES    0x0020
#define DIAG_CONTENT    0x0040
#define DIAG_RESPONSE   0x0080
#define DIAG_TIMING     0x0100
/*
** Mode flags
*/
#define DIAG_SQL        0x0200
#define DIAG_MEMORY     0x0400
#define DIAG_THREADS    0x0800
#define DIAG_ERROR      0x1000

/*
** LOB mode flags
*/
#define LOB_MODE_BIN    1           /* Binary LOBs only                  */
#define LOB_MODE_CHAR   2           /* Binary and Character LOBs         */
#define LOB_MODE_NCHAR  3           /* Binary, Character, and NCHAR LOBs */
#define LOB_MODE_BFILE  4           /* All LOB types including BFILE     */
/*
** Long mode flags
*/
#define LONG_MODE_RETURN_LEN  1     /* document_proc returns content len */
#define LONG_MODE_FETCH_LEN   2     /* Statement returns content length  */

/*
** Alternate OWA mode flags
*/
#define ALT_MODE_KEEP     1         /* Keep package state          */
#define ALT_MODE_RAW      2         /* Add RAW buffer for GET_PAGE */
#define ALT_MODE_LOBS     4         /* Use LOBs for GET_PAGE       */
#define ALT_MODE_CACHE    8         /* Enable GET_PAGE caching     */
#define ALT_MODE_SETSEC   16        /* Enable OWA_SEC workaround   */
#define ALT_MODE_GETRAW   32        /* Enable GET_PAGE_RAW         */
#define ALT_MODE_NOMERGE  64        /* Disable URL merge for POSTs */

/*
** Describe mode flags
*/
#define DESC_MODE_STRICT        0   /* Do not do describe           */
#define DESC_MODE_NORMAL        1   /* Do the collection describe   */
#define DESC_MODE_RELAXED       2   /* Do the advanced describe     */
#define DESC_MODE_POSITIONAL    3   /* Used for positional bindings */

/*
** Unicode mode flags
*/
#define UNI_MODE_NONE           0   /* Always use VARCHAR                  */
#define UNI_MODE_USER           1   /* Use NCHARs for user procecures only */
#define UNI_MODE_FULL           2   /* Use NCHARs for all OWA operations   */
#define UNI_MODE_RAW            4   /* Use RAWs for user procedures only   */

/*
** Reset mode flags
*/
#define RSET_MODE_NORMAL        0   /* Use RESET_PACKAGE                   */
#define RSET_MODE_FULL          1   /* Use FREE_ALL_RESOURCES              */
#define RSET_MODE_LAZY          2   /* Use REINITIALIZE                    */
#define RSET_MODE_INIT          3   /* Call HTP.INIT                       */

/*
** Authorization flags
*/
#define OWA_AUTH_NONE        0
#define OWA_AUTH_INIT        1
#define OWA_AUTH_CUSTOM      2
#define OWA_AUTH_PACKAGE  0x10

/*
** Apache request table types
*/
#define OWA_TABLE_SUBPROC    1  /* Table of subprocess environment vars    */
#define OWA_TABLE_HEADIN     2  /* Table of request header elements        */
#define OWA_TABLE_HEADOUT    3  /* Table of response header elements       */
#define OWA_TABLE_HEADERR    4  /* Table of response error header elements */

/*
** Cache performance flags
*/
#define CACHE_DEF_LIFE      0x7FFFFFFF  /* Infinite life                 */
#define CACHE_MAX_SIZE      0x100000    /* Largest file size             */
#define CACHE_MAX_ALIASES   8           /* Reallocation unit for aliases */

/*
** Buffer sizes
*/
#define HTBUF_LINE_CHARS     256    /* Size of GET_PAGE line value    */
#define HTBUF_LINE_LENGTH    768    /* Size of GET_PAGE line value X3 */
#define HTBUF_ARRAY_SIZE     256    /* Maximum array size             */
#define HTBUF_HEADER_MAX    4000    /* Maximum size of header element */
#define HTBUF_BLOCK_SIZE   32768    /* Size of a file upload block    */
#define HTBUF_BLOCK_READ   65536    /* Size of a file download block  */
#define HTBUF_PARAM_CHUNK     64    /* Increment for parameter allocs */
#ifdef OLD_ORACLE
#define HTBUF_ENV_MAX   HTBUF_HEADER_MAX/2
#else
#define HTBUF_ENV_MAX   HTBUF_HEADER_MAX
#endif
#define HTBUF_ENV_NAM   80          /* HTTP header element name size  */
#define HTBUF_ENV_ARR   40          /* Number of HTTP elements max    */

/*
** PL/SQL-related constants
*/
#define PLSQL_TAB     251                 /* Data type for PL/SQL table    */
#define PLSQL_ERR     6550                /* Oracle PL/SQL error code      */
#define PLSQL_ARGERR  "PLS-00306"         /* Wrong number or types of args */
#define LOGON_ERR     1017                /* Oracle logon failure          */

/*
** Connection pool locking states
*/
#define C_LOCK_UNUSED           0
#define C_LOCK_NEW              1
#define C_LOCK_AVAILABLE        2
#define C_LOCK_INUSE            3
#define C_LOCK_OFFLINE          4
#define C_LOCK_UNKNOWN          7
#define C_LOCK_MAXIMUM          8

/*
** Shared memory constants
*/
#define SHMEM_WAIT_MAX       100     /* milliseconds */
#ifdef WIN32
#define SHMEM_WAIT_INFINITE  INFINITE
#else
#define SHMEM_WAIT_INFINITE  0xFFFF  /* 1 minute */
#endif

typedef struct regnode regnode;

/*
** Global constants
*/
#define LONG_MAXSZ          0x7FFFFFFF  /* Maximum value for a signed long */
#define LONG_MAXSTRLEN      48          /* Maximum string width of oranum  */
#define ORA_MAXCOLNAME      32          /* Maximum length of column name   */
#define DESCRIBE_MAX        256         /* Maximum describe values         */

#ifdef WIN32

#define mem_zero(ptr, sz)    FillMemory((void *)(ptr),(DWORD)(sz),(BYTE)0)
#define mem_copy(dt, sc, sz) CopyMemory((void *)(dt),(void *)(sc),(SIZE_T)(sz))
#define os_str_length        lstrlen
#define os_str_copy          lstrcpy
#define os_str_print         wsprintf
#define os_alloca            _alloca
#define os_objhand           HANDLE
#define os_objptr            HANDLE
#define os_socket            SOCKET
#define os_dir_separator     '\\'
#define os_nullfilehand      INVALID_HANDLE_VALUE
#define os_badsocket         INVALID_SOCKET
#define os_nullmutex         INVALID_HANDLE_VALUE
#define InvalidFile(fp)      ((fp) == os_nullfilehand)
#define InvalidMutex(mx)     ((mx) == os_nullfilehand)

#else

#define mem_zero(ptr, sz)    memset((void *)(ptr),0,(sz))
#define mem_copy(dt, sc, sz) memcpy((void *)(dt),(void *)(sc),(sz))
#define os_str_length        strlen
#define os_str_copy          strcpy
#define os_str_print         sprintf
#define os_alloca            alloca
#define os_objhand           int
#define os_objptr            void *
#define os_socket            int
#define os_dir_separator     '/'
#define os_nullfilehand      -1
#define os_badsocket         -1
#define os_nullmutex         (void *)0
#define InvalidFile(fp)      ((fp) < 0)
#define InvalidMutex(mx)     (!(mx))

#endif

#define un_long unsigned long

#ifdef WITH_OCI
/*
** Type for passing SQL argument vectors
*/
typedef struct arg_record
{
    int    nargs;
    char **names;
    char **values;
    sb4   *widths;
    ub4   *counts;
    int    arrsz;
} arg_record;
#endif

/*
** Type for passing lists of name/value pairs
*/
typedef struct env_record
{
    int   count;
    char *names;
    char *values;
    int   nwidth;
    int   vwidth;
    char *authuser;
    char *authpass;
    char *session;
} env_record;

/*
** Handle to hold OCI connection information
*/
typedef struct connection connection;
#ifdef WITH_OCI
/* ### EVENTUALLY, MOVE ENTIRELY INTO owasql.c ### */
struct connection
{
#ifdef ORACLE7 /* UPI */
    ub4            host[256];
    struct hstdef *hstptr;
    struct oerdef *oerptr;
    word           stmhp1;
    word           stmhp2;
    word           stmhp3;
    word           stmhp4;
    word           stmhp5;
#else /* OCI 8.x */
    OCIEnv        *envhp;
    OCIError      *errhp;
    OCIServer     *srvhp;
    OCISvcCtx     *svchp;
    OCISession    *seshp;
    OCIStmt       *stmhp1;
    OCIStmt       *stmhp2;
    OCIStmt       *stmhp3;
    OCIStmt       *stmhp4;
    OCIStmt       *stmhp5;
    OCILobLocator *pblob;
    OCILobLocator *pclob;
    OCILobLocator *pnlob;
    OCILobLocator *pbfile;
#endif /* OCI 8.x */
    char          *session;
    char          *lastsql;
    char          *errbuf;
    int            mem_err;
    int            c_lock;
    int            timestamp;
    ub2            csid;
    int            ncflag;
    ub2            blob_ind;
    ub2            clob_ind;
    ub2            nlob_ind;
    ub2            bfile_ind;
    ub2            out_ind;
    ub2            rcode;
};

#ifdef OLD_ORACLE
# ifndef OCI_UCS2ID
#  define OCI_UCS2ID 1000
# endif
#endif

#ifdef ORACLE7
#define OraCursor word
#ifndef OCI_SUCCESS
# define OCI_SUCCESS 0
#endif
#ifndef OCI_ERROR
# define OCI_ERROR -1
#endif
#ifndef OCI_NEED_DATA
# define OCI_NEED_DATA 99
#endif
#ifndef OCI_ERROR_MAXMSG_SIZE
# define OCI_ERROR_MAXMSG_SIZE 512
#endif
#ifndef SQLT_CHR
# define SQLT_CHR 1
#endif
#ifndef SQLT_INT
# define SQLT_INT 3
#endif
#ifndef SQLT_STR
# define SQLT_STR 5
#endif
#ifndef SQLT_VCS
# define SQLT_VCS 5
#endif
#ifndef SQLT_VBI
# define SQLT_VBI 15
#endif
#ifndef SQLT_BIN
# define SQLT_BIN 23
#endif
#ifndef SQLT_LBI
# define SQLT_LBI 24
#endif
#ifndef SQLT_LNG
# define SQLT_LNG 8
#endif
#ifndef OCI_FIRST_PIECE
# define OCI_FIRST_PIECE UPI_FIRST_PIECE
#endif
#ifndef OCI_NEXT_PIECE
# define OCI_NEXT_PIECE  UPI_NEXT_PIECE
#endif
#ifndef OCI_LAST_PIECE
# define OCI_LAST_PIECE  UPI_LAST_PIECE
#endif
#ifndef OCI_ONE_PIECE
# define OCI_ONE_PIECE   UPI_ONE_PIECE
#endif
#else
#define OraCursor OCIStmt *
#endif

#endif

#ifdef ORACLE7
#define NEED_READ_DATA  3130
#define NEED_WRITE_DATA 3129
#else
#define NEED_READ_DATA  OCI_NEED_DATA
#define NEED_WRITE_DATA OCI_NEED_DATA
#endif

/*
** Sub-context for multi-threading support (visible only in modowa.c)
*/
typedef struct owa_mt_context owa_mt_context;

/*
** Structure to hold a cache alias
*/
typedef struct alias
{
    char    *logname;
    char    *physname;
    un_long  lifespan;
} alias;

/*
** Structure to hold an environment name/value pair
*/
typedef struct envstruct
{
    char    *name;
    char    *value;
} envstruct;

typedef struct shm_context
{
    int          mapsize;
    int          memthresh;
    int          filthresh;
    int          pagesize;
    os_objhand   map_fp;
    void        *map_ptr;
    os_objhand   f_mutex; /* HANDLE on WIN32, int on Unix */
} shm_context;

/*
** Main context area for an OWA Location
*/
typedef struct owa_context owa_context;
struct owa_context
{
    owa_mt_context *mtctx;          /* Back-pointer to module configuration */
    char           *location;       /* Location name                        */
    char           *oracle_userid;  /* Oracle username/password@dbname      */
    char           *doc_path;       /* Document path prefix                 */
    char           *doc_long;       /* Document LONG/LONG RAW path prefix   */
    char           *doc_file;       /* Document LONG/LONG RAW path prefix   */
    char           *doc_gen;        /* Generated document path prefix       */
    char           *doc_proc;       /* Document read access procedure       */
    char           *doc_start;      /* First-page procedure                 */
    char           *ora_before;
    char           *ora_after;
    char           *ora_proc;
    char           *diagfile;
    int             diagflag;
    int             descmode;
    char           *desc_schema;
    char           *alternate;
    int             altflags;
    int             rsetflag;
    int             ncflag;
    char           *nls_cs;
    char           *nls_lang;
    char           *nls_terr;
    char           *nls_datfmt;
    char           *nls_tz;
    char           *verstring;
    char           *froot;
    char           *authrealm;
    char           *session;
    char           *defaultcs;
    int             poolsize;
    int             upmax;
    int             version;
    int             lobtypes;
    int             lontypes;
    int             nls_init;
    int             ora_csid;
    int             dad_csid;
    int             db_csid;
    int             init_complete;
    char           *dav;
    int             keepalive_flag; /* ### Written at run-time on Unix */
    int             shm_offset;     /* ### Written at run-time on Unix */
    int             multithread;
    un_long         crtl_subnet;
    un_long         crtl_mask;
    int             realpid;
    connection     *c_pool;
    int             poolstats[C_LOCK_MAXIMUM];
    char           *reset_stmt;
    char            cgi_stmt[128];
    char            sec_stmt[256];
    char            get_stmt[256];
    char           *lob_stmt;
    char           *cpmv_stmt;
    char           *res_stmt;
    char           *prop_stmt;
    char           *sqlerr_uri;
    int             nlifes;
    alias          *lifes;
    int             nflex;
    char          **flexprocs;
    int             nreject;
    char          **rejectprocs;
    int             nenvs;
    envstruct      *envvars;
    un_long         scale_round;
    un_long         arr_round;
    shm_context    *mapmem;
    owa_context    *next;
};

/*
** Context used for directory reading
*/
typedef struct dir_record
{
    os_objptr dh;
    un_long   fsize;
    un_long   ftime;
    int       dir_flag;
    int       attrs;
    char      fname[HTBUF_HEADER_MAX];
} dir_record;


/*
** Structure used for random number generator states
*/
#define RAND_MAX_STATES  64
typedef struct randstate
{
    int     length;                             /* Length of the cycle */
    int     tap;                                /* Feedback tap */
    int     ind;                                /* Index to next value */
    int     carry;                              /* Carry flag */
    un_long state[RAND_MAX_STATES];             /* Previous states */
} randstate;

/*
** String Functions
*/
#define str_length      os_str_length
#define str_copy        os_str_copy

int   str_concat(char *dest, int dstart, const char *src, int maxlen);

int   str_compare(const char *s1, const char *s2, int maxlen, int caseflag);

char *str_char(const char *sptr, int ch, int reverse_flag);

char *str_substr(const char *sptr, const char *ssub, int caseflag);

char *str_dup(const char *sptr);

int   str_chrcmp(char *a, char *b, int caseflag);

void  str_prepend(char *base_str, char *prefix_str);

int   str_itoa(int i, char *s);

int   str_itox(un_long i, char *s);

int   str_atoi(char *s);

int   str_atox(char *s);

/*
** OS Functions
*/

void    os_exit(int status);

int     os_get_pid(void);

int     os_get_tid(void);

un_long os_get_time(un_long *musec);

void    os_milli_sleep(int ms);

un_long os_page_round(un_long sz);

int     os_get_errno(void);

void    os_set_errno(int e);

void    os_fd_close_exec(int fd);

char   *os_env_get(char *name);

char   *os_env_set(char *name, char *value);

/*
** Memory functions
*/

void *mem_alloc(int sz, int zero_flag);

void  mem_free(void *ptr);

/*
** IPC Functions
*/

os_objhand  os_sem_create(char *name, int sz, int secure_flag);

int         os_sem_acquire(os_objhand mh, int timeout);

int         os_sem_release(os_objhand mh);

int         os_sem_destroy(os_objhand mh);

os_objptr   os_mutex_create(char *name, int secure_flag);

int         os_mutex_acquire(os_objptr mh, int timeout);

int         os_mutex_release(os_objptr mh);

int         os_mutex_destroy(os_objptr mh);

os_objptr   os_cond_init(char *name, int sz, int secure_flag);

int         os_cond_wait(os_objptr mh, int timeout);

int         os_cond_signal(os_objptr mh);

int         os_cond_destroy(os_objptr mh);

/*
** File I/O Functions
*/

os_objhand  file_open_temp(char *fpath, char *tempname, int tmax);

os_objhand  file_open_write(char *fpath, int append_flag, int share_flag);

void        file_seek(os_objhand fp, long offset);

int         file_write_data(os_objhand fp, char *buffer, int buflen);

int         file_write_text(os_objhand fp, char *str, int slen);

int         file_read_data(os_objhand fp, char *buffer, int buflen);

void        file_close(os_objhand fp);

#ifndef NO_FILE_CACHE

os_objhand  file_open_read(char *fpath, un_long *fsz, un_long *fage);

os_objhand  file_map(os_objhand fp, un_long fsz,
                     char *mapname, int write_flag);

void       *file_view(os_objhand hnd, un_long fsz, int write_flag);

void        file_unmap(os_objhand hnd, void *ptr, un_long fsz);

void        file_delete(char *fname);

void        file_move(char *src, char *dest);

int         file_readdir(char *dirname, dir_record *dr);

int         file_mkdir(char *name, int mode);

int         file_rmdir(char *name);

os_objhand  os_shm_get(char *mapname, un_long fsz);

void       *os_shm_attach(os_objhand hnd, un_long fsz);

void        os_shm_detach(void *ptr);

void        os_shm_destroy(os_objhand hnd);

#endif /* FILE_CACHE */

/*
** Thread Functions
*/

os_objhand thread_spawn(void (*thread_start)(void *),
                        void *thread_context,
                        int *out_tid);

void       thread_exit(void);

void       thread_shutdown(os_objhand th);

void       thread_suspend(os_objhand th);

void       thread_resume(os_objhand th);

void       thread_cancel(os_objhand th);

void       thread_check(void);

void       thread_block_signals(void);

/*
** Socket Functions
*/
#ifndef NO_SOCKETS

void      socket_init(void);

void      socket_close(os_socket sock);

os_socket socket_listen(int port, char *ipaddr, int backlog);

os_socket socket_accept(os_socket sock);

os_socket socket_connect(int port, char *ipaddr);

int       socket_write(os_socket sock, char *buffer, int buflen);

int       socket_read(os_socket sock, char *buffer, int buflen);

int       socket_test(os_socket *sock, int nsock, int ms);

#endif

/*
** Debug log function
*/

void debug_out(char *diagfile, char *fmt,
               char *s1, char *s2, int i1, int i2);

/*
** Oracle NLS-related functions
*/

int   nls_cstype(int cs_id);

int   nls_csx_from_oracle(char *csname);

int   nls_csx_from_iana(char *csname);

int   nls_csid(int idx);

char *nls_iana(int idx);

int   nls_check_utf8(char *buf, int nbytes);

int   nls_conv_utf8(char *inbuf, char *outbuf);

int   nls_count_chars(int cs_id, char *outbuf, un_long *nbytes);

void  nls_validate_path(char *spath, int cs_id);

/*
** Utilities
*/
char    *util_get_method(int m);

void     util_print_time(un_long t, char *outbuf, un_long *ms);

void     util_set_mime(char *fpath, char *pmimetype, int bin_flag);

un_long  util_checksum(char *buffer, int buflen);

un_long  util_ipaddr(char *ipstr);

un_long  util_round(un_long n, un_long r);

int      util_decode64(char *inbuf, char *outbuf);

void     util_make_rand(randstate *prand, int length, int tap);

un_long  util_gen_rand(randstate *prand);

void     util_seed_rand(randstate *prand, randstate *srand);

un_long  util_get_rand(randstate *arand, randstate *brand);

void     util_init_rand(randstate *arand, randstate *brand,
                        char *key, int keylen);

/*
** SQL functions
*/

int   sql_conn_size(int arrsz);

int   sql_init(void);

#ifdef WITH_OCI

sword sql_parse(connection *c, OraCursor stmhp, char *stmt, int slen);

sword sql_fetch(connection *c, OraCursor stmhp);

sword sql_exec(connection *c, OraCursor stmhp, ub4 niters, int exact);

sword sql_transact(connection *c, int rollback_flag);

void  sql_get_nls(connection *c, owa_context *octx);

sb4   sql_get_error(connection *c);

int   sql_get_errpos(connection *c, OraCursor stmhp);

sword sql_connect(connection *c, owa_context *octx,
                  char *authuser, char *authpass);

sword sql_disconnect(connection *c);

sword sql_define(connection *c, OraCursor stmhp, ub4 pos,
                 dvoid *buf, sb4 buflen, ub2 dtype);

sword sql_bind_int(connection *c, OraCursor stmhp, ub4 pos, sb4 *val);

sword sql_bind_str(connection *c, OraCursor stmhp, ub4 pos,
                   char *str, sb4 slen);

sword sql_bind_raw(connection *c, OraCursor stmhp, ub4 pos,
                   char *rawbuf, ub2 *rlen, sb4 buflen);

sword sql_bind_arr(connection *c, OraCursor stmhp, ub4 pos,
                   char *aptr, ub2 *lens, sb4 awidth, sb2 *inds,
                   ub4 *asize, ub4 amax);

sword sql_bind_rarr(connection *c, OraCursor stmhp, ub4 pos,
                    char *aptr, ub2 *lens, sb4 awidth, sb2 *inds,
                    ub4 *asize, ub4 amax);

sword sql_bind_iarr(connection *c, OraCursor stmhp, ub4 pos,
                    sb4 *ints, ub4 *asize, ub4 amax);

sword sql_bind_vcs(connection *c, OraCursor stmhp, ub4 pos,
                   dvoid *buf, ub2 *lens, sb4 awidth, sb2 *inds,
                   ub4 *asize, ub4 amax, ub4 swidth);

sword sql_bind_stream(connection *c, OraCursor stmhp, ub4 pos, ub2 sql_type);

#ifndef ORACLE7

sword sql_bind_ptrs(connection *c, OraCursor stmhp, ub4 pos,
                    char **aptr, sb4 awidth, ub4 *asize, ub4 amax);

sword sql_bind_lob(connection *c, OraCursor stmhp, ub4 pos, ub2 flag);

#endif

sword sql_write_piece(connection *c, OraCursor stmhp,
                      dvoid *piecebuf, ub4 *nbytes, int pieceflag);

sword sql_read_piece(connection *c, OraCursor stmhp,
                     dvoid *piecebuf, ub4 *nbytes);

int   sql_describe(connection *c, char *pname, int descmode, char *schema,
                   int nargs, char *names[], ub4 counts[]);

#endif

/*
** Apache-based functions from module linkage code
*/

int   mowa_check_keepalive(int context_flag);

void  mowa_acquire_mutex(owa_context *octx);

void  mowa_release_mutex(owa_context *octx);

void  mowa_semaphore_create(owa_context *octx);

int   mowa_semaphore_get(owa_context *octx);

void  mowa_semaphore_put(owa_context *octx);

/*
** Apache-based request-related functions
*/
#ifdef APACHE_LINKAGE

void  morq_create_mutex(request_rec *request, owa_context *octx);

char *morq_get_buffer(request_rec *request, int sz);

void *morq_alloc(request_rec *request, int sz, int zero_flag);

int   morq_regex(request_rec *request,
                 char *pattern, char **str, int caseflag);

void  morq_send_header(request_rec *request);

void  morq_set_status(request_rec *request, int status);

void  morq_set_mimetype(request_rec *request, char *mtype);

void  morq_set_length(request_rec *request, long len, int range_flag);

int   morq_check_range(request_rec *request);

int   morq_get_range(request_rec *request, long *off, long *len);

char *morq_getword(request_rec *request,
                   const char **args, char ch, int un_escape);

int   morq_stream(request_rec *request, int flush_flag);

long  morq_read(request_rec *request, char *buffer, long buflen);

long  morq_write(request_rec *request, char *buffer, long buflen);

void  morq_print_int(request_rec *request, char *fmt, long ival);

void  morq_print_str(request_rec *request, char *fmt, char *sptr);

int   morq_table_get(request_rec *request,
                     int tableid, int index, char **name, char **value);

void  morq_table_put(request_rec *request, int tableid, int set_flag,
                     char *name, char *value);

char *morq_get_header(request_rec *request, char *name);

char *morq_get_env(request_rec *request, char *name);

char *morq_parse_auth(request_rec *request, char *buffer);

void  morq_add_context(request_rec *request, owa_context *octx);

void  morq_close_exec(request_rec *request, owa_context *octx, int first_flag);

void  morq_sql_error(request_rec *request,
                     int code, char *errbuf, char *lastsql);

#endif /* APACHE_LINKAGE */

/*
** Core gateway functions
*/
#ifdef APACHE_LINKAGE

int   owa_handle_request(owa_context *octx, request_rec *r,
                         char *req_args, int req_method);

int   owa_dav_request(owa_context *octx, request_rec *r,
                      char *req_args, char *req_uri, int req_method);

int   owa_writedata(connection *c, owa_context *octx, request_rec *r,
                    char *boundary, int clen, int long_flag,
                    char *outbuf, char *stmt, char *pmimetype, int nargs,
                    int nuargs, char **pnames, char **pvalues);

int   owa_readlob(connection *c, owa_context *octx, request_rec *r,
                  char *fpath, char *pmimetype, char *physical,
                  char *outbuf);

int   owa_readlong(connection *c, owa_context *octx, request_rec *r,
                   char *fpath, char *pmimetype, char *physical,
                   char *outbuf, int bin_flag);

int   owa_getpage(connection *c, owa_context *octx, request_rec *r,
                  char *physical, char *outbuf, int *rstatus);

void  owa_file_show(owa_context *octx, request_rec *r);

#ifdef WITH_OCI

int   owa_runplsql(connection *c, char *stmt, char *outbuf, int xargs,
                   int nargs, char *values[], ub4 counts[], sb4 widths[],
                   char **pointers, ub2 *plens, un_long arr_round, int csid);

int   owa_showerror(connection *c, char *stmt, int errcode);

#ifndef NO_FILE_CACHE

char *owa_map_cache(owa_context *octx, request_rec *r,
                    char *fpath, ub4 *life);

int   owa_download_file(owa_context *octx, request_rec *r,
                        char *fpath, char *pmimetype, ub4 life, char *outbuf);

int   owa_create_path(char *fpath, char *tempbuf);

#endif

#endif /* WITH_OCI */

#endif /* APACHE_LINKAGE */

void  owa_find_session(owa_context *octx, char *cookieval, char *session);

int   owa_reset(connection *c, owa_context *octx);

int   owa_passenv(connection *c, owa_context *octx, env_record *penv);

void  owa_set_statements(owa_context *octx);

void  owa_out_args(char *argarr[], int mode_flag);

void  owa_cleanup(owa_context *octx);

void  owa_pool_purge(owa_context *octx, int interval);

#ifndef NO_FILE_CACHE

void  owa_file_purge(owa_context *octx, int interval);

#endif

/*
** Shared-memory interfaces
*/

un_long owa_shmem_init(shm_context *map);

void    owa_shmem_update(shm_context *map, int *mapoff, int realpid,
                         char *location, int *poolstats);

int     owa_shmem_stats(shm_context *map, char *location, int *poolstats);

#endif /* MODOWA_H */
