/*
** Copyright (c) 1999-2004 Oracle Corporation, All rights reserved.
**
** File:         owadoc.c
** Description:  Gateway document-related routines.
** Author:       Doug McMahon      dmcmahon@us.oracle.com
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
** History
**
** 08/17/2000   D. McMahon      Created
** 05/31/2001   D. McMahon      Began revisioning with 1.3.9
** 06/04/2001   D. McMahon      Fix bug re. added headers in file uploads
** 06/16/2001   D. McMahon      Added support for RAW input/output bindings
** 09/01/2001   D. McMahon      Added array and scalar rounding
** 10/30/2001   D. McMahon      Fix bug with mime types and file caching
** 11/25/2001   D. McMahon      Return Content-Length:0 for empty docs
** 01/31/2002   D. McMahon      Always truncate LOB after write
** 02/11/2002   D. McMahon      Add OwaDocFile
** 03/27/2002   D. McMahon      Rework request read handling
** 09/18/2002   D. McMahon      Fix bug with width of reserved binding
** 11/14/2002   D. McMahon      Do not use RAW bind mode for OUT params
** 03/17/2003   D. McMahon      Merge URL arguments for POSTs
*/

#define WITH_OCI
#define APACHE_LINKAGE
#include <modowa.h>

/*
** Scan buffer for pattern or part of pattern
*/
static char *mem_find(char *mptr, int mlen, const char *mpat, int plen)
{
    int i, n;

    n = plen;

    if ((plen > 0) && (mptr) && (mpat))
        while (mlen > 0)
        {
            if (n > mlen) n = mlen;
            for (i = 0; i < n; ++i)
                if (mptr[i] != mpat[i])
                    break;
            if (i == n) return(mptr);
            ++mptr;
            --mlen;
        }
    return((char *)0);
}

/*
** Get file name for file-based transfer
*/
static int getfile(request_rec *r, connection *c, char *stmt, int first_flag,
                     sb4 nparams, char *pnams[], char *pvals[],
                     sb4 nam_width, sb4 val_width, char *pmimetype,
                     char **tmpbuf, int *tmpsize, char *outbuf, char *ret_file)
{
    sword    status;
    sb4      oerrno;
#ifdef ORACLE7
    char    *names;
    char    *values;
    int      sz;
#endif
    int      i;
    char     reserved[HTBUF_ARRAY_SIZE*2];
#ifdef ORACLE7
    ub2      nlens[HTBUF_ARRAY_SIZE];
    ub2      vlens[HTBUF_ARRAY_SIZE];
#endif
    ub2      rlens[HTBUF_ARRAY_SIZE];
    ub4      counts[3];

    mem_zero(reserved, sizeof(reserved));

    counts[0] = nparams; /* Names */
    counts[1] = nparams; /* Values */
    counts[2] = nparams; /* Reserved */

#ifdef ORACLE7
    sz = (nam_width + val_width + 2) * nparams;
    sz = (sz + HTBUF_BLOCK_SIZE - 1)/HTBUF_BLOCK_SIZE;

    /* Bind arrays too large, allocate */
    if (sz > *tmpsize)
    {
        *tmpsize = sz;
        sz *= HTBUF_BLOCK_SIZE;
        *tmpbuf = (char *)morq_alloc(r, sz, 0);
        if (!(*tmpbuf))
        {
            c->mem_err = sz;
            *tmpsize = 0;
            return(OCI_SUCCESS);
        }
    }
    names = *tmpbuf;
    values = names + (nam_width + 1) * nparams;

    for (i = 0; i < nparams; ++i)
    {
        str_copy(names + i * (nam_width + 1), pnams[i]);
        str_copy(values + i * (val_width + 1), pvals[i]);
    }
#endif

    c->lastsql = stmt;

    /*
    ** ### first_flag SHOULD BE USED FOR EFFICIENCY BUT stmhp3 USED LATER
    */
    status = sql_parse(c, c->stmhp3, stmt, -1);
    if (status != OCI_SUCCESS) goto filerror;

    *ret_file = '\0';

    status = sql_bind_int(c, c->stmhp3, (ub4)1, &nparams);
    if (status != OCI_SUCCESS) goto filerror;

#ifdef ORACLE7
    if (c->ncflag & UNI_MODE_RAW)
    {
      for (i = 0; i < (int)counts[2]; ++i) nlens[i] = str_length(names + i);
      for (i = 0; i < (int)counts[1]; ++i) vlens[i] = str_length(values + i);

      status = sql_bind_rarr(c, c->stmhp3, (ub4)2, names, nlens,
                             (sb4)(nam_width+1), (sb2 *)0, counts + 2,
                             counts[2]);
      if (status != OCI_SUCCESS) goto filerror;

      status = sql_bind_rarr(c, c->stmhp3, (ub4)3, values, vlens,
                             (sb4)(val_width+1), (sb2 *)0, counts + 1,
                             counts[1]);
      if (status != OCI_SUCCESS) goto filerror;
    }
    else
    {
      status = sql_bind_arr(c, c->stmhp3, (ub4)2, names, (ub2 *)0,
                            (sb4)(nam_width+1), (sb2 *)0, counts + 2,
                            counts[2]);
      if (status != OCI_SUCCESS) goto filerror;

      status = sql_bind_arr(c, c->stmhp3, (ub4)3, values, (ub2 *)0,
                            (sb4)(val_width+1), (sb2 *)0, counts + 1,
                            counts[1]);
      if (status != OCI_SUCCESS) goto filerror;
    }
#else
    status = sql_bind_ptrs(c, c->stmhp3, (ub4)2, pnams,
                           (sb4)(nam_width+1), counts + 2, counts[2]);
    if (status != OCI_SUCCESS) goto filerror;

    status = sql_bind_ptrs(c, c->stmhp3, (ub4)3, pvals,
                           (sb4)(val_width+1), counts + 1, counts[1]);
    if (status != OCI_SUCCESS) goto filerror;
#endif

    if (c->ncflag & UNI_MODE_RAW)
    {
      for (i = 0; i < HTBUF_ARRAY_SIZE; ++i) rlens[i] = 0;
      status = sql_bind_rarr(c, c->stmhp3, (ub4)4, reserved, rlens,
                             (sb4)2, (sb2 *)0, counts, *counts);
      if (status != OCI_SUCCESS) goto filerror;
    }
    else
    {
      status = sql_bind_arr(c, c->stmhp3, (ub4)4, reserved, (ub2 *)0,
                            (sb4)2, (sb2 *)0, counts, *counts);
      if (status != OCI_SUCCESS) goto filerror;
    }

    status = sql_bind_str(c, c->stmhp3, (ub4)5, pmimetype,
                          (sb4)HTBUF_LINE_LENGTH);
    if (status != OCI_SUCCESS) goto filerror;

    status = sql_bind_str(c, c->stmhp3, (ub4)6, ret_file,
                          (sb4)HTBUF_HEADER_MAX);
    if (status != OCI_SUCCESS) goto filerror;

    status = sql_exec(c, c->stmhp3, (ub4)1, 0);
    if (status != OCI_SUCCESS) goto filerror;

filerror:
    if (status != OCI_SUCCESS)
    {
        oerrno = sql_get_error(c); 
        if (oerrno) status = oerrno;
    }
    else
    {
        c->lastsql = (char *)0;
    }

    return(status);
}

#ifndef ORACLE7
/*
** Run PL/SQL to return a LOB handle for writing
*/
static int getlob(request_rec *r, connection *c, char *stmt,
                  int nargs, int first_flag,
                  sb4 nparams, char *pnams[], char *pvals[],
                  sb4 nam_width, sb4 val_width, char *pmimetype,
                  char **tmpbuf, int *tmpsize, char *outbuf)
{
    sword  status;
    sb4    oerrno;
#ifdef ORACLE7
    char  *names;
    char  *values;
    int    sz;
#endif
    int    i;
    char   reserved[HTBUF_ARRAY_SIZE*2];
#ifdef ORACLE7
    ub2    nlens[HTBUF_ARRAY_SIZE];
    ub2    vlens[HTBUF_ARRAY_SIZE];
#endif
    ub2    rlens[HTBUF_ARRAY_SIZE];
    ub4    counts[3];

    mem_zero(reserved, sizeof(reserved));

    counts[0] = nparams; /* Names */
    counts[1] = nparams; /* Values */
    counts[2] = nparams; /* Reserved */

#ifdef ORACLE7
    sz = (nam_width + val_width + 2) * nparams;
    sz = (sz + HTBUF_BLOCK_SIZE - 1)/HTBUF_BLOCK_SIZE;
    if (sz <= 1)
    {
        /* Use portion of outbuf */
        if (*tmpsize == 0) *tmpbuf = outbuf;
    }
    else
    {
        /* Bind arrays too large, allocate */
        if (sz > *tmpsize)
        {
            *tmpsize = sz;
            sz *= HTBUF_BLOCK_SIZE;
            *tmpbuf = (char *)morq_alloc(r, sz, 0);
            if (!(*tmpbuf))
            {
                c->mem_err = sz;
                return(OCI_SUCCESS);
            }
        }
    }
    names = *tmpbuf;
    values = names + (nam_width + 1) * nparams;

    for (i = 0; i < nparams; ++i)
    {
        str_copy(names + i * (nam_width + 1), pnams[i]);
        str_copy(values + i * (val_width + 1), pvals[i]);
    }
#endif

    c->lastsql = stmt;

    if (first_flag)
    {
        status = sql_parse(c, c->stmhp3, stmt, -1);
        if (status != OCI_SUCCESS) goto glberror;
    }


    status = sql_bind_int(c, c->stmhp3, (ub4)1, &nparams);
    if (status != OCI_SUCCESS) goto glberror;

#ifdef ORACLE7
    if (c->ncflag & UNI_MODE_RAW)
    {
      for (i = 0; i < (int)counts[2]; ++i) nlens[i] = str_length(names + i);
      for (i = 0; i < (int)counts[1]; ++i) vlens[i] = str_length(values + i);

      status = sql_bind_rarr(c, c->stmhp3, (ub4)2, names, nlens,
                             (sb4)(nam_width+1), (sb2 *)0, counts + 2,
                             counts[2]);
      if (status != OCI_SUCCESS) goto glberror;

      status = sql_bind_rarr(c, c->stmhp3, (ub4)3, values, vlens,
                             (sb4)(val_width+1), (sb2 *)0, counts + 1,
                             counts[1]);
      if (status != OCI_SUCCESS) goto glberror;
    }
    else
    {
      status = sql_bind_arr(c, c->stmhp3, (ub4)2, names, (ub2 *)0,
                            (sb4)(nam_width+1), (sb2 *)0, counts + 2,
                            counts[2]);
      if (status != OCI_SUCCESS) goto glberror;

      status = sql_bind_arr(c, c->stmhp3, (ub4)3, values, (ub2 *)0,
                            (sb4)(val_width+1), (sb2 *)0, counts + 1,
                            counts[1]);
      if (status != OCI_SUCCESS) goto glberror;
    }
#else
    status = sql_bind_ptrs(c, c->stmhp3, (ub4)2, pnams,
                           (sb4)(nam_width+1), counts + 2, counts[2]);
    if (status != OCI_SUCCESS) goto glberror;

    status = sql_bind_ptrs(c, c->stmhp3, (ub4)3, pvals,
                           (sb4)(val_width+1), counts + 1, counts[1]);
    if (status != OCI_SUCCESS) goto glberror;
#endif

    if (c->ncflag & UNI_MODE_RAW)
    {
      for (i = 0; i < HTBUF_ARRAY_SIZE; ++i) rlens[i] = 0;
      status = sql_bind_rarr(c, c->stmhp3, (ub4)4, reserved, rlens,
                             (sb4)2, (sb2 *)0, counts, *counts);
      if (status != OCI_SUCCESS) goto glberror;
    }
    else
    {
      status = sql_bind_arr(c, c->stmhp3, (ub4)4, reserved, (ub2 *)0,
                            (sb4)2, (sb2 *)0, counts, *counts);
      if (status != OCI_SUCCESS) goto glberror;
    }

    status = sql_bind_str(c, c->stmhp3, (ub4)5,
                          pmimetype, (sb4)HTBUF_LINE_LENGTH);
    if (status != OCI_SUCCESS) goto glberror;

    /* Set indicators for unbound arguments to null */
    c->bfile_ind = (ub2)-1;
    c->nlob_ind  = (ub2)-1;
    c->clob_ind  = (ub2)-1;
    c->blob_ind  = (ub2)-1;

    if (nargs >= 6)
    {
        status = sql_bind_lob(c, c->stmhp3, (ub4)6, SQLT_BLOB);
        if (status != OCI_SUCCESS) goto glberror;
    }

    if (nargs >= 7)
    {
        status = sql_bind_lob(c, c->stmhp3, (ub4)7, SQLT_CLOB);
        if (status != OCI_SUCCESS) goto glberror;
    }

    if (nargs >= 8)
    {
        status = sql_bind_lob(c, c->stmhp3, (ub4)8, 0);
        if (status != OCI_SUCCESS) goto glberror;
    }

    if (nargs >= 9) /* ### BFILE writes aren't yet supported */
    {
        status = sql_bind_lob(c, c->stmhp3, (ub4)9, SQLT_BFILE);
        if (status != OCI_SUCCESS) goto glberror;
    }

    status = sql_exec(c, c->stmhp3, (ub4)1, 0);

glberror:
    if (status != OCI_SUCCESS)
    {
        oerrno = sql_get_error(c); 
        if (oerrno) status = oerrno;
    }
    else
    {
        c->lastsql = (char *)0;
    }

    return(status);
}
#endif

/*
** Open SQL statement for LONG or LONG RAW UPDATE statement
*/
static int openlong(request_rec *r, connection *c,
                    char *stmt, char *sbind, int bin_flag, ub4 nbytes)
{
    sword    status;
    sb4      oerrno;
    int      stmtlen;
    ub2      sql_type;
    ub2      rind = (ub2)-1;

    stmtlen = str_length(stmt);
    if (stmtlen == 0) return(OCI_SUCCESS); /* Nothing to do */

    /*
    ** Build the UPDATE (or INSERT) statement and execute it.
    */
    c->lastsql = stmt;
    status = sql_parse(c, c->stmhp3, stmt, stmtlen);
    if (status != OCI_SUCCESS) goto oplerror;

    if (nbytes == 0)
    {
        /*
        ** For 0-length transfer, bind a null
        */
        if (bin_flag)
            status = sql_bind_raw(c,c->stmhp3,(ub4)1, (char *)"",&rind,(sb4)0);
        else
            status = sql_bind_str(c,c->stmhp3,(ub4)1, (char *)"",(sb4)0);
    }
    else
    {
        /*
        ** SQL_LBI is a LONG RAW.  Another possibility is SQLT_BIN.
        ** SQL_LNG is a LONG.  Other possibilities are SQLT_CHR and SQLT_STR.
        ** In case needed later:
        **   SQLT_LVB is 4-byte leading long raw
        **   SQLT_LVC is 4-byte leading long varchar
        */
        sql_type = (bin_flag) ? SQLT_LBI : SQLT_LNG;

        status = sql_bind_stream(c, c->stmhp3, (ub4)1, sql_type);
    }
    if (status != OCI_SUCCESS) goto oplerror;

    if (*sbind)
    {
        status = sql_bind_str(c, c->stmhp3, (ub4)2, sbind,
                              (sb4)(str_length(sbind) + 1));
        if (status != OCI_SUCCESS) goto oplerror;
    }

    status = sql_exec(c, c->stmhp3, (ub4)1, 0);
    /* Statement should now be ready for piecewise insert */
    if (status == NEED_WRITE_DATA) status = OCI_SUCCESS;

oplerror:
    if (status != OCI_SUCCESS)
    {
        oerrno = sql_get_error(c); 
        if (oerrno) status = oerrno;
    }
    else
    {
        c->lastsql = (char *)0;
    }

    return(status);
}

/*
** Get SQL INSERT or UPDATE statement for LONG or LONG RAW transfer
*/
static int getcursor(request_rec *r, connection *c, char *stmt, int first_flag,
                     sb4 nparams, char *pnams[], char *pvals[],
                     sb4 nam_width, sb4 val_width, char *pmimetype,
                     char **tmpbuf, int *tmpsize, char *outbuf,
                     char *ret_sql, char *ret_bind, int *bin_flag)
{
    sword    status;
    sb4      oerrno;
#ifdef ORACLE7
    char    *names;
    char    *values;
    int      sz;
#endif
    int      i;
    char     reserved[HTBUF_ARRAY_SIZE*2];
#ifdef ORACLE7
    ub2      nlens[HTBUF_ARRAY_SIZE];
    ub2      vlens[HTBUF_ARRAY_SIZE];
#endif
    ub2      rlens[HTBUF_ARRAY_SIZE];
    char     prawchar[10];
    ub4      counts[3];

    mem_zero(reserved, sizeof(reserved));

    counts[0] = nparams; /* Names */
    counts[1] = nparams; /* Values */
    counts[2] = nparams; /* Reserved */

#ifdef ORACLE7
    sz = (nam_width + val_width + 2) * nparams;
    sz = (sz + HTBUF_BLOCK_SIZE - 1)/HTBUF_BLOCK_SIZE;

    /* Bind arrays too large, allocate */
    if (sz > *tmpsize)
    {
        *tmpsize = sz;
        sz *= HTBUF_BLOCK_SIZE;
        *tmpbuf = (char *)morq_alloc(r, sz, 0);
        if (!(*tmpbuf))
        {
            c->mem_err = sz;
            *tmpsize = 0;
            return(OCI_SUCCESS);
        }
    }
    names = *tmpbuf;
    values = names + (nam_width + 1) * nparams;

    for (i = 0; i < nparams; ++i)
    {
        str_copy(names + i * (nam_width + 1), pnams[i]);
        str_copy(values + i * (val_width + 1), pvals[i]);
    }
#endif

    c->lastsql = stmt;

    /*
    ** ### first_flag SHOULD BE USED FOR EFFICIENCY BUT stmhp3 USED LATER
    */
    status = sql_parse(c, c->stmhp3, stmt, -1);
    if (status != OCI_SUCCESS) goto curerror;

    *ret_bind = '\0';
    *ret_sql = '\0';
    *prawchar = '\0';

    status = sql_bind_int(c, c->stmhp3, (ub4)1, &nparams);
    if (status != OCI_SUCCESS) goto curerror;

#ifdef ORACLE7
    if (c->ncflag & UNI_MODE_RAW)
    {
      for (i = 0; i < (int)counts[2]; ++i) nlens[i] = str_length(names + i);
      for (i = 0; i < (int)counts[1]; ++i) vlens[i] = str_length(values + i);

      status = sql_bind_rarr(c, c->stmhp3, (ub4)2, names, nlens,
                             (sb4)(nam_width+1), (sb2 *)0, counts + 2,
                             counts[2]);
      if (status != OCI_SUCCESS) goto curerror;

      status = sql_bind_rarr(c, c->stmhp3, (ub4)3, values, vlens,
                             (sb4)(val_width+1), (sb2 *)0, counts + 1,
                             counts[1]);
      if (status != OCI_SUCCESS) goto curerror;
    }
    else
    {
      status = sql_bind_arr(c, c->stmhp3, (ub4)2, names, (ub2 *)0,
                            (sb4)(nam_width+1), (sb2 *)0, counts + 2,
                            counts[2]);
      if (status != OCI_SUCCESS) goto curerror;

      status = sql_bind_arr(c, c->stmhp3, (ub4)3, values, (ub2 *)0,
                            (sb4)(val_width+1), (sb2 *)0, counts + 1,
                            counts[1]);
      if (status != OCI_SUCCESS) goto curerror;
    }
#else
    status = sql_bind_ptrs(c, c->stmhp3, (ub4)2, pnams,
                           (sb4)(nam_width+1), counts + 2, counts[2]);
    if (status != OCI_SUCCESS) goto curerror;

    status = sql_bind_ptrs(c, c->stmhp3, (ub4)3, pvals,
                           (sb4)(val_width+1), counts + 1, counts[1]);
    if (status != OCI_SUCCESS) goto curerror;
#endif

    if (c->ncflag & UNI_MODE_RAW)
    {
      for (i = 0; i < HTBUF_ARRAY_SIZE; ++i) rlens[i] = 0;
      status = sql_bind_rarr(c, c->stmhp3, (ub4)4, reserved, rlens,
                             (sb4)(val_width+1), (sb2 *)0, counts, *counts);
      if (status != OCI_SUCCESS) goto curerror;
    }
    else
    {
      status = sql_bind_arr(c, c->stmhp3, (ub4)4, reserved, (ub2 *)0,
                            (sb4)(val_width+1), (sb2 *)0, counts, *counts);
      if (status != OCI_SUCCESS) goto curerror;
    }

    status = sql_bind_str(c, c->stmhp3, (ub4)5, pmimetype,
                          (sb4)HTBUF_LINE_LENGTH);
    if (status != OCI_SUCCESS) goto curerror;

    status = sql_bind_str(c, c->stmhp3, (ub4)6, ret_sql,
                          (sb4)HTBUF_BLOCK_SIZE);
    if (status != OCI_SUCCESS) goto curerror;

    status = sql_bind_str(c, c->stmhp3, (ub4)7, ret_bind,
                          (sb4)HTBUF_HEADER_MAX);
    if (status != OCI_SUCCESS) goto curerror;

    status = sql_bind_str(c, c->stmhp3, (ub4)8, prawchar,
                          (sb4)sizeof(prawchar));
    if (status != OCI_SUCCESS) goto curerror;

    status = sql_exec(c, c->stmhp3, (ub4)1, 0);
    if (status != OCI_SUCCESS) goto curerror;

    *bin_flag = (*prawchar != '\0');

curerror:
    if (status != OCI_SUCCESS)
    {
        oerrno = sql_get_error(c); 
        if (oerrno) status = oerrno;
    }
    else
    {
        c->lastsql = (char *)0;
    }

    return(status);
}

/*
** Write a piece of a LONG or LONG RAW
*/
static int write_piece(connection *c, dvoid *lbuf, ub4 nbytes, int lob_flag)
{
    sword    status;
    status = sql_write_piece(c, c->stmhp3, lbuf, &nbytes, lob_flag);
    if (status == OCI_SUCCESS) status = sql_exec(c, c->stmhp3, (ub4)1, 0);
    return(status);
}

/*
** Read content data stream and write to LOBs or LONGs
**
** ### Add the compatibility mode that writes directly to document_table:
** ###    NAME         varchar2(256) unique not null,
** ###    MIME_TYPE    varchar2(128)
** ###    DOC_SIZE     number
** ###    DAD_CHARSET  varchar2(128)
** ###    LAST_UPDATED date
** ###    CONTENT_TYPE varchar2(128)
** ###    <cont_col1>,<cont_col2>,...
** ### Legacyd table has NAME, MIME_TYPE, DOC_SIZE, and:
** ###    CONTENT      long raw
** ### Call action procedure at the end with argument vectors in
** ### built in the normal mode.
** ### Need to generate unique name from SYSDATE+OSPID/filename.
** ### Need directives "upload_blob", "upload_clob", etc. that
** ### classify upload type by file extension.
** ### Need to investigate document download case.
*/
int owa_writedata(connection *c, owa_context *octx, request_rec *r,
                  char *boundary, int clen, int long_flag,
                  char *outbuf, char *stmt, char *pmimetype, int nargs,
                  int nuargs, char **pnames, char **pvalues)
{
    sword          status = OCI_SUCCESS;
    sb4            oerrno;
    int            i, j, n;
    int            end_data;
    int            lob_flag;
    int            first_flag;
    int            bound_flag;
    int            bin_flag = 0;
    int            blen;
    char          *aptr;
    char          *sptr;
    char          *filpath;
    char          *fldname;
    char          *ret_sql;
    char          *ret_bind;
    char          *ret_file;
    sb4            nparams;
    un_long        nbytes;
    ub4            total;
    ub4            amount;
    ub4            offset;
    ub2            cs_id;
    int            nfiles;
    ub4            file_sizes[HTBUF_ARRAY_SIZE];
    char          *param_name[HTBUF_ARRAY_SIZE];
    char          *param_value[HTBUF_ARRAY_SIZE];
    char          *tempname;
    char           cmax[32];
    sb4            nam_width;
    sb4            val_width;
    dvoid         *lbuf;
    char          *tmpbuf;
    int            tmpsize;
#ifndef NO_FILE_CACHE
    os_objhand  fp = os_nullfilehand;
#endif
#ifndef ORACLE7
    OCILobLocator *plob = (OCILobLocator *)0;
#endif

    switch (long_flag)
    {
    case 1:
        param_name[0] = (char *)"document_long";
        param_value[0] = octx->doc_long;
        break;
    case 2:
        param_name[0] = (char *)"document_file";
        param_value[0] = octx->doc_long;
        break;
    default:
        param_name[0] = (char *)"document_path";
        param_value[0] = octx->doc_path;
        break;
    }

    if (!param_value[0]) param_value[0] = (char *)"";
    nparams = 2;
    nam_width = str_length(*param_name);
    val_width = str_length(*param_value);
    if (val_width < 2) val_width = 2;

    /* Transfer the URL parameters */
    for (n = 0; n < nuargs; ++n)
    {
        param_name[nparams] = pnames[n];
        param_value[nparams] = pvalues[n];
        i = str_length(param_name[nparams]);
        if (i > nam_width) nam_width = i;
        i = str_length(param_value[nparams]);
        if (i > val_width) val_width = i;
        if (++nparams == HTBUF_ARRAY_SIZE) break;
    }

    /* Initialize the temporary buffer in case it's needed later */
    tmpbuf = (char *)0;
    tmpsize = 0;

    /*
    ** Assumes the caller only needs space for 2X the block size,
    ** and that there is another 2X available.
    */
    ret_sql  = outbuf + (HTBUF_BLOCK_SIZE << 1);
    ret_bind = ret_sql + HTBUF_BLOCK_SIZE;
    tempname = ret_bind;
    ret_file = ret_sql;

    blen = str_length(boundary);

    cs_id = nls_csid(octx->dad_csid);

    first_flag = 1;

    offset = (ub4)1;

    lob_flag = OCI_LAST_PIECE;
    end_data = 0;

    c->mem_err = 0;

    nfiles = 0;

    /* Limit content upload */
    if ((octx->upmax > 0) && (clen > octx->upmax))
    {
        str_itoa(octx->upmax, cmax);
        param_name[0] = (char *)"upload_max";
        param_value[0] = cmax;
        goto max_err;
    }
    /* Force streaming read of chunked data */
    if (clen < 0) clen = LONG_MAXSZ;

    morq_stream(r, 0);
    str_copy(outbuf, "\r\n");
    i = 2;
    while ((i > 0) || (!end_data))
    {
        if ((i < HTBUF_BLOCK_SIZE) && (!end_data))
        {
            j = HTBUF_BLOCK_SIZE;
            if (j > clen) j = clen;
            n = morq_read(r, outbuf + i, j);
            if (n < 0)
            {
                n = 0;
                clen = 0;
            }
            clen -= n;
            end_data = (clen == 0);
            n += i;
        }
        else
        {
            n = i;
        }

        sptr = mem_find(outbuf, n, boundary, blen);
        if (sptr)
        {
            i = sptr - outbuf;
            bound_flag = ((n - i) >= blen);
        }
        else
        {
            sptr = outbuf + n;
            bound_flag = 0;
        }
        i = sptr - outbuf;

        /* If boundary at top of buffer, ensure adequate data has been read */
        if ((n < HTBUF_BLOCK_SIZE) && (!end_data) && (i == 0))
        {
            i = n;
            continue;
        }

        if (lob_flag != OCI_LAST_PIECE)
        {
            amount = i;
            if ((i > 0) && (nls_cstype(cs_id)) && (!bin_flag))
            {
                nbytes = (un_long)i;
                amount = nls_count_chars((int)cs_id, outbuf, &nbytes);
                j = (int)nbytes;
                if (j < i)
                {
                    if (!end_data)
                        i = j; /* Leave fractional character for next pass */
                    else
                        amount += (i - j); /* Error - trailing fragment! */
                }
            }

            /* Consume all data and continue */
            if (lob_flag == OCI_ONE_PIECE)
            {
                if (long_flag == 1)
                {
                    status = openlong(r, c, ret_sql, ret_bind,
                                      bin_flag, (ub4)i);
                    if (status != OCI_SUCCESS) goto wrterror;
                }
#ifndef OLD_ORACLE
#ifdef NEVER /* ### DOESN'T WORK FOR UPDATE CASE! ### */
                else if (!long_flag)
                {
                    status = OCILobOpen(c->svchp, c->errhp, plob,
                                        (ub4)OCI_LOB_READWRITE);
                    if (status != OCI_SUCCESS) goto wrterror;
                }
#endif
#endif
                offset = (ub4)1;
                if ((i > 0) && (!end_data) && (!bound_flag))
                    lob_flag = OCI_FIRST_PIECE;
            }
            else /* OCI_FIRST_PIECE || OCI_NEXT_PIECE */
            {
                if ((i == 0) || (bound_flag))
                    lob_flag = OCI_LAST_PIECE;
                else
                    lob_flag = OCI_NEXT_PIECE;
            }

            if ((i == 0) && (!long_flag))
            {
                lbuf = (dvoid *)"";
                nbytes = 1;
            }
            else
            {
                lbuf = (dvoid *)outbuf;
                nbytes = i;
            }

            total = (lob_flag == OCI_ONE_PIECE) ? nbytes : 0;

#ifndef NO_FILE_CACHE
            if (long_flag == 2)
            {
                if ((nbytes > 0) && (!InvalidFile(fp)))
                    file_write_data(fp, (char *)lbuf, nbytes);
            }
            else
#endif
            if (long_flag)
            {
                if (nbytes > 0)
                    status = write_piece(c, lbuf, (ub4)nbytes, lob_flag);
            }
#ifndef ORACLE7
            else
            {
                status = OCILobWrite(c->svchp, c->errhp, plob,
                                     &total, offset, lbuf, (ub4)nbytes,
                                     lob_flag, (dvoid *)0,
                                     NULL, c->csid, (ub1)0);
            }
#endif
            if (status == NEED_WRITE_DATA) status = OCI_SUCCESS;
            if (status != OCI_SUCCESS) goto wrterror;

            offset += amount;

            if (lob_flag == OCI_ONE_PIECE) lob_flag = OCI_LAST_PIECE;

            if (lob_flag == OCI_LAST_PIECE)
            {
                /* Save the file size for later */
                file_sizes[nfiles++] = --offset;
#ifndef NO_FILE_CACHE
                if (long_flag == 2)
                {
                    file_close(fp);
                    file_move(tempname, ret_file);
                }
#endif
#ifndef ORACLE7
                if (!long_flag)
                {
                    /*
                    ** Trim the LOB (in case this was an update to a LOB that
                    ** was previously larger), then close it and commit it.
                    */
                    /*
                    ** ### THERE IS UNFORTUNATELY NO WAY TO TRIM A LONG
                    ** ### OR LONG RAW.  TO CLOSE THE TRANSACTION
                    ** ### YOU HAVE TO HAVE CALLED write_piece, WHICH
                    ** ### WON'T TAKE A NON-ZERO VALUE!  TO FIX THIS,
                    ** ### WE'LL HAVE TO DEFER THE BIND FOR THE LONG
                    ** ### STATEMENT UNTIL THE FIRST TRANSFER NEEDS TO
                    ** ### BE DONE.  AT THAT TIME, IF THE FIRST TRANSFER
                    ** ### IS ZERO BYTES, WE WOULD BIND IN THE IMMEDIATE
                    ** ### MODE WITH A NULL BUFFER.  UGH!
                    */
                    status = OCILobTrim(c->svchp, c->errhp, plob, offset);
                    if (status != OCI_SUCCESS) goto wrterror;
#ifndef OLD_ORACLE
#ifdef NEVER /* ### DOESN'T WORK FOR UPDATE CASE! ### */
                    status = OCILobClose(c->svchp, c->errhp, plob);
                    if (status != OCI_SUCCESS) goto wrterror;
#endif
#endif
                }
#endif
                status = sql_transact(c, 0); /* Commit */
                if (status != OCI_SUCCESS) goto wrterror;
            }
        }

        if (i < n) /* Boundary or boundary fragment has been encountered */
        {
            outbuf[n] = '\0';
            j = n - i;
            if ((j >= blen) && (i == 0))
            {
                /*
                ** Boundary has been found
                ** Boundary aligned at beginning of buffer
                */
                sptr += blen;
                if (!str_compare(sptr, "--", 2, 0))
                {
                    /* End of content stream, exit loop */
                    break;
                }
                if (str_compare(sptr,
                                "\r\nContent-Disposition: form-data;", 33, 1))
                    break; /* Some sort of streaming error ### */

                /* Look for trailing double-newline */
                sptr += 33;
                aptr = str_substr(sptr, "\r\n", 0);
                if (!aptr) break; /* Another class of streaming error ### */
                *(aptr++) = ';';
                *(aptr++) = '\0';
                while (*sptr == ' ') ++sptr;
                j = (aptr - sptr);
                fldname = (char *)morq_alloc(r, j, 0);
                if (!fldname)
                {
                    c->mem_err = j;
                    return(OCI_SUCCESS);
                }
                filpath = (char *)0;

                j = (aptr - outbuf);

                /*
                ** sptr now points to one of the following two forms:
                ** 1.  name="ordinary_field_name";
                ** 2.  name="filefield_name"; filename="filename.ext";
                */
                if (str_compare(sptr, "name=\"", 6, 1))
                    break; /* Stream parsing error ### */
                sptr += 6;
                aptr = fldname;
                while (*sptr != '\"') *(aptr++) = *(sptr++);
                i = (aptr - fldname);
                if (i > nam_width) nam_width = i;
                *(aptr++) = '\0';
                sptr += 2;
                while (*sptr == ' ') ++sptr;
                if (!str_compare(sptr, "filename=\"", 10, 1))
                {
                    sptr += 10;
                    filpath = aptr;
                    while (*sptr != '\"') *(aptr++) = *(sptr++);
                    *aptr = '\0';
                    i = aptr - filpath;
                    if (i > val_width) val_width = i;
                }
                /* Now look for optional Content-type */
                *pmimetype = '\0';
                sptr = outbuf + j;
                if (!str_compare(sptr, "Content-Type: ", 14, 1))
                {
                    aptr = str_substr(sptr, "\r\n", 0);
                    if (!aptr) break; /* Another streaming error ### */
                    *aptr = ' ';
                    j = (aptr - outbuf) + 2;
                    sptr += 14;
                    *aptr = '\0';
                    if (*sptr) str_copy(pmimetype, sptr);
                    sptr = outbuf + j;
                }

                /* Skip any other header fields until blank line is found */
                while (1)
                {
                    aptr = str_substr(sptr, "\r\n", 1);
                    if (!aptr) break;
                    j += 2;
                    if (aptr == sptr) break;
                    j += (aptr - sptr);
                    sptr = aptr + 2;
                }

                sptr = outbuf + j;

                if (!filpath)
                {
                    /*
                    ** Data until next boundary is content for this field
                    ** Contents limited to about 16K by this code
                    */
                    aptr = mem_find(sptr, n - j, boundary, blen);
                    if (!aptr) break; /* Error - field item too large ### */
                    i = aptr - sptr;
                    if (nparams < HTBUF_ARRAY_SIZE)
                    {
                        param_name[nparams] = fldname;
                        param_value[nparams] = (char *)morq_alloc(r, i + 1, 0);
                        if (!param_value[nparams])
                        {
                            c->mem_err = i + 1;
                            return(OCI_SUCCESS);
                        }
                        mem_copy(param_value[nparams], sptr, i);
                        param_value[nparams][i] = '\0';
                        if (i > val_width) val_width = i;
                        ++nparams;
                    }
                    sptr = aptr;
                    j = n - j - i;
                }
                else
                {
                    /* Commence a file upload */
                    param_name[1] = fldname;
                    param_value[1] = filpath;

                    if (long_flag == 2)
                    {
                        status = getfile(r, c, stmt, first_flag,
                                         nparams, param_name, param_value,
                                         nam_width, val_width, pmimetype,
                                         &tmpbuf, &tmpsize, outbuf, ret_file);
                        if (c->mem_err > 0) return(status);
                        lob_flag = OCI_ONE_PIECE;
                        first_flag = 0;
                        bin_flag = 1;
#ifndef NO_FILE_CACHE
                        /* Open/Create file for writing */
                        /* ### Need to check directory path
                           ### and create directories if necessary */
                        fp = os_nullfilehand;
                        if ((status == OCI_SUCCESS) && (*ret_file))
                        {
                            str_prepend(ret_file, octx->doc_file);
                            if (!owa_create_path(ret_file, tempname))
                                fp = file_open_temp(ret_file, tempname,
                                                    HTBUF_HEADER_MAX);
                        }
#endif
                    }
                    else if (long_flag)
                    {
                        status = getcursor(r, c, stmt, first_flag,
                                           nparams, param_name, param_value,
                                           nam_width, val_width, pmimetype,
                                           &tmpbuf, &tmpsize, outbuf,
                                           ret_sql, ret_bind, &bin_flag);
                        if (c->mem_err > 0) return(status);
                        lob_flag = OCI_ONE_PIECE;
                        first_flag = 0;
                    }
#ifndef ORACLE7
                    else
                    {
                        status = getlob(r, c, stmt, nargs, first_flag,
                                        nparams, param_name, param_value,
                                        nam_width, val_width, pmimetype,
                                        &tmpbuf, &tmpsize, outbuf + n);
                        if (status == OCI_SUCCESS)
                        {
                            if (c->mem_err > 0) return(status);

                            if (c->blob_ind != (ub2)-1)
                            {
                                plob = c->pblob;
                                lob_flag = OCI_ONE_PIECE;
                                bin_flag = 1;
                            }
                            else if (c->clob_ind != (ub2)-1)
                            {
                                plob = c->pclob;
                                lob_flag = OCI_ONE_PIECE;
                                bin_flag = 0;
                            }
                            else if (c->nlob_ind != (ub2)-1)
                            {
                                plob = c->pnlob;
                                lob_flag = OCI_ONE_PIECE;
                                bin_flag = 0;
                            }
                            first_flag = 0;
                        }
                    }
#endif
                    /* else LOB operation failed, skip this content */
                    j = n - j;
                }
            }

            /* Shift buffer and continue */
            for (i = 0; i < j; ++i) outbuf[i] = sptr[i];
        }
        else
        {
            i = 0;
        }
    }

max_err:
    /* Final call to LOB writer to generate return page */
    param_name[1] = (char *)"";
    param_value[1] = (char *)"";
    *pmimetype = '\0';

    /*
    ** Add arguments for file sizes
    */
    aptr = (char *)"content_length";
    n = str_length(aptr);
    if (n > nam_width) nam_width = n;
    n = 16;
    if (n > val_width) val_width = n;
    sptr = (char *)morq_alloc(r, n * nfiles, 0);
    if (sptr)
      for (i = 0; i < nfiles; ++i)
        if (nparams < HTBUF_ARRAY_SIZE)
        {
            param_name[nparams] = aptr;
            param_value[nparams] = sptr;
            str_itoa((int)file_sizes[i], sptr);
            sptr += n;
            ++nparams;
        }

    if (long_flag == 2)
    {
        status = getfile(r, c, stmt, first_flag,
                         nparams, param_name, param_value,
                         nam_width, val_width, pmimetype,
                         &tmpbuf, &tmpsize, outbuf, ret_file);
    }
    else if (long_flag)
        status = getcursor(r, c, stmt, first_flag,
                           nparams, param_name, param_value,
                           nam_width, val_width, pmimetype,
                           &tmpbuf, &tmpsize, outbuf,
                           ret_sql, ret_bind, &bin_flag);
#ifndef ORACLE7
    else
        status = getlob(r, c, stmt, nargs, first_flag,
                        nparams, param_name, param_value,
                        nam_width, val_width, pmimetype,
                        &tmpbuf, &tmpsize, outbuf);
#endif

wrterror:
    if (status != OCI_SUCCESS)
    {
        oerrno = sql_get_error(c); 
        if (oerrno) status = oerrno;
        sql_transact(c, 1); /* Rollback */
    }
    else
    {
        c->lastsql = (char *)0;
    }

    return(status);
}

/*
** Streaming range transfer, returning new range length
*/
static void transfer_ranges(request_rec *r, char *outbuf,
                            long offset, long nbytes, long *rlen, long *roff)
{
    long range_length = *rlen;
    long range_offset = *roff;
    long range_start;
    long range_nbytes;

    /*
    ** For range transfers, continue looping until either the
    ** range_length goes negative (indicating that there are no
    ** more ranges to transfer) or the range_offset is beyond
    ** the end of the current block.
    */
    while ((range_length >= 0) && (range_offset < (offset + nbytes)))
    {
        /*
        ** This should never happen, unless the ranges are out of
        ** order.  Since it's too late to transfer this range,
        ** mark it as done and get the next range.
        */
        if (range_offset < offset) range_length = 0;

        if (range_length > 0)
        {
            /*
            ** See if the bytes of this buffer overlap with the current
            ** range.  If so, compute how many bytes of the current block
            ** can be used to satisfy the range, transmit them, and
            ** subtract them.
            */
            range_start = range_offset - offset;
            range_nbytes = nbytes - range_start;
            if (range_nbytes > range_length) range_nbytes = range_length;
            morq_write(r, outbuf + range_start, (int)range_nbytes);
            range_length -= range_nbytes;
            range_offset += range_nbytes;
        }
        /*
        ** If range length is zero bytes, it's time to get the next
        ** range, which might also be inside this buffered region.
        */
        if (range_length == 0)
            if (!morq_get_range(r, &range_offset, &range_length))
                range_length = -1;
    }
    *rlen = range_length;
    *roff = range_offset;
}

/*
** Parse mimetype buffer for additional header elements
*/
static void parse_mime(owa_context *octx, request_rec *r, char *pmimetype)
{
    char *sptr, *eptr, *headval;

    sptr = str_char(pmimetype, '\n', 0);
    if (!sptr) return;
    *(sptr++) = '\0'; /* Null-terminate after mime type */
    while (*sptr)
    {
        headval = sptr;
        eptr = str_char(headval, '\n', 0);
        if (eptr) *eptr = '\0';
        while (*sptr != '\0')
        {
            if (*sptr == ':')
            {
                *(sptr++) = '\0';
                while (*sptr == ' ') ++sptr;
                break;
            }
            else
                ++sptr;
        }
        morq_table_put(r, OWA_TABLE_HEADOUT, 0, headval, sptr);
        sptr = (eptr) ? eptr + 1 : (char *)"";
    }
}

#ifndef ORACLE7
/*
** Read LOB data and download file
*/
int owa_readlob(connection *c, owa_context *octx, request_rec *r,
                char *fpath, char *pmimetype, char *physical, char *outbuf)
{
    sword          status;
    sb4            oerrno;
    ub4            nbytes;
    ub4            amount;
    ub4            total;
    ub4            offset;
    ub4            buflen = (ub4)HTBUF_BLOCK_READ;
    ub2            cs_id;
    int            bin_flag;
    int            last_flag;
    ub2            n;
    int            i;
    char          *sptr;
    ub2           *unistr;
    OCILobLocator *plob;
    int            range_flag = 0;
    long           range_offset = 0;
    long           range_length = -1;
#ifndef NO_FILE_CACHE
    char          *tempname = (char *)0;
    os_objhand     fp = os_nullfilehand;
#endif

    if (c->blob_ind != (ub2)-1)
    {
        plob = c->pblob;
        bin_flag = 1;
    }
    else if (c->clob_ind != (ub2)-1)
    {
        plob = c->pclob;
        bin_flag = 0;
    }
    else if (c->nlob_ind != (ub2)-1)
    {
        plob = c->pnlob;
        bin_flag = 0;
    }
    else if (c->bfile_ind != (ub2)-1)
    {
        plob = c->pbfile;
        bin_flag = 1;
    }
    else
    {
        /* Nothing to do - send header and exit */
        morq_send_header(r);
        return(OCI_SUCCESS);
    }

    /*
    ** Default character-set conversions for CLOBs and NCLOBs:
    */
    cs_id = (ub2)0;
    if ((octx->dad_csid) && (!bin_flag))
    {
        cs_id = nls_csid(octx->dad_csid);
#ifdef COUNT_CHARACTERS
        if (nls_cstype(cs_id) < 0)
        {
            /*
            ** Multibyte character sets not known to count_chars
            ** are fetched as UCS2 and stripped to ASCII.
            **
            ** ### THIS CODE IS NO LONGER USED BECAUSE WE USE LOB STREAMING
            ** ### TO AVOID THE NEED TO COUNT CHARACTERS DURING READS.
            */
            cs_id = (ub2)OCI_UCS2ID;
        }
#endif
    }

    /* Treat single-byte character sets as if they were binary */
    if (nls_cstype(nls_csid(octx->dad_csid)) == 0) bin_flag = 1;

    status = OCILobGetLength(c->svchp, c->errhp, plob, &total);
    if (status != OCI_SUCCESS) goto readerr;

#ifndef OLD_ORACLE
    status = OCILobOpen(c->svchp, c->errhp, plob, OCI_LOB_READONLY);
    if (status != OCI_SUCCESS) goto readerr;
#endif

    /*
    ** Set header Content-Type and status (unless called from owa_getpage)
    */
    if ((fpath) && (pmimetype))
    {
        /* Parse mimetype buffer for additional header elements */
        parse_mime(octx, r, pmimetype);

        sptr = pmimetype;
        if ((physical) && (*sptr))
        {
            sptr += (str_length(sptr) + 1);
            *sptr = '\0';
        }
        util_set_mime(fpath, sptr, bin_flag);
        if (sptr > pmimetype)
        {
          i = str_length(sptr);
          if (str_compare(sptr, pmimetype, i, 1) ||
              ((pmimetype[i] > ' ') && (pmimetype[i] != ';')))
            physical = (char *)0;
        }
        if ((!bin_flag) && (octx->dad_csid) &&
            (!str_substr(pmimetype, "charset=", 1)))
        {
            i = str_length(pmimetype);
            i = str_concat(pmimetype, i, "; charset=", HTBUF_LINE_LENGTH-1);
            str_concat(pmimetype, i, nls_iana(octx->dad_csid),
                       HTBUF_LINE_LENGTH - 1);
        }

        /*
        ** Set up the return header
        */
        morq_set_mimetype(r, pmimetype);
    }

    /*
    ** Content-Length returned only for binary LOBs; for character
    ** LOBs, the value of total indicates the number of characters,
    ** not bytes, so only return it if the DB character-set is single-byte.
    */
    if (bin_flag)
    {
        if (mowa_check_keepalive(octx->keepalive_flag))
        {
            morq_set_length(r, (long)total, 0);
            range_flag = morq_check_range(r);
            if (range_flag) range_length = 0;
        }
    }

    /*
    ** Send the header
    */
    morq_send_header(r);

    /*
    ** For range-based transfers from binary or single-byte data,
    ** transfer LOB data using seek-like semantics (for performance).
    ** This works only if you assume bytes == characters for the
    ** return from OCILobRead.  Otherwise, the more complex
    ** streaming version of the code must be used.
    */
    if ((bin_flag) && (range_flag))
    {
        while (morq_get_range(r, &range_offset, &range_length))
        {
            offset = (ub4)(range_offset + 1);
            for (total = (ub4)range_length; total > 0; total -= nbytes)
            {
                nbytes = buflen;
                nbytes = (buflen > total) ? total : buflen;
                status = OCILobRead(c->svchp, c->errhp, plob,
                                    &nbytes, offset, (dvoid *)outbuf, buflen,
                                    (dvoid *)0, NULL, cs_id, (ub1)0);
                if (status == NEED_READ_DATA) status = OCI_SUCCESS;
                if (status != OCI_SUCCESS) goto readerr;
                if (nbytes == 0) break; /* ### SOME SORT OF ERROR ### */
                morq_write(r, outbuf, (int)nbytes);
                offset += nbytes;
            }
        }
        goto closelob;
    }

#ifndef NO_FILE_CACHE
    if ((physical) && (total <= CACHE_MAX_SIZE))
    {
        tempname = outbuf + HTBUF_BLOCK_READ;
        owa_create_path(physical, tempname);
        fp = file_open_temp(physical, tempname, HTBUF_HEADER_MAX);
    }
#endif

    offset = 0;
    last_flag = 0;

    while (total > 0)
    {
        if ((bin_flag) && (buflen > total)) buflen = total;
        nbytes = 0; /* Read data in LOB streaming mode */
        status = OCILobRead(c->svchp, c->errhp, plob,
                            &nbytes, (ub4)1, (dvoid *)outbuf, buflen,
                            (dvoid *)0, NULL, cs_id, (ub1)0);
        if (status == NEED_READ_DATA) status = OCI_SUCCESS;
        else                         last_flag = 1;
        if (status != OCI_SUCCESS) goto readerr;
        if (nbytes == 0) break; /* ### SOME SORT OF ERROR ### */

        if (cs_id == OCI_UCS2ID)
        {
            amount = nbytes; /* UCS2 returns character count */
            nbytes <<= 1;
        }
#ifdef COUNT_CHARACTERS
        else if (nls_cstype(cs_id))
            amount = (ub4)nls_count_chars((int)cs_id, outbuf, &nbytes);
#endif
        else
            amount = nbytes;

        if (bin_flag) total -= amount;

        /*
        ** ### UCS2 fetch is used for all multi-byte client character
        ** ### sets because there's no other way to measure the return
        ** ### lengths in characters.  After the transfer, the UCS2
        ** ### string is stripped to plain ASCII.
        */
        if (cs_id == OCI_UCS2ID)
        {
            unistr = (ub2 *)(dvoid *)outbuf;
            sptr = outbuf;
            for (i = 0; i < (int)amount; ++i)
            {
                n = *(unistr++);
                *(sptr++) = (n < 0x80) ? (char)n : '?';
            }
            nbytes = amount;
        }

#ifndef NO_FILE_CACHE
        if (!InvalidFile(fp)) file_write_data(fp, outbuf, (int)nbytes);
#endif
        if (!range_flag) /* Do normal file download */
            morq_write(r, outbuf, (int)nbytes);
        else             /* Do range transfer */
            transfer_ranges(r, outbuf, (long)offset, (long)nbytes,
                            &range_length, &range_offset);
                            

        offset += nbytes;
        if (last_flag) total = 0; /* Ensure loop exit */
    }

closelob:
#ifndef OLD_ORACLE
    {
      int is_temp = 0; /* ### Should be of Oracle type "boolean" */
      status = OCILobIsTemporary(c->envhp, c->errhp, plob, &is_temp);
      if (status != OCI_SUCCESS) is_temp = 0;
      if (is_temp) status = OCILobFreeTemporary(c->svchp, c->errhp, plob);
      else         status = OCILobClose(c->svchp, c->errhp, plob);
    }
#endif

readerr:
    if (status != OCI_SUCCESS)
    {
        oerrno = sql_get_error(c); 
        if (oerrno) status = oerrno;
        physical = (char *)0; /* Ensure file_move does a delete */
    }
    else
    {
        c->lastsql = (char *)0;
    }

#ifndef NO_FILE_CACHE
    if (!InvalidFile(fp))
    {
        file_close(fp);
        file_move(tempname, physical);
    }
#endif
    return(status);
}
#endif

/*
** Read LONG or LONG RAW data and download file
*/
int owa_readlong(connection *c, owa_context *octx, request_rec *r,
                 char *fpath, char *pmimetype, char *physical, char *outbuf,
                 int bin_flag)
{
    sword       status;
    sb4         oerrno;
    ub4         nbytes;
    ub4         total = (ub4)0;
    int         slen;
    sb4         clen = (sb4)-1;
    ub2         sql_type;
    char       *sptr;
    char       *stmt;
    char       *sbind;
    char       *piecebuf;
    int         range_flag = 0;
    long        range_offset = 0;
    long        range_length = -1;
#ifndef NO_FILE_CACHE
    char       *tempname = (char *)0;
    os_objhand  fp = os_nullfilehand;
#endif

    if (bin_flag)
        sql_type = SQLT_LBI; /* SQLT_LVB is 4-byte leading long raw */
                             /* Another possibility is SQLT_BIN */
    else
        sql_type = SQLT_LNG; /* SQLT_LVC is 4-byte leading long varchar */
                             /* Other possibilities are SQLT_CHR, SQLT_STR */

    /* Parse mimetype buffer for additional header elements */
    parse_mime(octx, r, pmimetype);

    sptr = pmimetype;
    if ((physical) && (*sptr))
    {
        sptr += (str_length(sptr) + 1);
        *sptr = '\0';
    }
    util_set_mime(fpath, sptr, bin_flag);
    if (sptr > pmimetype)
      if (str_compare(sptr, pmimetype, -1, 1))
        physical = (char *)0;

    if (!bin_flag)
    {
         if ((octx->dad_csid) &&
             (!str_substr(pmimetype, "charset=", 1)))
         {
             slen = str_length(pmimetype);
             slen = str_concat(pmimetype, slen,
                               "; charset=", HTBUF_LINE_LENGTH - 1);
             str_concat(pmimetype, slen, nls_iana(octx->dad_csid),
                        HTBUF_LINE_LENGTH - 1);
         }
         if (!nls_cstype(octx->dad_csid)) bin_flag = 1;
    }

    stmt = outbuf;
    sbind = outbuf + HTBUF_BLOCK_SIZE;

    if (octx->lontypes == LONG_MODE_RETURN_LEN)
    {
        /* Get content length returned by document procedure */
        sptr = sbind + HTBUF_HEADER_MAX;
        while (*sptr == ' ') ++sptr;
        clen = (sb4)0;
        while ((*sptr >= '0') && (*sptr <= '9'))
            clen = clen * 10 + (*(sptr++) - '0');
    }
    c->lastsql = stmt;
    status = sql_parse(c, c->stmhp3, stmt, -1);
    if (status != OCI_SUCCESS) goto longerr;

    piecebuf = sbind;
    if (*sbind)
    {
        slen = str_length(sbind) + 1;
        status = sql_bind_str(c, c->stmhp3, (ub4)1, sbind, (sb4)slen);
        if (status != OCI_SUCCESS) goto longerr;
        piecebuf += slen;
    }

    status = sql_exec(c, c->stmhp3, (ub4)0, 0);
    if ((status != NEED_READ_DATA) && (status != OCI_SUCCESS)) goto longerr;

    if (octx->lontypes == LONG_MODE_FETCH_LEN)
    {
        status = sql_define(c, c->stmhp3, (ub4)2,
                            (dvoid *)&clen, (sb4)sizeof(clen),
                            (ub2)SQLT_INT);
        if (status != OCI_SUCCESS) goto longerr;
    }

    status = sql_define(c, c->stmhp3, (ub4)1,
                        (dvoid *)0, (sb4)LONG_MAXSZ, sql_type);
    if (status != OCI_SUCCESS) goto longerr;

    status = sql_fetch(c, c->stmhp3);

    /*
    ** Set up the return header
    */
    morq_set_mimetype(r, pmimetype);

    if ((bin_flag) && (clen >= 0))
        if (mowa_check_keepalive(octx->keepalive_flag))
        {
            morq_set_length(r, (long)clen, 0);
            range_flag = morq_check_range(r);
            if (range_flag) range_length = 0;
        }

    morq_send_header(r);

#ifndef NO_FILE_CACHE
    if ((physical) && (clen < CACHE_MAX_SIZE))
    {
        tempname = outbuf + HTBUF_BLOCK_READ;
        fp = file_open_temp(physical, tempname, HTBUF_HEADER_MAX);
    }
#endif

    while (status == NEED_READ_DATA)
    {
        nbytes = HTBUF_BLOCK_READ;

        status = sql_read_piece(c, c->stmhp3, (dvoid *)piecebuf, &nbytes);
        if (status != OCI_SUCCESS) goto longerr;

        status = sql_fetch(c, c->stmhp3);
        if ((status != NEED_READ_DATA) && (status != OCI_SUCCESS))
            goto longerr;

#ifndef NO_FILE_CACHE
        if (!InvalidFile(fp)) file_write_data(fp, piecebuf, (int)nbytes);
#endif
        if (!range_flag) /* Do normal file download */
            morq_write(r, piecebuf, (int)nbytes);
        else             /* Do range transfer */
            transfer_ranges(r, piecebuf, (long)total, (long)nbytes,
                            &range_length, &range_offset);

        total += nbytes;
    }

longerr:
    if (status != OCI_SUCCESS)
    {
        oerrno = sql_get_error(c); 
        if (oerrno) status = oerrno;
        physical = (char *)0; /* Ensure file_move does a delete */
    }
    else
    {
        c->lastsql = (char *)0;
    }

#ifndef NO_FILE_CACHE
    if (!InvalidFile(fp))
    {
        if (total > CACHE_MAX_SIZE) physical = (char *)0;

        file_close(fp);
        file_move(tempname, physical);
    }
#endif
    return(status);
}
