/*
** Copyright (c) 1999-2004 Oracle Corporation, All rights reserved.
**
** File:         owaplsql.c
** Description:  Basic gateway routines.
** Author:       Doug McMahon      dmcmahon@us.oracle.com
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
** History
**
** 08/17/2000   D. McMahon      Created
** 05/31/2001   D. McMahon      Began revisioning with 1.3.9
** 06/16/2001   D. McMahon      Added support for RAW input/output bindings
** 09/01/2001   D. McMahon      Added array and scalar rounding
** 11/25/2001   D. McMahon      Support returning Content-Length:0
** 02/11/2002   D. McMahon      Add OwaDocFile support (file name return)
** 06/06/2003   D. McMahon      Add owa_showerror
** 12/17/2003   D. McMahon      Fixed redirect handling (for Apache 2.0)
** 01/19/2004   D. McMahon      Added owa_find_session and related code
*/

#define WITH_OCI
#define APACHE_LINKAGE
#include <modowa.h>

/*
** Dummy names for output bind variables
**
** ### NEED TO FIND A WAY TO ALSO USE THIS MECHANISM FOR IN/OUT AND
** ### SOME MISC. OUT BINDS, SO THAT DIAGNOSTICS WILL BE EASIER TO READ.
*/
static char var_blob[]  = "OUT: <BLOB>";
static char var_clob[]  = "OUT: <CLOB>";
static char var_nlob[]  = "OUT: <NCLOB>";
static char var_bfile[] = "OUT: <BFILE>";
static char var_stmt[]  = "OUT: <SQL STMT>";
static char var_bind[]  = "OUT: <SQL BIND>";
static char var_clen[]  = "OUT: <CONTENT LENGTH>";
static char var_curi[]  = "OUT: <CACHE URI>";
static char var_csum[]  = "OUT: <CHECKSUM STRING>";
static char var_fname[] = "OUT: <FILE NAME>";

/*
** SQL Statements
*/
static char sql_stmt1[]  = "begin "
                           "OWA.INIT_CGI_ENV(:ecount, :namarr, :valarr); "
                           "end;";
static char sql_stmt2a[] = "begin\n"
                           "  OWA.USER_ID := :usr;\n"
                           "  OWA.PASSWORD := :pwd;\n"
                           "end;";
static char sql_stmt2b[] = "begin\n"
                           "  OWA.SET_USER_ID(:usr);\n"
                           "  OWA.SET_PASSWORD(:pwd);\n"
                           "end;";
static char sql_stmt2c[] = "begin\n"
                           "  OWA.SET_USER_ID(:usr);\n"
                           "  OWA.SET_PASSWORD(:pwd);\n"
                           "  OWA.ENABLE_RAW_MODE;\n"
                           "end;";
static char sql_stmt4a[] = "begin OWA.GET_PAGE(:linearr, :nlines); end;";
static char sql_stmt4b[] = "begin OWA.GET_PAGE_RAW(:linearr, :nlines); end;";
static char sql_stmt5a[] = "begin DBMS_SESSION.RESET_PACKAGE; end;";
static char sql_stmt5b[] = "begin DBMS_SESSION.MODIFY_PACKAGE_STATE(1); end;";
static char sql_stmt5c[] = "begin DBMS_SESSION.MODIFY_PACKAGE_STATE(2); end;";
static char sql_stmt5d[] = "begin HTP.INIT; end;";

/*
** Reset PL/SQL package state
*/
int owa_reset(connection *c, owa_context *octx)
{
    sword    status;
    sb4      oerrno;

    if (octx->altflags & ALT_MODE_KEEP) return(OCI_SUCCESS);
    if (octx->authrealm) return(OCI_SUCCESS);

    status = sql_exec(c, c->stmhp5, (ub4)1, 0);

    if (status != OCI_SUCCESS)
    {
        if (c->errbuf) c->lastsql = octx->reset_stmt;
        oerrno = sql_get_error(c); 
        if (oerrno) status = oerrno;
    }
    else if (c->errbuf)
    {
        c->errbuf[0] = '\0';
        c->lastsql = (char *)0;
    }

    return(status);
}

/*
** Pass environment information to OWA
*/
int owa_passenv(connection *c, owa_context *octx, env_record *penv)
{
    sword    status;
    sb4      oerrno;
    sb4      ecount;
    ub4      asize;
    ub4      amax;
    char    *usr;
    char    *pwd;

    status = OCI_SUCCESS;
    c->errbuf[0] = '\0';

#ifndef RESET_AFTER_EXEC
    /*
    ** ### CONSIDER DOING THIS AFTER THE CURRENT OPERATION, IT MIGHT
    ** ### IMPROVE PERCEIVED USER RESPONSE TIME.
    */
    if (c->c_lock == C_LOCK_INUSE)
    {
        /*
        ** No need to execute this on the very first call for this connection
        ** Saves overhead in the case where the connection is not being reused
        */
        status = owa_reset(c, octx);
        if (status != OCI_SUCCESS) return(status);
    }
#endif

    ecount = (sb4)penv->count;
    asize = (ub4)penv->count;
    amax = (ub4)asize;
    if (amax < (ub4)HTBUF_ENV_ARR) amax = (ub4)HTBUF_ENV_ARR;

    c->lastsql = octx->cgi_stmt;

    status = sql_bind_int(c, c->stmhp1, (ub4)1, &ecount);
    if (status != OCI_SUCCESS) goto passerr;
    status = sql_bind_arr(c, c->stmhp1, (ub4)2, penv->names,
                          (ub2 *)0, (sb4)(penv->nwidth), (sb2 *)0,
                          &asize, amax);
    if (status != OCI_SUCCESS) goto passerr;
    status = sql_bind_arr(c, c->stmhp1, (ub4)3, penv->values,
                          (ub2 *)0, (sb4)(penv->vwidth), (sb2 *)0,
                          &asize, amax);
    if (status != OCI_SUCCESS) goto passerr;

    status = sql_exec(c, c->stmhp1, (ub4)1, 0);
    if (status != OCI_SUCCESS) goto passerr;

    /*
    ** If Basic authentication information is present,
    ** pass it to OWA so that OWA_SEC will work.
    ** ### NOTE THAT hostname AND ip_address ARE NOT SET.
    ** ### THE hostname SHOULD BE AVAILABLE IN A CGI VARIABLE.
    ** ### SO IS THE ip_address, THOUGH IN THAT CASE THE OWA
    ** ### REQUIRES THAT IT BE BOUND AS A TABLE OF INTEGER VALUES,
    ** ### NOT A STRING.
    */
    usr = penv->authuser;
    pwd = penv->authpass;
    if (!usr) usr = (char *)"";
    if (!pwd) pwd = (char *)"";

    if ((*usr) || (*pwd) ||
        (octx->altflags & ALT_MODE_GETRAW) || (octx->altflags & ALT_MODE_KEEP))
    {
        c->lastsql = octx->sec_stmt;

        status = sql_bind_str(c, c->stmhp2, (ub4)1, usr, (sb4)-1);
        if (status != OCI_SUCCESS) goto passerr;
        status = sql_bind_str(c, c->stmhp2, (ub4)2, pwd, (sb4)-1);
        if (status != OCI_SUCCESS) goto passerr;

        if (!(*usr) && !(*pwd)) c->out_ind = -1; /* Null bindings */

        status = sql_exec(c, c->stmhp2, (ub4)1, 0);
    }

passerr:
    if (status != OCI_SUCCESS)
    {
        oerrno = sql_get_error(c); 
        if (oerrno) status = oerrno;
    }
    else
        c->lastsql = (char *)0;
    return(status);
}

/*
** Run PL/SQL procedure with arguments
*/
int owa_runplsql(connection *c, char *stmt, char *outbuf, int xargs,
                 int nargs, char *values[], ub4 counts[], sb4 widths[],
                 char **pointers, ub2 *plens, un_long arr_round, int csid)
{
    sword  status;
    sb4    oerrno;
    int    i, j;
    char  *optr;
    ub4    amax;
    int    ocsid = c->csid;

    status = OCI_SUCCESS;

    c->lastsql = stmt;
    c->errbuf[0] = '\0';

    status = sql_parse(c, c->stmhp3, stmt, -1);
    if (status != OCI_SUCCESS) goto runerr;

    /* Set indicators for unbound arguments to null */
    c->bfile_ind = (ub2)-1;
    c->nlob_ind  = (ub2)-1;
    c->clob_ind  = (ub2)-1;
    c->blob_ind  = (ub2)-1;

    optr = outbuf;

    if (csid) c->csid = csid;

    j = 0;
    for (i = 0; i < nargs; ++i)
    {
        if (i == (nargs - xargs))
            c->ncflag &= ~(UNI_MODE_USER | UNI_MODE_RAW);

        if (counts[i] == LONG_MAXSZ) continue; /* Skip binding */

        if (values[i] == (char *)0)
        {
            amax = (ub4)util_round((un_long)counts[i], arr_round);
            status = sql_bind_ptrs(c, c->stmhp3, (ub4)(j + 1), pointers,
                                   widths[i], counts + i, amax);
            if (c->ncflag & UNI_MODE_RAW) plens += counts[i];
        }
        else if (counts[i] > 0)
        {
          amax = (ub4)util_round((un_long)counts[i], arr_round);
          if (c->ncflag & UNI_MODE_RAW)
          {
            status = sql_bind_rarr(c, c->stmhp3, (ub4)(j + 1), values[i],
                                   plens, widths[i], (sb2 *)0, counts + i,
                                   amax);
            plens += counts[i];
          }
          else
            status = sql_bind_arr(c, c->stmhp3, (ub4)(j + 1), values[i],
                                  (ub2 *)0, widths[i], (sb2 *)0, counts + i,
                                  amax);
        }
#ifndef ORACLE7
        else if (values[i] == var_blob)
        {
            status = sql_bind_lob(c, c->stmhp3, (ub4)(j + 1), SQLT_BLOB);
            if (c->ncflag & UNI_MODE_RAW) ++plens;
        }
        else if (values[i] == var_clob)
        {
            c->csid = ocsid;
            status = sql_bind_lob(c, c->stmhp3, (ub4)(j + 1), SQLT_CLOB);
            if (c->ncflag & UNI_MODE_RAW) ++plens;
        }
        else if (values[i] == var_bfile)
        {
            status = sql_bind_lob(c, c->stmhp3, (ub4)(j + 1), SQLT_BFILE);
            if (c->ncflag & UNI_MODE_RAW) ++plens;
        } 
        else if (values[i] == var_nlob)
        {
            status = sql_bind_lob(c, c->stmhp3, (ub4)(j + 1), 0);
            if (c->ncflag & UNI_MODE_RAW) ++plens;
        }
#endif
        else if ((values[i] == var_stmt) || (values[i] == var_bind) ||
                 (values[i] == var_clen) || (values[i] == var_fname) ||
                 (values[i] == var_curi) || (values[i] == var_csum))
        {
            c->csid = ocsid;
            *optr = '\0';
            status = sql_bind_str(c, c->stmhp3, (ub4)(j + 1), optr, widths[i]);
            optr += widths[i];
            if (c->ncflag & UNI_MODE_RAW) ++plens;
        }
        else
        {
          if (c->ncflag & UNI_MODE_RAW)
          {
            status = sql_bind_raw(c, c->stmhp3, (ub4)(j + 1),
                                  values[i], plens, widths[i]);
            ++plens;
          }
          else
          {
            status = sql_bind_str(c, c->stmhp3, (ub4)(j + 1),
                                  values[i], widths[i]);
          }
        }
        if (status != OCI_SUCCESS) goto runerr;
        ++j;
    }

    status = sql_exec(c, c->stmhp3, (ub4)1, 0);

runerr:
    if (status != OCI_SUCCESS)
    {
        oerrno = sql_get_error(c); 
        if (oerrno) status = oerrno;
    }
    else
        c->lastsql = (char *)0;
    c->csid = ocsid;
    return(status);
}

/*
** Show special error page for SQL errors
*/
int owa_showerror(connection *c, char *stmt, int errcode)
{
    sword status;
    sb4   oerrno;
    sb4   val = errcode;
    status = OCI_SUCCESS;

    c->lastsql = stmt;
    c->errbuf[0] = '\0';

    status = sql_parse(c, c->stmhp3, stmt, -1);
    if (status != OCI_SUCCESS) goto errerr;

    status = sql_bind_int(c, c->stmhp3, (ub4)1, &val);
    if (status != OCI_SUCCESS) goto errerr;
    status = sql_bind_str(c, c->stmhp3, (ub4)2, (char *)"", (sb4)0);
    if (status != OCI_SUCCESS) goto errerr;

    status = sql_exec(c, c->stmhp3, (ub4)1, 0);

errerr:
    if (status != OCI_SUCCESS)
    {
        oerrno = sql_get_error(c); 
        if (oerrno) status = oerrno;
    }
    else
        c->lastsql = (char *)0;
    return(status);
}

/*
** Scan a cookie string for a match with the configured session cookie name.
** If found, parse off the value for that cookie and return its value
** in the designated buffer (up to 4000 bytes).
*/
void owa_find_session(owa_context *octx, char *cookieval, char *session)
{
    char *sptr;
    char *cptr;
    char *nptr;
    int   clen;

    *session = '\0';
    if (!(octx->session)) return;
    clen = str_length(octx->session);
    if (clen == 0) return;

    cptr = cookieval;

    while (1)
    {
        while (*cptr == ' ') ++cptr;
        sptr = cptr;
        while (*sptr != '=')
        {
            if (*sptr == '\0') return; /* ### Premature end of cookie string */
            ++sptr;
        }
        nptr = sptr;
        while (*nptr != ';')
        {
            if (*nptr == '\0') break;
            ++nptr;
        }

        if ((sptr - cptr) == clen)
            if (!str_compare(octx->session, cptr, clen, 1))
            {
                /*
                ** Cookie name matches session identifier;
                ** copy cookie contents into session string buffer
                */
                ++sptr;
                clen = nptr - sptr;
                if (clen >= HTBUF_HEADER_MAX) clen = HTBUF_HEADER_MAX - 1;
                mem_copy(session, sptr, clen);
                session[clen] = '\0';
                break;
            }

        if (*nptr == '\0') break;

        cptr = nptr + 1;
    }
}

/*
** Get and print results from OWA
*/
int owa_getpage(connection *c, owa_context *octx, request_rec *r,
                char *physical, char *outbuf, int *rstatus)
{
    sword    status;
    sb4      oerrno;
    int      diagflag;
    char    *optr;
    char    *sptr;
    char    *eptr;
    char    *rawbuf;
    ub2      rlen;
    int      i, j;
    int      bufwidth;
    int      arrlength;
    sb4      maxwidth;
    int      clen, olen;
    int      dbcs, apcs;
    int      content_flag;
    int      errhdr_code;
    char    *header_type;
    char    *header_len;
    char    *header_status;
    char    *headval;
    sb4      ecount;
    ub4      asize;
    ub4      amax;
    ub4      total = 0;
    char     pmore[LONG_MAXSTRLEN];
    sb2      inds[HTBUF_ARRAY_SIZE * 3];
    ub2      lens[HTBUF_ARRAY_SIZE * 3];
    char     ctype[HTBUF_LINE_LENGTH];
    char     session[HTBUF_HEADER_MAX];
#ifndef NO_FILE_CACHE
    char       *tempname = session;
    os_objhand  fp = os_nullfilehand;
#endif

    diagflag = octx->diagflag;

    status = OCI_SUCCESS;

    c->lastsql = octx->get_stmt;
    c->errbuf[0] = '\0';

    /*
    ** Get character set IDs for client and server
    */
    dbcs = octx->db_csid;
    if (dbcs > 0) dbcs = nls_csid(dbcs);
    apcs = octx->dad_csid;
    if (apcs > 0) apcs = nls_csid(apcs);

    /*
    ** If running single-byte mode, reduce the width of the buffer
    ** (in bytes) to the maximum required, and increase the array
    ** size correspondingly.
    */
    arrlength = HTBUF_ARRAY_SIZE;
    bufwidth = HTBUF_LINE_LENGTH;
    if ((octx->altflags & ALT_MODE_GETRAW) ||
        ((nls_cstype(dbcs) == 0) && (nls_cstype(apcs) == 0)))
    {
        arrlength = (bufwidth * arrlength)/HTBUF_LINE_CHARS;
        if (arrlength > sizeof(lens)/sizeof(*lens))
            arrlength = sizeof(lens)/sizeof(*lens);
        bufwidth = HTBUF_LINE_CHARS;
    }
    maxwidth = (sb4)(HTBUF_BLOCK_SIZE * 4); /* Allow for 4x expansion */

    rawbuf = outbuf + maxwidth;

    asize = (ub4)arrlength;
    amax = (ub4)util_round((un_long)asize, octx->arr_round);

    /* Set indicators for unbound arguments to null */
    c->bfile_ind = (ub2)-1;
    c->nlob_ind  = (ub2)-1;
    c->clob_ind  = (ub2)-1;
    c->blob_ind  = (ub2)-1;

    if (octx->altflags & ALT_MODE_GETRAW)
    {
        status = sql_bind_rarr(c, c->stmhp4, (ub4)1, outbuf, lens,
                               (sb4)bufwidth, inds, &asize, amax);
        if (status != OCI_SUCCESS) goto geterr;

        status = sql_bind_int(c, c->stmhp4, (ub4)2, &ecount);
        if (status != OCI_SUCCESS) goto geterr;
    }
    else
#ifndef ORACLE7
    if (octx->altflags & ALT_MODE_LOBS)
    {
        status = sql_bind_str(c, c->stmhp4, (ub4)1, outbuf, maxwidth);
        if (status != OCI_SUCCESS) goto geterr;

        status = sql_bind_lob(c, c->stmhp4, (ub4)2, SQLT_BLOB);
        if (status != OCI_SUCCESS) goto geterr;

        status = sql_bind_lob(c, c->stmhp4, (ub4)3, SQLT_CLOB);
        if (status != OCI_SUCCESS) goto geterr;
    }
    else
#endif
    if (octx->altflags & ALT_MODE_RAW)
    {
        status = sql_bind_str(c, c->stmhp4, (ub4)1, outbuf, maxwidth);
        if (status != OCI_SUCCESS) goto geterr;

        status = sql_bind_raw(c, c->stmhp4, (ub4)2,
                              rawbuf, &rlen, (sb4)HTBUF_BLOCK_SIZE);
        if (status != OCI_SUCCESS) goto geterr;

        status = sql_bind_str(c, c->stmhp4, (ub4)3, pmore, (sb4)sizeof(pmore));
        if (status != OCI_SUCCESS) goto geterr;

    }
    else
    {
        status = sql_bind_arr(c, c->stmhp4, (ub4)1, outbuf, lens,
                              (sb4)bufwidth, inds, &asize, amax);
        if (status != OCI_SUCCESS) goto geterr;

        status = sql_bind_int(c, c->stmhp4, (ub4)2, &ecount);
        if (status != OCI_SUCCESS) goto geterr;
    }

    clen = -1;

    *ctype = '\0';

    content_flag = 0;
    errhdr_code = 0;
    ecount = (sb4)arrlength;
    while (ecount == (sb4)arrlength)
    {
        *pmore = '\0';

        if (!(octx->altflags & ALT_MODE_GETRAW) &&
            (octx->altflags & (ALT_MODE_LOBS | ALT_MODE_RAW)))
        {
            outbuf[0] = '\n';
            outbuf[1] = '\0';
            rawbuf[0] = '\0';
        }
        else
        {
            rawbuf = (char *)"";
            optr = outbuf;
            for (i = 0; i < (int)asize; ++i)
            {
                *optr = '\0';
                optr += bufwidth;
                lens[i] = 0;
            }
            asize = 0;
        }

        rlen = (ub2)0;
        status = sql_exec(c, c->stmhp4, (ub4)1, 0);
        if (status != OCI_SUCCESS) goto geterr;

        /*
        ** Concatenate all the returned elements into a single string.
        ** This will make it easier to guarantee that pieces that
        ** belong together are kept together.  It will also allow
        ** a single write to the content stream to be used later.
        */
        if (!(octx->altflags & ALT_MODE_GETRAW) &&
            (octx->altflags & (ALT_MODE_LOBS | ALT_MODE_RAW)))
        {
            if ((rlen) && (content_flag))
            {
                optr = rawbuf;
                olen = (int)rlen;
            }
            else
            {
                optr = outbuf;
                olen = str_length(outbuf);
            }
        }
        else
        {
            sptr = outbuf;
            for (i = 0; i < (int)ecount; ++i)
            {
                optr = outbuf + (i * bufwidth);
                for (j = 0; j < lens[i]; ++j) *(sptr++) = *(optr++);
            }
            *sptr = '\0';
            olen = sptr - outbuf;
            optr = outbuf;
        }

        /* Search the result for content header */
        if (!content_flag)
        {
            /*
            ** No data returned from the OWA.  This could have happened
            ** as a result of a document read that failed to find a match.
            ** The best result in that case is "Document not found".
            ** In other cases, an error page might make more sense, but
            ** I can't distinguish those cases here.
            */
            if (ecount == 0)
            {
                *rstatus = HTTP_NOT_FOUND;
                return(status);
            }

            /*
            ** Unfortunately the OWA doesn't cleanly separate the header
            ** from the content, and doesn't separate the lines of the
            ** header from each other.  Instead, the returned result is
            ** a series of string fragments representing parts of a
            ** continuous stream.  This is fine when transferring the
            ** content but Apache will require us to separate out each
            ** header element and put it in a table.  Complicating matters,
            ** the OWA doesn't always return a header.  If it does, then
            ** the header will be separated from the content by a single
            ** blank line.  However, blank lines can also occur in the
            ** content, too, so it's important to look for clues in the
            ** preceeding lines that indicate it is in fact a header.
            */

            /* 1. Parse for newlines, which are taken to be separating
            **    header elements.  Look for a few key indications that
            **    these lines might be part of a header.  Stop at the
            **    first blank line.
            */
            sptr = outbuf;
            eptr = (char *)0;
            header_type = (char *)0;
            header_len = (char *)0;
            header_status = (char *)0;
            while (1)
            {
                headval = (char *)0;
                for (eptr = sptr; *eptr != '\0'; ++eptr)
                {
                    if (*eptr == '\n')     break;
                    else if (*eptr == ':') headval = eptr;
                }
                if (*eptr == '\0') break; /* No more newlines found */
                if (eptr == sptr)  break; /* Blank line encountered */
                if (!headval)      break; /* Can't be a header line */

                if (!str_compare(sptr, "Content-Type:", 13, 1))
                {
                    header_type = sptr;
                    sptr += 13;
                    while (*sptr == ' ') ++sptr;
                    j = (eptr - sptr);
                    if (j >= (int)sizeof(ctype)) j = sizeof(ctype) - 1;
                    j = str_concat(ctype, 0, sptr, j);
                    if ((octx->dad_csid) &&
                        (!str_substr(ctype, "charset=", 1)))
                    {
                        j = str_concat(ctype, j, "; charset=",
                                       sizeof(ctype) - 1);
                        str_concat(ctype, j, nls_iana(octx->dad_csid),
                                   sizeof(ctype) - 1);
                    }
                    content_flag = 1;
                }
                else if (!str_compare(sptr, "Content-Length:", 15, 1))
                {
                    header_len = sptr;
                    sptr += 15;
                    while (*sptr == ' ') ++sptr;
                    for (clen = 0; (*sptr >= '0') && (*sptr <= '9'); ++sptr)
                        clen = clen * 10 + (*sptr - '0');
                    content_flag = 1;
                }
                else if ((!str_compare(sptr, "Location:", 9, 1)) ||
                         (!str_compare(sptr, "Refresh:", 8, 1)))
                {
                    content_flag = 1;
                    errhdr_code = HTTP_MOVED_TEMPORARILY;
                }
                else if (!str_compare(sptr, "WWW-Authenticate:", 17, 1))
                {
                    content_flag = 1;
                    errhdr_code = HTTP_UNAUTHORIZED;
                }
                else if (!str_compare(sptr, "Set-Cookie:", 11, 1))
                {
                    /*
                    ** If sessioning is being used, check to see if
                    ** the PL/SQL code is returning a session cookie of
                    ** some sort.  If so, capture it and if necessary
                    ** update the connection to reflect it.
                    */
                    content_flag = 1;
                    if (octx->session)
                    {
                        for (sptr += 11; *sptr == ' '; ++sptr);
                        owa_find_session(octx, sptr, session);
                        if (*session)
                        {
                            if (c->session)
                            {
                                if (str_compare(session, c->session, -1, 0))
                                {
                                    mem_free((void *)(c->session));
                                    c->session = (char *)0;
                                }
                            }
                            if (!(c->session)) c->session = str_dup(session);
                        }
                        else
                        {
                            mem_free((void *)(c->session));
                            c->session = (char *)0;
                        }
                    }
                }
                else if (!str_compare(sptr, "Status:", 7, 1))
                {
                    header_status = sptr;
                    content_flag = 1;
                    for (sptr += 7; *sptr == ' '; ++sptr);
                    errhdr_code = str_atoi(sptr);
                    if (errhdr_code == HTTP_OK)
                        errhdr_code = 0;
                }
                sptr = eptr + 1;
            }

            if (*ctype == '\0')
            {
                /* Add a default return type */
                j = str_concat(ctype, 0, "text/html", -1);
                if (octx->dad_csid)
                {
                    j = str_concat(ctype, j, "; charset=", -1);
                    str_concat(ctype, j, nls_iana(octx->dad_csid),
                               sizeof(ctype) - 1);
                }
            }

            morq_set_mimetype(r, ctype);

            if (diagflag & DIAG_RESPONSE)
                debug_out(octx->diagfile, "Response\n  Content-Type: %s\n",
                          ctype, (char *)0, 0, 0);

            /*
            ** 2. If we found a blank line preceeded by one of the
            **    key indicators, then we've found a header.
            */
            if ((sptr == eptr) && (content_flag))
            {
                /* Transfer the header elements to the table */
                sptr = outbuf;
                optr = outbuf;
                while (sptr < eptr)
                {
                    for (optr = sptr; *optr != '\n'; ++optr);
                    *(optr++) = '\0';
                    if ((sptr != header_type) && (sptr != header_len) &&
                        (sptr != header_status))
                    {
                        for (headval = sptr; *sptr != '\0'; ++sptr)
                            if (*sptr == ':')
                            {
                                *(sptr++) = '\0';
                                while (*sptr == ' ') ++sptr;
                                break;
                            }
                        if (errhdr_code != 0)
                            morq_table_put(r,OWA_TABLE_HEADERR,0,headval,sptr);
                        else
                            morq_table_put(r,OWA_TABLE_HEADOUT,0,headval,sptr);
                        if (diagflag & DIAG_RESPONSE)
                            debug_out(octx->diagfile,
                                      "  %s: %s\n", headval, sptr, 0, 0);
                    }
                    sptr = optr;
                }
                ++optr;
                olen -= (optr - outbuf);
            }
            else
            {
                optr = outbuf;
            }

            /*
            ** ### THE LENGTH IS VIRTUALLY CERTAIN TO BE WRONG FOR
            ** ### MULTIBYTE CHARACTER SETS.  IN THE CASE WHERE THE
            ** ### CLIENT IS MULTIBYTE BUT THE SERVER IS NOT, THE ERROR
            ** ### WILL OCCUR BECAUSE THE BYTE LENGTHS WILL BE DIFFERENT.
            ** ### EVEN IF BOTH CLIENT AND SERVER ARE SET TO THE SAME
            ** ### MULTIBYTE CHARACTER SET, IF YOU'RE USING AN OLDER OWA
            ** ### YOU'RE LIKELY TO GET AN ERROR BECAUSE THE OWA PL/SQL
            ** ### CODE MEASURES USING length(), SO IT'S COUNTING
            ** ### CHARACTERS, NOT BYTES.  AS A TEMPORARY WORK-AROUND
            ** ### THIS CODE IGNORES CONTENT-LENGTHS WHEN THE CLIENT IS
            ** ### RUNNING A MULTI-BYTE SET (UNLESS THE CLIENT IS DOING
            ** ### A RAW TRANSFER).
            */
            if (!(octx->altflags & ALT_MODE_GETRAW) &&
                (clen >= 0) && (rlen == (ub2)0) && (c->blob_ind == (ub2)-1))
            {
                if ((dbcs != apcs) &&
                    ((nls_cstype(dbcs)) || (nls_cstype(apcs))))
                {
                    clen = -1;
                    if (diagflag & DIAG_RESPONSE)
                        debug_out(octx->diagfile,
                                  "  DAD CS = %d, DB CS = %d\n",
                                  (char *)0, (char *)0, apcs, dbcs);
                }
            }

            /*
            ** If transferring data from the raw buffer, reset
            ** the buffer pointer and length value.
            */
            if ((octx->altflags & ALT_MODE_RAW) && (rlen > (ub2)0))
            {
                olen = (int)rlen;
                optr = rawbuf;
            }

            /*
            ** If more flag indicates end of content,
            ** set ecount to ensure loop exit.
            */
            if ((octx->altflags & ALT_MODE_RAW) && (*pmore == '\0'))
                ecount = 0;

            /*
            ** If this is the only block of content, use the client-side
            ** length (to account for any character-set conversion);
            ** otherwise just hope clen is not set or was set properly
            ** on the server side.
            */
            if ((olen > 0) && (ecount < (sb4)arrlength))
                clen = olen;

            /*
            ** Set content length (if known) and return the header
            */
            if (clen >= 0)
            {
                if (mowa_check_keepalive(octx->keepalive_flag))
                    morq_set_length(r, (long)clen, 0);

                if (diagflag & DIAG_RESPONSE)
                    debug_out(octx->diagfile, "  Content-Length: %d\n",
                              (char *)0, (char *)0, clen, 0);
            }

#ifndef NO_FILE_CACHE
            if (physical)
            {
                /* Make sure file extension matches mime type */
                *tempname = '\0';
                util_set_mime(physical, tempname,
                              ((rlen > 0) || (c->blob_ind != (ub2)-1)));
                sptr = str_char(ctype, ';', 0);
                j = (sptr) ? (sptr - ctype) : str_length(ctype);
                if (str_compare(tempname, ctype, j, 1) || (tempname[j]))
                    physical = (char *)0;
            }
#endif

            if (errhdr_code != 0)
            {
                if (errhdr_code == HTTP_MOVED_TEMPORARILY)
                {
#ifdef NEVER
                    /* ### WHY IS THIS NECESSARY FOR POST REQUESTS? ### */
                    *rstatus = errhdr_code;
                    return(status);
#endif
                }
                else if (((olen == 0) || (!header_len)) && (!header_type))
                {
                    /* Allow Apache to return the error page */
                    *rstatus = errhdr_code;
                    return(status);
                }
                /* Otherwise handle error page in this routine */
                morq_set_status(r, errhdr_code);
            }

            /*
            ** 3. Now send the remainder of this content block.
            **    - For LOB mode, use LOB streaming
            **    - For RAW mode, transfer RAW content
            **    - Otherwise transfer remainder of character content
            */
#ifndef ORACLE7
            if (!(octx->altflags & ALT_MODE_GETRAW) &&
                (octx->altflags & ALT_MODE_LOBS))
            {
                olen = 0;   /* Avoid transferring any more from header */
                ecount = 0; /* Ensure exit from surrounding fetch loop */
                status = owa_readlob(c, octx, r, (char *)0, (char *)0,
                                     physical, outbuf);
                physical = (char *)0;
            }
            else
#endif
            /*
            ** LOB mode sent the header itself (to allow owa_readlob
            ** to compute the LOB length and add it to the header).
            ** For all other modes, transmit the header now.
            */
            morq_send_header(r);

#ifndef NO_FILE_CACHE
            if (physical)
            {
                fp = file_open_temp(physical, tempname, HTBUF_HEADER_MAX);
                if (InvalidFile(fp)) physical = (char *)0;
            }
#endif

            content_flag = 1;

            /*
            ** Fall through to transfer the remaining data, if any
            ** At this point optr should point to either the remainder
            ** of the character data, or to the raw buffer, and olen
            ** should be the length (in bytes).
            */
        }

        if (olen > 0)
        {
            /*
            ** Transfer block of raw or character data
            */
            if ((diagflag & DIAG_CONTENT) && (optr != rawbuf))
                debug_out(octx->diagfile, "%s\n", optr, (char *)0, 0, 0);
            morq_write(r, optr, olen);
#ifndef NO_FILE_CACHE
            if (!InvalidFile(fp)) file_write_data(fp, optr, olen);
#endif
            total += (ub4)olen;
            if (diagflag & DIAG_RESPONSE)
                debug_out(octx->diagfile, "  Wrote block of %d bytes\n",
                          (char *)0, (char *)0, olen, 0);
        }
    }

geterr:
    if (status != OCI_SUCCESS)
    {
        oerrno = sql_get_error(c); 
        if (oerrno) status = oerrno;
#ifndef NO_FILE_CACHE
        if (!InvalidFile(fp)) total = (~(ub4)0); /* Ensure delete */
#endif
    }
    else
    {
        c->lastsql = (char *)0;
    }

#ifndef NO_FILE_CACHE
    if (!InvalidFile(fp))
    {
        if (total > CACHE_MAX_SIZE) physical = (char *)0;

        file_close(fp);
        file_move(tempname, physical);
    }
#endif

    return(status);
}

void owa_set_statements(owa_context *octx)
{
    char *sptr;
    char *aptr;
    char *nptr;

    switch (octx->rsetflag)
    {
    case RSET_MODE_FULL:  sptr = sql_stmt5b; break;
    case RSET_MODE_LAZY:  sptr = sql_stmt5c; break;
    case RSET_MODE_INIT:  sptr = sql_stmt5d; break;
    default:              sptr = sql_stmt5a; break;
    }
    octx->reset_stmt = sptr;

    if (octx->alternate)
    {
        sptr = octx->cgi_stmt;
        aptr = sql_stmt1;

        do
        {
            nptr = str_substr(aptr, "OWA.", 0);
            if (nptr)
            {
                sptr += str_concat(sptr, 0, aptr, (nptr - aptr));
                sptr += str_concat(sptr, 0, octx->alternate, 32);
                aptr = nptr + 3;
            }
            else
                str_copy(sptr, aptr);
        }
        while (nptr);

        sptr = octx->sec_stmt;
        if (octx->altflags & ALT_MODE_GETRAW)       aptr = sql_stmt2c;
        else if (octx->altflags & ALT_MODE_SETSEC)  aptr = sql_stmt2b;
        else                                        aptr = sql_stmt2a;

        do
        {
            nptr = str_substr(aptr, "OWA.", 0);
            if (nptr)
            {
                sptr += str_concat(sptr, 0, aptr, (nptr - aptr));
                sptr += str_concat(sptr, 0, octx->alternate, 32);
                aptr = nptr + 3;
            }
            else
                str_copy(sptr, aptr);
        }
        while (nptr);

        sptr = octx->get_stmt;
        if (octx->altflags & ALT_MODE_GETRAW)
        {
          sptr += str_concat(sptr, 0, "begin ", -1);
          sptr += str_concat(sptr, 0, octx->alternate, 32);
          sptr += str_concat(sptr, 0,
                             ".GET_PAGE_RAW(:linearr, :nlines); end;\n", -1);
        }
        else
        {
          sptr += str_concat(sptr, 0, "begin ", -1);
          sptr += str_concat(sptr, 0, octx->alternate, 32);
#ifndef ORACLE7
          if (octx->altflags & ALT_MODE_LOBS)
            sptr += str_concat(sptr, 0,
                               ".GET_PAGE(:header, :bptr, :cptr); end;", -1);
          else
#endif
          if (octx->altflags & ALT_MODE_RAW)
              sptr += str_concat(sptr, 0,
                         ".GET_PAGE(:cbuffer, :rbuffer, :moreflag); end;", -1);
          else
              sptr += str_concat(sptr, 0,
                                 ".GET_PAGE(:linearr, :nlines); end;", -1);
        }
    }
    else
    {
        str_copy(octx->cgi_stmt, sql_stmt1);
        if (octx->altflags & ALT_MODE_GETRAW)
        {
            str_copy(octx->get_stmt, sql_stmt4b);
            str_copy(octx->sec_stmt, sql_stmt2c);
        }
        else
        {
            str_copy(octx->get_stmt, sql_stmt4a);
            str_copy(octx->sec_stmt,
                     (octx->altflags & ALT_MODE_SETSEC) ? sql_stmt2b:
                                                          sql_stmt2a);
        }
    }
}

/*
** Set up output arguments for special operations
**   0 - Returning <32K statement>, <4K bind>, <integer content-length>
**   1 - Returning <blob>, <clob>, <nclob>, <bfile>
**   2 - Returning <4K URI>, <32K checksum string>
**   3 = Returning <4K file name>
*/
void owa_out_args(char *argarr[], int mode_flag)
{
    switch (mode_flag)
    {
    case 0:
        argarr[0] = var_stmt;
        argarr[1] = var_bind;
        argarr[2] = var_clen;
        break;
    case 1:
        argarr[0] = var_blob;
        argarr[1] = var_clob;
        argarr[2] = var_nlob;
        argarr[3] = var_bfile;
        break;
    case 2:
        argarr[0] = var_curi;
        argarr[1] = var_csum;
        break;
    case 3:
        argarr[0] = var_fname;
        break;
    default:
        break;
    }
}
