/*
** Copyright (c) 1999-2004 Oracle Corporation, All rights reserved.
**
** File:         owasql.c
** Description:  OCI structures and functions
** Author:       Doug McMahon      dmcmahon@us.oracle.com
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
** History
**
** 08/17/2000   D. McMahon      Created
** 05/31/2001   D. McMahon      Began revisioning with 1.3.9
** 06/16/2001   D. McMahon      Added support for RAW input/output bindings
** 09/01/2001   D. McMahon      Added array and scalar rounding
** 02/18/2002   D. McMahon      Ensure oerrno value is zero-initialized
** 11/14/2002   D. McMahon      Add sql_bind_vcs
** 05/22/2003   D. McMahon      Scan for last @ in connect string
** 12/17/2003   D. McMahon      Normal describe skips signature mismatches
** 01/19/2004   D. McMahon      Free session string on disconnect
** 01/29/2004   D. McMahon      Patch sql_bind_ptrs to avoid OCI indicator bug
*/

#define WITH_OCI
#include <modowa.h>

#ifdef OLD_ORACLE
#ifndef WIN32
/*
** ### On Unix platforms, the 8.0.x versions of libclntsh.so had an
** ### undefined reference to slpmprodstab[], which is a global
** ### used by some Oracle core routines.  To get mod_owa to link,
** ### it is necessary to define this global somewhere, so this
** ### code defines it here.
*/

#ifndef ORACLE7
#ifdef USE_SLPM

#include <slpm.h>

#else
/*
** ### But there's a catch: Oracle doesn't ship customers the
** ### internal slpm.h header file needed to define the structure
** ### of the global as well as its initialization!  To get around
** ### this problem, mod_owa has it's own definition, used if
** ### you don't define USE_SLPM (which I use to compile it
** ### internally).
*/

typedef struct slpmpt
{
    text  *name_slpmpt;         /* Name of the program object   */
    uword  nlen_slpmpt;         /* Length of name_slpmpt        */
    text  *file_slpmpt;         /* File containing executable   */
    uword  flen_slpmpt;         /* Length of file_slpmpt        */
    text  *sym_slpmpt;          /* Symbol name                  */
    uword  slen_slpmpt;         /* Length of sym_slpmpt         */
    void  *fp_slpmpt;           /* Entry point if linked in     */
    ub2    flags_slpmpt;        /* Flags desc location of prog  */
} slpmpt;

#define SLPMLINKED     (ub2)1           /* Program to run is linked in  */
#define SLPMBIND       (ub2)8           /* Program dynamically bound    */
#define SLPMEND        (ub2)32768       /* Flag to mark end of table    */

#endif /* !USE_SLPM */

const text empty_name[1] = {0};
const slpmpt slpmprodstab[] =
{
    {0, 0, 0, 0, 0, 0, 0, SLPMEND}
};
#endif /* !ORACLE7 */

#endif /* !WIN32 */
#endif /* OLD_ORACLE */

#define MAX_CSNAME    64                  /* Longest character set name    */

/*
** Get size of array of connection objects
*/
int sql_conn_size(int arrsz)
{
    return(arrsz * sizeof(connection));
}

/*
** One-time initialization of OCI.  This is the old technique that is
** compatible with 8.0; for 8i and beyond we don't need it.
*/
int sql_init(void)
{
#ifndef ORACLE7
#ifdef OLD_ORACLE
    sword status;
    status = OCIInitialize((ub4)OCI_THREADED, (dvoid *)0,
                           (dvoid * (*)(dvoid *, size_t))0,
                           (dvoid * (*)(dvoid *, dvoid *, size_t))0,
                           (dvoid (*)(dvoid *, dvoid *))0);
    if (status != OCI_SUCCESS) return(status);
#endif
#endif
    return(0); /* ### ASSUMES ERROR STATUS IS ALWAYS NON-ZERO */
}

/*
** Parse SQL statement
*/
sword sql_parse(connection *c, OraCursor stmhp, char *stmt, int slen)
{
    sword status;
    ub4   stmtlen;
    stmtlen = (slen < 0) ? (ub4)str_length(stmt) : (ub4)slen;
#ifdef ORACLE7
    status = (sword)upiosd(c->hstptr, stmhp, stmt, stmtlen, (ub4)1, (word)1);
#else
    status = OCIStmtPrepare(stmhp, c->errhp, (text *)stmt, stmtlen,
                            (ub4)OCI_NTV_SYNTAX, (ub4)OCI_DEFAULT);
#endif
    return(status);
}

/*
** Fetch next row or piece
*/
sword sql_fetch(connection *c, OraCursor stmhp)
{
#ifdef ORACLE7
    return((sword)upifch(c->hstptr, stmhp));
#else
    return(OCIStmtFetch(stmhp, c->errhp, (ub4)1,
                        (ub2)OCI_FETCH_NEXT, (ub4)OCI_DEFAULT));
#endif
}

/*
** Execute PL/SQL statement through OCI
*/
sword sql_exec(connection *c, OraCursor stmhp, ub4 niters, int exact)
{
#ifdef ORACLE7
    if (exact)
        return((sword)upiefn(c->hstptr, stmhp, niters, (word)0, (word)exact));
    else if (niters == 0)
        return((sword)upiexe(c->hstptr, stmhp));
    else
        return((sword)upiexn(c->hstptr, stmhp, (word)niters, (word)0));
#else
    return(OCIStmtExecute(c->svchp, stmhp, c->errhp, niters, (ub4)0,
                          (OCISnapshot *)0, (OCISnapshot *)0,
                          (exact) ? (ub4)OCI_EXACT_FETCH : (ub4)OCI_DEFAULT));
#endif
}

/*
** Commit or roll back a transaction
*/
sword sql_transact(connection *c, int rollback_flag)
{
    sword status;
#ifdef ORACLE7
    if (rollback_flag)
        status = (sword)upirol(c->hstptr);
    else
        status = (sword)upicom(c->hstptr);
#else
    if (rollback_flag)
        status = OCITransRollback(c->svchp, c->errhp, (ub4)OCI_DEFAULT);
    else
        status = OCITransCommit(c->svchp, c->errhp, (ub4)OCI_DEFAULT);
#endif
    return(status);
}

/*
** Get NLS character set from OCI
*/
void sql_get_nls(connection *c, owa_context *octx)
{
    sword      status;
    char       buf[MAX_CSNAME + 1];
    int        i;
    char      *sptr;

    if (octx->nls_init) return;

    mowa_acquire_mutex(octx);
    if (octx->nls_init)
    {
        mowa_release_mutex(octx);
        return;
    }

    octx->ora_csid = 0;
#ifndef OLD_ORACLE
    /*
    ** First, attempt to find a match to the OCI's NLS character set
    */
    status = OCINlsGetInfo(c->envhp, c->errhp, (text *)buf, sizeof(buf) - 1,
                           (ub2)OCI_NLS_CHARACTER_SET);
    if (status == OCI_SUCCESS)
    {
        i = nls_csx_from_oracle(buf);
        if (i > 0) octx->ora_csid = i;
    }
#endif

    /*
    ** If that fails, then default to the character set from httpd.conf
    */
    if ((octx->ora_csid == 0) && (octx->nls_cs[0] != '\0'))
    {
        i = nls_csx_from_oracle(octx->nls_cs);
        if (i > 0) octx->ora_csid = i;
    }

    if (octx->ora_csid)
    {
      if (octx->dad_csid)
      {
        if (nls_csid(octx->dad_csid) == nls_csid(octx->ora_csid))
            octx->ora_csid = octx->dad_csid;
        else
            c->csid = nls_csid(octx->dad_csid);
      }
      else
        octx->dad_csid = octx->ora_csid;
    }

    /*
    ** Get database character set from V$NLS_PARAMETERS
    */
    octx->db_csid = octx->ora_csid;
    sptr = (char *)"select VALUE from V$NLS_PARAMETERS"
                   " where PARAMETER='NLS_CHARACTERSET'";
    status = sql_parse(c, c->stmhp3, sptr, -1);
    if (status == OCI_SUCCESS)
    {
        status = sql_define(c, c->stmhp3, (ub4)1,
                            (dvoid *)buf, (sb4)MAX_CSNAME, (ub2)SQLT_STR);
        if (status == OCI_SUCCESS)
        {
            status = sql_exec(c, c->stmhp3, (ub4)1, 1);
            if (status == OCI_SUCCESS)
            {
                i = nls_csx_from_oracle(buf);
                if (i > 0) octx->db_csid = i;
            }
        }
    }

    octx->nls_init = 1;
    mowa_release_mutex(octx);
}

/*
** Get a normal error from the OCI error handle and return the code
*/
sb4 sql_get_error(connection *c)
{
    sb4   oerrno = 0;
    char  msgbuf[OCI_ERROR_MAXMSG_SIZE];
    char *errbuf;
#ifdef ORACLE7
    word  msgl;
#endif

    errbuf = c->errbuf;
    if (!errbuf) errbuf = msgbuf;
#ifdef ORACLE7
    msgl = upigml(c->hstptr, errbuf, OCI_ERROR_MAXMSG_SIZE - 1);
    errbuf[msgl] = (text)'\0';
    oerrno =  (sb4)(c->oerptr->oerrcd);
#else
    OCIErrorGet((dvoid *)(c->errhp), (ub4)1, (text *)0,
                &oerrno, (text *)errbuf,
                (ub4)OCI_ERROR_MAXMSG_SIZE, (ub4)OCI_HTYPE_ERROR);
#endif

    return(oerrno);
}

/*
** Get statement error position
*/
int sql_get_errpos(connection *c, OraCursor stmhp)
{
#ifdef ORACLE7
    return((int)(c->oerptr->oerepo));
#else
    ub4   errpos;
    sword status;

    status = OCIAttrGet(stmhp, (ub4)OCI_HTYPE_STMT, (dvoid *)&errpos,
                        (ub4 *)0, (ub4)OCI_ATTR_PARSE_ERROR_OFFSET, c->errhp);

    return((status == OCI_SUCCESS) ? (int)errpos : 0);
#endif
}

/*
** Create OCI connection
*/
sword sql_connect(connection *c, owa_context *octx,
                  char *authuser, char *authpass)
{
    sword  status;
    sb4    oerrno = 0;
    char  *sptr;
    char  *optr;
    char  *username;
    char  *password;
    char  *database;
    int    ulen, plen, dlen;
#ifndef ORACLE7
    ub4    lobempty;
    ub4    buflen;
#endif
    char   buf[1024];

    c->mem_err = 0;
    c->lastsql = (char *)0;
    c->errbuf[0] = '\0';

    if (c->c_lock == C_LOCK_INUSE)
    {
#ifndef ORACLE7
        if (octx->authrealm)
        {
            /* Check current session username/password */
            username = (char *)0;
            buflen = sizeof(username);
            status = OCIAttrGet(c->seshp, (ub4)OCI_HTYPE_SESSION,
                                (dvoid *)&username, &buflen,
                                (ub4)OCI_ATTR_USERNAME, c->errhp);
            if (status != OCI_SUCCESS) username = (char *)0;
            password = (char *)0;
            buflen = sizeof(password);
            status = OCIAttrGet(c->seshp, (ub4)OCI_HTYPE_SESSION,
                                (dvoid *)&password, &buflen,
                                (ub4)OCI_ATTR_PASSWORD, c->errhp);
            if (status != OCI_SUCCESS) password = (char *)0;
            if (!username) username = (char *)"";
            if (!password) password = (char *)"";

            /* If they match, just reset and reuse this session */
            if ((!str_compare(authuser, username, -1, 0)) &&
                (!str_compare(authpass, password, -1, 0)))
                return(owa_reset(c, octx));

            username = authuser;
            password = authpass;
            ulen = str_length(authuser);
            plen = str_length(password);

            /* Otherwise close old session and create new session for user */
            status = OCISessionEnd(c->svchp, c->errhp, c->seshp,
                                   (ub4)OCI_DEFAULT);
            if (status != OCI_SUCCESS) goto connerr;
            status = OCIAttrSet(c->seshp, (ub4)OCI_HTYPE_SESSION, username,
                                (ub4)ulen, (ub4)OCI_ATTR_USERNAME, c->errhp);
            if (status != OCI_SUCCESS) goto connerr;
            status = OCIAttrSet(c->seshp, (ub4)OCI_HTYPE_SESSION, password,
                                (ub4)plen, (ub4)OCI_ATTR_PASSWORD, c->errhp);
            if (status != OCI_SUCCESS) goto connerr;
            status = OCISessionBegin(c->svchp, c->errhp, c->seshp,
                                     (ub4)OCI_CRED_RDBMS, (ub4)OCI_DEFAULT);
            if (status != OCI_SUCCESS) goto connerr;
            status = OCIAttrSet(c->svchp, (ub4)OCI_HTYPE_SVCCTX, c->seshp,
                                (ub4)0, (ub4)OCI_ATTR_SESSION, c->errhp);
            if (status != OCI_SUCCESS) goto connerr;
        }
#endif
        return(OCI_SUCCESS);
    }

#ifdef ORACLE7
    mem_zero(&(c->host), sizeof(c->host));
    c->hstptr = (struct hstdef *)0;
    c->oerptr = (struct oerdef *)0;
    c->stmhp1 = 0;
    c->stmhp2 = 0;
    c->stmhp3 = 0;
    c->stmhp4 = 0;
    c->stmhp5 = 0;
#else
    c->envhp = (OCIEnv *)0;
    c->errhp = (OCIError *)0;
    c->srvhp = (OCIServer *)0;
    c->svchp = (OCISvcCtx *)0;
    c->seshp = (OCISession *)0;
    c->stmhp1 = (OCIStmt *)0;
    c->stmhp2 = (OCIStmt *)0;
    c->stmhp3 = (OCIStmt *)0;
    c->stmhp4 = (OCIStmt *)0;
    c->stmhp5 = (OCIStmt *)0;
    c->pblob = (OCILobLocator *)0;
    c->pclob = (OCILobLocator *)0;
    c->pnlob = (OCILobLocator *)0;
    c->pbfile = (OCILobLocator *)0;
#endif

    optr = octx->oracle_userid;
    if (!optr) return(OCI_ERROR);
    sptr = buf;
    username = sptr;
    for (ulen = 0; *optr != '/'; ++ulen)
    {
        *(sptr++) = *(optr++);
        if ((ulen + 1) == sizeof(buf)) return(OCI_ERROR);
        if (*optr == '\0') return(OCI_ERROR);
    }
    *(sptr++) = '\0';
    password = sptr;
    if (*(++optr) == '\0') return(OCI_ERROR);
    sptr = optr;
    database = (char *)0;
    while (*sptr != '\0')
    {
        if (*sptr == '@') database = sptr;
        ++sptr;
    }
    if (database)
    {
        sptr = (database++);
        if (*database == '\0') database = (char *)0;
    }
    plen = (sptr - optr);
    if ((plen + ulen + 2) >= sizeof(buf)) return(OCI_ERROR);
    mem_copy(password, optr, plen);
    password[plen] = '\0';
    dlen = (database) ? str_length(database) : 0;

    if (octx->authrealm)
    {
        username = authuser;
        password = authpass;
        ulen = str_length(username);
        plen = str_length(password);
    }

#ifdef ORACLE7
    /*
    ** Log on via UPI
    */
    c->hstptr = (struct hstdef *)(dvoid *)(c->host);

    status = (sword)upilog(c->hstptr, username, ulen, password, plen,
                           database, dlen, (ub4)0, (ub4)0);
    if (status != 0) goto connerr;

    c->oerptr = upih2o(c->hstptr);

    /*
    ** Create four UPI cursors
    */
    status = (sword)upiopn(c->hstptr, &(c->stmhp1), (sword)-1);
    if (status != 0) goto connerr;
    status = (sword)upiopn(c->hstptr, &(c->stmhp2), (sword)-1);
    if (status != 0) goto connerr;
    status = (sword)upiopn(c->hstptr, &(c->stmhp3), (sword)-1);
    if (status != 0) goto connerr;
    status = (sword)upiopn(c->hstptr, &(c->stmhp4), (sword)-1);
    if (status != 0) goto connerr;
    status = (sword)upiopn(c->hstptr, &(c->stmhp5), (sword)-1);
    if (status != 0) goto connerr;

#else /* OCI 8.x */
    /*
    ** Initialize OCI handles
    */
#ifdef OLD_ORACLE
    status = OCIEnvInit(&(c->envhp), (ub4)OCI_DEFAULT, (size_t)0, (dvoid *)0);
#else
    status = OCIEnvCreate(&(c->envhp), (ub4)OCI_THREADED, (dvoid *)0,
                          (dvoid * (*)(dvoid *, size_t))0,
                          (dvoid * (*)(dvoid *, dvoid *, size_t))0,
                          (dvoid (*)(dvoid *, dvoid *))0,
                          (size_t)0, (dvoid **)0);
#endif

    if (status != OCI_SUCCESS) goto handerr;
    status = OCIHandleAlloc(c->envhp, (dvoid **)&(c->errhp),
                            (ub4)OCI_HTYPE_ERROR, (size_t)0, (dvoid **)0);
    if (status != OCI_SUCCESS) goto handerr;
    status = OCIHandleAlloc(c->envhp, (dvoid **)&(c->svchp),
                            (ub4)OCI_HTYPE_SVCCTX, (size_t)0, (dvoid **)0);
    if (status != OCI_SUCCESS) goto handerr;
    status = OCIHandleAlloc(c->envhp, (dvoid **)&(c->srvhp),
                            (ub4)OCI_HTYPE_SERVER, (size_t)0, (dvoid **)0);
    if (status != OCI_SUCCESS) goto handerr;

    /*
    ** Attach to specified database server
    */
    status = OCIServerAttach(c->srvhp, c->errhp, (text *)database, (sb4)dlen,
                             (ub4)OCI_DEFAULT);
    if (status != OCI_SUCCESS) goto connerr;

    /* Set the server for use with the service context */
    status = OCIAttrSet(c->svchp, (ub4)OCI_HTYPE_SVCCTX, c->srvhp, (ub4)0,
                        (ub4)OCI_ATTR_SERVER, c->errhp);
    if (status != OCI_SUCCESS) goto connerr;

    /*
    ** Log on and create a new session
    */
    status = OCIHandleAlloc(c->envhp, (dvoid **)&(c->seshp),
                            (ub4)OCI_HTYPE_SESSION, (size_t)0, (dvoid **)0);
    if (status != OCI_SUCCESS) goto handerr;
    status = OCIAttrSet(c->seshp, (ub4)OCI_HTYPE_SESSION, username, (ub4)ulen,
                        (ub4)OCI_ATTR_USERNAME, c->errhp);
    if (status != OCI_SUCCESS) goto connerr;
    status = OCIAttrSet(c->seshp, (ub4)OCI_HTYPE_SESSION, password, (ub4)plen,
                        (ub4)OCI_ATTR_PASSWORD, c->errhp);
    if (status != OCI_SUCCESS) goto connerr;
    status = OCISessionBegin(c->svchp, c->errhp, c->seshp,
                             (ub4)OCI_CRED_RDBMS, (ub4)OCI_DEFAULT);
    if (status != OCI_SUCCESS) goto connerr;
    status = OCIAttrSet(c->svchp, (ub4)OCI_HTYPE_SVCCTX, c->seshp, (ub4)0,
                        (ub4)OCI_ATTR_SESSION, c->errhp);
    if (status != OCI_SUCCESS) goto connerr;

    /*
    ** Create reusable cursor handles
    */
    status = OCIHandleAlloc(c->envhp, (dvoid **)&(c->stmhp1),
                            (ub4)OCI_HTYPE_STMT, (size_t)0, (dvoid **)0);
    if (status != OCI_SUCCESS) goto handerr;
    status = OCIHandleAlloc(c->envhp, (dvoid **)&(c->stmhp2),
                            (ub4)OCI_HTYPE_STMT, (size_t)0, (dvoid **)0);
    if (status != OCI_SUCCESS) goto handerr;
    status = OCIHandleAlloc(c->envhp, (dvoid **)&(c->stmhp3),
                            (ub4)OCI_HTYPE_STMT, (size_t)0, (dvoid **)0);
    if (status != OCI_SUCCESS) goto handerr;
    status = OCIHandleAlloc(c->envhp, (dvoid **)&(c->stmhp4),
                            (ub4)OCI_HTYPE_STMT, (size_t)0, (dvoid **)0);
    if (status != OCI_SUCCESS) goto handerr;
    status = OCIHandleAlloc(c->envhp, (dvoid **)&(c->stmhp5),
                            (ub4)OCI_HTYPE_STMT, (size_t)0, (dvoid **)0);
    if (status != OCI_SUCCESS) goto handerr;

    /*
    ** Create reusable BLOB and CLOB descriptors
    */
    status = OCIDescriptorAlloc(c->envhp, (dvoid **)(dvoid *)&(c->pblob),
                                (ub4)OCI_DTYPE_LOB, (size_t)0, (dvoid **)0);
    if (status != OCI_SUCCESS) goto handerr;
    status = OCIDescriptorAlloc(c->envhp, (dvoid **)(dvoid *)&(c->pclob),
                                (ub4)OCI_DTYPE_LOB, (size_t)0, (dvoid **)0);
    if (status != OCI_SUCCESS) goto handerr;
    status = OCIDescriptorAlloc(c->envhp, (dvoid **)(dvoid *)&(c->pnlob),
                                (ub4)OCI_DTYPE_LOB, (size_t)0, (dvoid **)0);
    if (status != OCI_SUCCESS) goto handerr;
    status = OCIDescriptorAlloc(c->envhp, (dvoid **)(dvoid *)&(c->pbfile),
                                (ub4)OCI_DTYPE_FILE, (size_t)0, (dvoid **)0);
    if (status != OCI_SUCCESS) goto handerr;

    lobempty = (ub4)0;
    status = OCIAttrSet(c->pblob, (ub4)OCI_DTYPE_LOB, &lobempty, (ub4)0,
                        (ub4)OCI_ATTR_LOBEMPTY, c->errhp);
    if (status != OCI_SUCCESS) goto connerr;
    status = OCIAttrSet(c->pclob, (ub4)OCI_DTYPE_LOB, &lobempty, (ub4)0,
                        (ub4)OCI_ATTR_LOBEMPTY, c->errhp);
    if (status != OCI_SUCCESS) goto connerr;
    status = OCIAttrSet(c->pnlob, (ub4)OCI_DTYPE_LOB, &lobempty, (ub4)0,
                        (ub4)OCI_ATTR_LOBEMPTY, c->errhp);
    if (status != OCI_SUCCESS) goto connerr;
#endif /* OCI 8.x */

    if (!(octx->nls_lang) || !(octx->nls_terr)) return(status);

    /*
    ** If NLS_LANGUAGE and NLS_TERRITORY specified, do ALTER SESSION
    */
    c->lastsql = (char *)"ALTER SESSION";
    if (octx->nls_lang[0])
    {
        sptr = buf + str_concat(buf, 0,
                                (char *)"alter session set NLS_LANGUAGE='",-1);
        sptr += str_concat(sptr, 0, octx->nls_lang, -1);
        sptr += str_concat(sptr, 0, (char *)"'", -1);
        status = sql_parse(c, c->stmhp3, buf, (sptr - buf));
        if (status == OCI_SUCCESS)
            status = sql_exec(c, c->stmhp3, (ub4)1, 0);
        status = OCI_SUCCESS;
    }
    if (octx->nls_terr[0])
    {
        sptr = buf + str_concat(buf, 0,
                                "alter session set NLS_TERRITORY='", -1);
        sptr += str_concat(sptr, 0, octx->nls_terr, -1);
        sptr += str_concat(sptr, 0, "'", -1);
        status = sql_parse(c, c->stmhp3, buf, (sptr - buf));
        if (status == OCI_SUCCESS)
            status = sql_exec(c, c->stmhp3, (ub4)1, 0);
        status = OCI_SUCCESS;
    }

    /*
    ** If TIME_ZONE is specified, do ALTER SESSION (for Oracle 9i)
    ** If date format specified, also do ALTER SESSION to set up formats.
    */
    if (octx->nls_tz[0])
    {
        sptr = buf + str_concat(buf, 0,
                                "alter session set TIME_ZONE='", -1);
        sptr += str_concat(sptr, 0, octx->nls_tz, -1);
        sptr += str_concat(sptr, 0, "'", -1);
        status = sql_parse(c, c->stmhp3, buf, (sptr - buf));
        if (status == OCI_SUCCESS)
            status = sql_exec(c, c->stmhp3, (ub4)1, 0);
        status = OCI_SUCCESS;
    }
    if (octx->nls_datfmt[0])
    {
        sptr = buf + str_concat(buf, 0,
                                "alter session set NLS_DATE_FORMAT='", -1);
        optr = sptr;
        sptr += str_concat(sptr, 0, octx->nls_datfmt, -1);
        optr = str_char(optr, ' ', 0);
        if (optr) sptr = optr; /* Truncate time portion */
        sptr += str_concat(sptr, 0, "'", -1);
        status = sql_parse(c, c->stmhp3, buf, (sptr - buf));
        if (status == OCI_SUCCESS)
            status = sql_exec(c, c->stmhp3, (ub4)1, 0);
        status = OCI_SUCCESS;
    }
    if (octx->nls_datfmt[0])
    {
        sptr = buf + str_concat(buf, 0,
                                "alter session set NLS_TIMESTAMP_FORMAT='",-1);
        optr = sptr;
        sptr += str_concat(sptr, 0, octx->nls_datfmt, -1);
        optr = str_char(optr, ' ', 0);
        if (optr) optr = str_char(++optr, ' ', 0);
        if (optr) sptr = optr; /* Truncate zone portion */
        sptr += str_concat(sptr, 0, "'", -1);
        status = sql_parse(c, c->stmhp3, buf, (sptr - buf));
        if (status == OCI_SUCCESS)
            status = sql_exec(c, c->stmhp3, (ub4)1, 0);
        status = OCI_SUCCESS;
    }
    if (octx->nls_datfmt[0])
    {
        sptr = buf + str_concat(buf, 0,
                                "alter session set NLS_TIMESTAMP_TZ_FORMAT='",
                                -1);
        sptr += str_concat(sptr, 0, octx->nls_datfmt, -1);
        sptr += str_concat(sptr, 0, "'", -1);
        status = sql_parse(c, c->stmhp3, buf, (sptr - buf));
        if (status == OCI_SUCCESS)
            status = sql_exec(c, c->stmhp3, (ub4)1, 0);
        status = OCI_SUCCESS;
    }

    if (octx->dav)
    {
        /*
        ** Prepare cursors for the five DAV request statements
        */
        c->lastsql = octx->cgi_stmt;
        status = sql_parse(c, c->stmhp1, c->lastsql, -1);
        if (status != OCI_SUCCESS) goto connerr;

        c->lastsql = octx->lob_stmt;
        status = sql_parse(c, c->stmhp2, c->lastsql, -1);
        if (status != OCI_SUCCESS) goto connerr;

        c->lastsql = octx->cpmv_stmt;
        status = sql_parse(c, c->stmhp3, c->lastsql, -1);
        if (status != OCI_SUCCESS) goto connerr;

        c->lastsql = octx->res_stmt;
        status = sql_parse(c, c->stmhp4, c->lastsql, -1);
        if (status != OCI_SUCCESS) goto connerr;

        c->lastsql = octx->prop_stmt;
        status = sql_parse(c, c->stmhp5, c->lastsql, -1);
        if (status != OCI_SUCCESS) goto connerr;
    }
    else
    {
        /*
        ** Prepare cursors for the four statements used on every request:
        */
        c->lastsql = octx->reset_stmt;
        status = sql_parse(c, c->stmhp5, c->lastsql, -1);
        if (status != OCI_SUCCESS) goto connerr;

        c->lastsql = octx->cgi_stmt;
        status = sql_parse(c, c->stmhp1, c->lastsql, -1);
        if (status != OCI_SUCCESS) goto connerr;

        c->lastsql = octx->sec_stmt;
        status = sql_parse(c, c->stmhp2, c->lastsql, -1);
        if (status != OCI_SUCCESS) goto connerr;

        c->lastsql = octx->get_stmt;
        status = sql_parse(c, c->stmhp4, c->lastsql, -1);
        if (status != OCI_SUCCESS) goto connerr;
    }

    /*
    ** Set alternate character set for strings (if necessary for DAD)
    */
    if ((octx->ora_csid) && (octx->dad_csid != octx->ora_csid))
        c->csid = nls_csid(octx->dad_csid);
    else if ((octx->ncflag & (~UNI_MODE_RAW)) && (octx->dad_csid))
        c->csid = nls_csid(octx->dad_csid);
    c->ncflag = (octx->ncflag & UNI_MODE_FULL);

    return(status);

connerr:
    if (status != OCI_SUCCESS)
    {
        oerrno = sql_get_error(c); 
        if (oerrno) status = oerrno;
    }
    return(status);
#ifndef ORACLE7
handerr:
    if (status != OCI_SUCCESS)
    {
        if (!(c->envhp))
            str_copy(c->errbuf, "OCI Error: Environment not created\n");
        else
        {
            OCIErrorGet((dvoid *)(c->envhp), (ub4)1, (text *)0, &oerrno,
                        (text *)(c->errbuf),
                        (ub4)OCI_ERROR_MAXMSG_SIZE, (ub4)OCI_HTYPE_ENV);
            if (oerrno) status = oerrno;
        }
    }
    return(status);
#endif
}

/*
** Close OCI connection
*/
sword sql_disconnect(connection *c)
{
    sword status;
    sb4   oerrno = 0;

    c->c_lock = C_LOCK_NEW;
    c->mem_err = 0;

    if (c->session)
    {
        mem_free((void *)(c->session));
        c->session = (char *)0;
    }

#ifdef ORACLE7
    status = (sword)upicls(c->hstptr, c->stmhp1);
    if (status != 0) goto closeerr;
    status = (sword)upicls(c->hstptr, c->stmhp2);
    if (status != 0) goto closeerr;
    status = (sword)upicls(c->hstptr, c->stmhp3);
    if (status != 0) goto closeerr;
    status = (sword)upicls(c->hstptr, c->stmhp4);
    if (status != 0) goto closeerr;
    status = (sword)upicls(c->hstptr, c->stmhp5);
    if (status != 0) goto closeerr;
    status = (sword)upilof(c->hstptr);

#else /* OCI 8.x */
    status = OCISessionEnd(c->svchp, c->errhp, c->seshp, (ub4)OCI_DEFAULT);
    if (status != OCI_SUCCESS) goto closeerr;
    status = OCIServerDetach(c->srvhp, c->errhp, (ub4)OCI_DEFAULT);
    if (status != OCI_SUCCESS) goto closeerr;
    if (c->stmhp1)
        status = OCIHandleFree((dvoid *)(c->stmhp1), (ub4)OCI_HTYPE_STMT);
    if (status != OCI_SUCCESS) goto closehand;
    if (c->stmhp2)
        status = OCIHandleFree((dvoid *)(c->stmhp2), (ub4)OCI_HTYPE_STMT);
    if (status != OCI_SUCCESS) goto closehand;
    if (c->stmhp3)
        status = OCIHandleFree((dvoid *)(c->stmhp3), (ub4)OCI_HTYPE_STMT);
    if (status != OCI_SUCCESS) goto closehand;
    if (c->stmhp4)
        status = OCIHandleFree((dvoid *)(c->stmhp4), (ub4)OCI_HTYPE_STMT);
    if (status != OCI_SUCCESS) goto closehand;
    if (c->stmhp5)
        status = OCIHandleFree((dvoid *)(c->stmhp5), (ub4)OCI_HTYPE_STMT);
    if (status != OCI_SUCCESS) goto closehand;
    if (c->pblob)
        status = OCIDescriptorFree((dvoid *)(c->pblob), (ub4)OCI_DTYPE_LOB);
    if (status != OCI_SUCCESS) goto closehand;
    if (c->pclob)
        status = OCIDescriptorFree((dvoid *)(c->pclob), (ub4)OCI_DTYPE_LOB);
    if (status != OCI_SUCCESS) goto closehand;
    if (c->pnlob)
        status = OCIDescriptorFree((dvoid *)(c->pnlob), (ub4)OCI_DTYPE_LOB);
    if (status != OCI_SUCCESS) goto closehand;
    if (c->pbfile)
        status = OCIDescriptorFree((dvoid *)(c->pbfile), (ub4)OCI_DTYPE_FILE);
    if (status != OCI_SUCCESS) goto closehand;
    status = OCIHandleFree((dvoid *)(c->seshp), (ub4)OCI_HTYPE_SESSION);
    if (status != OCI_SUCCESS) goto closehand;
    status = OCIHandleFree((dvoid *)(c->srvhp), (ub4)OCI_HTYPE_SERVER);
    if (status != OCI_SUCCESS) goto closehand;
    status = OCIHandleFree((dvoid *)(c->svchp), (ub4)OCI_HTYPE_SVCCTX);
    if (status != OCI_SUCCESS) goto closehand;
    status = OCIHandleFree((dvoid *)(c->errhp), (ub4)OCI_HTYPE_ERROR);
    if (status != OCI_SUCCESS) goto closehand;
    status = OCIHandleFree((dvoid *)(c->envhp), (ub4)OCI_HTYPE_ENV);
#endif /* OCI 8.x */

    return(status);

closeerr:
    if ((status != OCI_SUCCESS) && (c->errbuf))
    {
        oerrno = sql_get_error(c); 
        if (oerrno) status = oerrno;
    }
    return(status);

#ifndef ORACLE7
closehand:
    if (status != OCI_SUCCESS)
    {
        OCIErrorGet((dvoid *)(c->envhp), (ub4)1, (text *)0, &oerrno,
                    (text *)(c->errbuf),
                    (ub4)OCI_ERROR_MAXMSG_SIZE, (ub4)OCI_HTYPE_ENV);
        if (oerrno) status = oerrno;
    }
    return(status);
#endif
}

/*
** Define fetch column
*/
sword sql_define(connection *c, OraCursor stmhp, ub4 pos,
                 dvoid *buf, sb4 buflen, ub2 dtype)
{
    sword status;
#ifdef ORACLE7
    if (!buf)
      status = (sword)upidfps(c->hstptr, stmhp, (ub1)UPIPCWS, (word)pos,
                            (ptr_t)buf, (size_t)buflen, (word)dtype,
                            (b2 *)0, (text *)0, (size_t)0, (ub2 *)0, (ub2 *)0,
                            (sb4)0, (sb4)0, (sb4)0, (sb4)0, (word)0);
    else
      status = (sword)upidfn(c->hstptr, stmhp, (word)pos, (ptr_t)buf,
                             (size_t)buflen, (word)dtype,
                             (b2 *)(dvoid *)&(c->out_ind),
                             (text *)0, (size_t)0,
                             (ub2 *)0, (ub2 *)0, (word)0);
#else
    OCIDefine *dhand;

    status = OCIDefineByPos(stmhp, &dhand, c->errhp, pos,
                            buf, buflen, dtype,
                            (buf ? (dvoid *)&(c->out_ind) : (dvoid *)0),
                            (ub2 *)0, (ub2 *)0,
                            (ub4)(buf ? OCI_DEFAULT : OCI_DYNAMIC_FETCH));

    if (((dtype == SQLT_STR) || (dtype == SQLT_LNG)) &&
        (status == OCI_SUCCESS) && (c->csid))
        status = OCIAttrSet(dhand, (ub4)OCI_HTYPE_DEFINE, &(c->csid), (ub4)0,
                            (ub4)OCI_ATTR_CHARSET_ID, c->errhp);
#endif
    return(status);
}

#ifndef ORACLE7
/*
** Set character set ID and form-of-use flag
*/
static sword set_cs_info(connection *c, OCIBind *bhand)
{
    sword status;
    ub1   cs_form;

    status = OCI_SUCCESS;

    if (c->csid)
    {
        status = OCIAttrSet(bhand, (ub4)OCI_HTYPE_BIND, &(c->csid), (ub4)0,
                            (ub4)OCI_ATTR_CHARSET_ID, c->errhp);
        if (status != OCI_SUCCESS) return(status);
    }

    if (c->ncflag & (~UNI_MODE_RAW))
    {
        cs_form = (ub1)SQLCS_NCHAR;
        status = OCIAttrSet(bhand, (ub4)OCI_HTYPE_BIND, &cs_form, (ub4)0,
                            (ub4)OCI_ATTR_CHARSET_FORM, c->errhp);
    }

    return(status);
}
#endif

/*
** Bind a signed 4-byte integer to an OCI statement
*/
sword sql_bind_int(connection *c, OraCursor stmhp, ub4 pos, sb4 *val)
{
    sword    status;

#ifdef ORACLE7
    status = (sword)upibrp(c->hstptr, stmhp, (word)pos, (ptr_t)val,
                           sizeof(*val), (word)SQLT_INT,
                           (b2 *)0, (text *)0, (size_t)0);
#else
    OCIBind *bhand;

    status = OCIBindByPos(stmhp, &bhand, c->errhp, pos,
                          (dvoid *)val, (sb4)sizeof(*val),
                          (ub2)SQLT_INT, (dvoid *)0,
                          (ub2 *)0, (ub2 *)0, 0, (ub4 *)0, (ub4)OCI_DEFAULT);
#endif
    return(status);
}

/*
** Bind a null-terminated string buffer to an OCI statement
*/
sword sql_bind_str(connection *c, OraCursor stmhp, ub4 pos,
                   char *str, sb4 slen)
{
    sword    status;

    /* ### ROUND slen? ### */

#ifdef ORACLE7
    int      ncflag = c->ncflag & (~UNI_MODE_RAW);

    if (slen < 0) slen = str_length(str) + 1;

    c->out_ind = 0;
    status = (sword)upibrp(c->hstptr, stmhp, (word)pos, (ptr_t)str,
                           (size_t)slen, (word)SQLT_STR,
                           (b2 *)(dvoid *)&(c->out_ind), (text *)0, (size_t)0);
#ifdef NEVER
    if ((status == 0) && ((c->csid) || (ncflag)))
        status = upibnchr(c->hstptr, stmhp,(sword)pos, c->csid,
                          (ub1)(ncflag ? SQLCS_NCHAR : SQLCS_IMPLICIT),
                          (sb4)slen, (ub1)0);
#endif
#else
    OCIBind *bhand;

    if (slen < 0) slen = str_length(str) + 1;

    c->out_ind = 0;
    status = OCIBindByPos(stmhp, &bhand, c->errhp, pos,
                          (dvoid *)str, slen,
                          (ub2)SQLT_STR, (dvoid *)&(c->out_ind),
                          (ub2 *)0, (ub2 *)0, 0, (ub4 *)0, (ub4)OCI_DEFAULT);

    if (status == OCI_SUCCESS) status = set_cs_info(c, bhand);
#endif

    return(status);
}

/*
** Bind a RAW buffer and length variable to an OCI statement
*/
sword sql_bind_raw(connection *c, OraCursor stmhp, ub4 pos,
                   char *rawbuf, ub2 *rlen, sb4 buflen)
{
    sword    status;
#ifndef ORACLE7
    OCIBind *bhand;
#endif

    c->out_ind = 0;
#ifdef ORACLE7
    status = (sword)upibrp(c->hstptr, stmhp, (word)pos, (ptr_t)rawbuf,
                           (size_t)buflen, (word)SQLT_BIN,
                           (b2 *)(dvoid *)&(c->out_ind), (text *)0, (size_t)0);
#else
    /* ### COULD USE SQLT_LBI BELOW ### */
    status = OCIBindByPos(stmhp, &bhand, c->errhp, pos,
                          (dvoid *)rawbuf, buflen,
                          (ub2)SQLT_BIN, (dvoid *)&(c->out_ind),
                          rlen, (ub2 *)0, 0, (ub4 *)0, (ub4)OCI_DEFAULT);
#endif

    return(status);
}

/*
** Bind an array of null-terminated strings to an OCI statement
*/
sword sql_bind_arr(connection *c, OraCursor stmhp, ub4 pos,
                   char *aptr, ub2 *lens, sb4 awidth, sb2 *inds,
                   ub4 *asize, ub4 amax)
{
    sword    status;

#ifdef ORACLE7
    int      ncflag = c->ncflag & (~UNI_MODE_RAW);

    status = (sword)upibap(c->hstptr, stmhp, (word)pos, (ptr_t)aptr,
                           (size_t)awidth, (word)SQLT_STR,
                           (b2 *)(dvoid *)inds, lens, (ub2 *)0,
                           amax, asize, (text *)0, (size_t)0);
#ifdef NEVER
    if ((status == 0) && ((c->csid) || (ncflag)))
        status = upibnchr(c->hstptr, stmhp,(sword)pos, c->csid,
                          (ub1)(ncflag ? SQLCS_NCHAR : SQLCS_IMPLICIT),
                          awidth, (ub1)0);
#endif
#else
    OCIBind *bhand;

    status = OCIBindByPos(stmhp, &bhand, c->errhp, pos,
                          (dvoid *)aptr, (sb4)awidth, (ub2)SQLT_STR,
                          (dvoid *)inds, lens, (ub2 *)0,
                          amax, asize, (ub4)OCI_DEFAULT);

    if (status == OCI_SUCCESS) status = set_cs_info(c, bhand);
#endif
    return(status);
}

/*
** Bind an array of raws to an OCI statement
*/
sword sql_bind_rarr(connection *c, OraCursor stmhp, ub4 pos,
                    char *aptr, ub2 *lens, sb4 awidth, sb2 *inds,
                    ub4 *asize, ub4 amax)
{
    sword    status;

#ifdef ORACLE7
    status = (sword)upibap(c->hstptr, stmhp, (word)pos, (ptr_t)aptr,
                           (size_t)awidth, (word)SQLT_BIN,
                           (b2 *)(dvoid *)inds, lens, (ub2 *)0,
                           amax, asize, (text *)0, (size_t)0);
#else
    OCIBind *bhand;

    status = OCIBindByPos(stmhp, &bhand, c->errhp, pos,
                          (dvoid *)aptr, (sb4)awidth, (ub2)SQLT_BIN,
                          (dvoid *)inds, lens, (ub2 *)0,
                          amax, asize, (ub4)OCI_DEFAULT);
#endif
    return(status);
}

/*
** Bind an array of integers to an OCI statement
*/
sword sql_bind_iarr(connection *c, OraCursor stmhp, ub4 pos,
                    sb4 *ints, ub4 *asize, ub4 amax)
{
    sword    status;

#ifdef ORACLE7
    status = (sword)upibap(c->hstptr, stmhp, (word)pos, (ptr_t)ints,
                           sizeof(*ints), (word)SQLT_INT,
                           (b2 *)0, (b2 *)0, (ub2 *)0,
                           amax, asize, (text *)0, (size_t)0);
#else
    OCIBind *bhand;

    status = OCIBindByPos(stmhp, &bhand, c->errhp, pos,
                          (dvoid *)ints, (sb4)sizeof(*ints), (ub2)SQLT_INT,
                          (dvoid *)0, (ub2 *)0, (ub2 *)0,
                          amax, asize, (ub4)OCI_DEFAULT);
#endif
    return(status);
}

/*
** Bind an array of length-leading string structures to an OCI statement
*/
sword sql_bind_vcs(connection *c, OraCursor stmhp, ub4 pos,
                   dvoid *buf, ub2 *lens, sb4 awidth, sb2 *inds,
                   ub4 *asize, ub4 amax, ub4 swidth)
{
    sword    status;
    int      dtype = (c->ncflag & UNI_MODE_RAW) ? SQLT_VBI : SQLT_VCS;

#ifdef ORACLE7
    int      ncflag = c->ncflag & (~UNI_MODE_RAW);

    status = (sword)upibpps(c->hstptr, stmhp, UPISKIP, (word)pos,
                            (ptr_t)buf, (size_t)awidth, (word)dtype,
                            (b2 *)(dvoid *)inds, lens, (ub2 *)0,
                            (sb4)swidth, (sb4)sizeof(*inds),
                            (sb4)sizeof(*lens), (sb4)0,
                            amax, asize, (text *)0, (size_t)0);

#ifdef NEVER
    if ((status == 0) && (dtype == SQLT_VCS) && ((c->csid) || (ncflag)))
        status = upibnchr(c->hstptr, stmhp,(sword)pos, c->csid,
                          (ub1)(ncflag ? SQLCS_NCHAR : SQLCS_IMPLICIT),
                          awidth, (ub1)0);
#endif
#else
    OCIBind *bhand;

    status = OCIBindByPos(stmhp, &bhand, c->errhp, pos,
                          buf, (sb4)awidth, (ub2)dtype,
                          (dvoid *)inds, lens, (ub2 *)0,
                          amax, asize, (ub4)OCI_DEFAULT);

    if ((status == OCI_SUCCESS) && (dtype == SQLT_VCS))
        status = set_cs_info(c, bhand);

    if ((status == OCI_SUCCESS) && ((sb4)swidth > awidth) && (amax > (ub4)1))
        status = OCIBindArrayOfStruct(bhand, c->errhp, swidth,
                                      (ub4)sizeof(*inds), (ub4)sizeof(*lens),
                                      (ub4)0);
#endif
    return(status);
}

sword sql_bind_stream(connection *c, OraCursor stmhp, ub4 pos, ub2 sql_type)
{
    sword status;

#ifdef ORACLE7
    status = (sword)upibpps(c->hstptr, stmhp, (ub1)UPIPCWS, (word)pos,
                            (ptr_t)0, (size_t)LONG_MAXSZ, (word)sql_type,
                            (b2 *)0, (ub2 *)0, (ub2 *)0,
                            (sb4)0, (sb4)0, (sb4)0, (sb4)0,
                            (ub4)0, (ub4 *)0, (text *)0, (size_t)0);
#else
    OCIBind *bhand;

    status = OCIBindByPos(stmhp, &bhand, c->errhp, pos,
                          (dvoid *)0, (sb4)LONG_MAXSZ, sql_type,
                          (dvoid *)0, (ub2 *)0, (ub2 *)0, (ub4)0, (ub4 *)0,
                          (ub4)OCI_DATA_AT_EXEC);
#endif
    return(status);
}

#ifndef ORACLE7

/*
** Dynamic data binding for arrays of string pointers:
**   sql_get_input()  - callback for output values
**   sql_put_output() - callback for return values (not implemented)
*/

static sb4 sql_get_input(dvoid *ctx, OCIBind *bindp, ub4 iter, ub4 index,
                         dvoid **bufpp, ub4 *alenp, ub1 *piecep, dvoid **indpp)
{
    char **arr = (char **)ctx;
    ub4    alen;

    alen = (arr[index]) ? str_length(arr[index]) : 0;
    *piecep = OCI_ONE_PIECE;
    *bufpp = (alen == 0) ? (dvoid *)"\0\0" : (dvoid *)arr[index];
    *indpp = (dvoid *)0; /* No indicator */
    *alenp = alen; /* (alen + 1) required for null-terminated strings */
    return(OCI_CONTINUE);
}

static sb4 sql_put_output(dvoid *ctx, OCIBind *bindp, ub4 iter, ub4 index,
                          dvoid **bufpp, ub4 **alenpp,
                          ub1 *piecep, dvoid **indpp, ub2 **rcodepp)
{
    /* ### NOT IMPLEMENTED ### */
    return(OCI_CONTINUE);
}

/*
** Bind array of string pointers
*/
sword sql_bind_ptrs(connection *c, OraCursor stmhp, ub4 pos,
                    char **aptr, sb4 awidth, ub4 *asize, ub4 amax)
{
    sword    status;
    sword    sqltype;
    OCIBind *bhand;

    /* Bind strings as CHR to avoid need to pass the null terminator */
    sqltype = (c->ncflag & UNI_MODE_RAW) ? SQLT_BIN : SQLT_CHR;

    status = OCIBindByPos(stmhp, &bhand, c->errhp, pos,
                          (dvoid *)0, awidth, sqltype,
                          (dvoid *)0, (ub2 *)0, (ub2 *)0,
                          amax, asize, (ub4)OCI_DATA_AT_EXEC);

    if (!(c->ncflag & UNI_MODE_RAW))
      if (status == OCI_SUCCESS)
        status = set_cs_info(c, bhand);

#ifdef NEVER
    if (status == OCI_SUCCESS)
        status = OCIBindArrayOfStruct(bhand, c->errhp,
                                      (ub4)0, (ub4)0, (ub4)0, (ub4)0);
#endif

    if (status == OCI_SUCCESS)
        status = OCIBindDynamic(bhand, c->errhp,
                                (dvoid *)aptr, sql_get_input,
                                (dvoid *)aptr, sql_put_output);

    return(status);
}

/*
** Bind a LOB locator to an OCI statement
*/
sword sql_bind_lob(connection *c, OraCursor stmhp, ub4 pos, ub2 flag)
{
    sword           status;
    ub2             cs_id;
    ub2            *indp;
    ub1             cs_form;
    OCIBind        *bhand;
    OCILobLocator **pl;

    if (flag == SQLT_BLOB)
    {
        pl = &(c->pblob);
        indp = &(c->blob_ind);
        c->blob_ind = 0;
        cs_form = (ub1)SQLCS_IMPLICIT;
        cs_id = 0;
    }
    else if (flag == SQLT_CLOB)
    {
        pl = &(c->pclob);
        indp = &(c->clob_ind);
        c->clob_ind = 0;
        cs_form = (ub1)SQLCS_IMPLICIT;
        cs_id = c->csid;
    }
    else if (flag == SQLT_BFILE)
    {
        pl = &(c->pbfile);
        indp = &(c->bfile_ind);
        c->bfile_ind = 0;
        cs_form = (ub1)SQLCS_IMPLICIT;
        cs_id = 0;
    }
    else
    {
        pl = &(c->pnlob);
        indp = &(c->nlob_ind);
        c->nlob_ind = 0;
        flag = SQLT_CLOB;
        cs_form = (ub1)SQLCS_NCHAR;
        cs_id = c->csid;
    }

    status = OCIBindByPos(stmhp, &bhand, c->errhp, pos, (dvoid *)pl,
                          (sb4)sizeof(*pl), flag, indp, (ub2 *)0, (ub2 *)0,
                          (ub4)0, (ub4 *)0, (ub4)OCI_DEFAULT);

    if ((status == OCI_SUCCESS) && (cs_form != (ub1)SQLCS_IMPLICIT))
        status = OCIAttrSet(bhand, (ub4)OCI_HTYPE_BIND, &cs_form, (ub4)0,
                            (ub4)OCI_ATTR_CHARSET_FORM, c->errhp);

    /* ### cs_id NOT USED FOR LOB BINDINGS -- ELIMINATE THIS CODE? ### */
    if ((status == OCI_SUCCESS) && (cs_id > 0))
        status = OCIAttrSet(bhand, (ub4)OCI_HTYPE_BIND, &cs_id, (ub4)0,
                            (ub4)OCI_ATTR_CHARSET_ID, c->errhp);

    return(status);
}
#endif

/*
** Set next piece to be written to database
*/
sword sql_write_piece(connection *c, OraCursor stmhp,
                      dvoid *piecebuf, ub4 *nbytes, int pieceflag)
{
    sword    status;
    ub1      piece;
#ifndef ORACLE7
    ub4      htype = OCI_HTYPE_BIND;
    ub1      in_out = OCI_PARAM_IN;
    ub4      iter;
    ub4      idx;
    OCIBind *bhand;

    status = OCIStmtGetPieceInfo(c->stmhp3, c->errhp, (dvoid **)&bhand,
                                 &htype, &in_out, &iter, &idx, &piece);
    if (status != OCI_SUCCESS) return(status);
#endif

    piece = pieceflag;

#ifdef ORACLE7
    status = (sword)upispi(c->hstptr, c->stmhp3, piece, piecebuf, nbytes);
#else
    c->out_ind = (ub2)0;
    c->rcode = (ub2)0;
    status = OCIStmtSetPieceInfo(bhand, htype, c->errhp,
                                 piecebuf, nbytes, piece,
                                 &(c->out_ind), &(c->rcode));
#endif
    return(status);
}

/*
** Set next piece to be read from database
*/
sword sql_read_piece(connection *c, OraCursor stmhp,
                     dvoid *piecebuf, ub4 *nbytes)
{
    sword      status;
    ub1        piece;
    ub4        iter;
    ub4        idx;
#ifdef ORACLE7
    dvoid     *ctxp;
#else
    ub4        htype = OCI_HTYPE_DEFINE;
    ub1        in_out = OCI_PARAM_OUT;
    OCIDefine *dhand;
#endif

#ifdef ORACLE7
    status = (sword)upigpi(c->hstptr, stmhp,
                           &piece, &ctxp, &iter, &idx);
#else
    status = OCIStmtGetPieceInfo(c->stmhp3, c->errhp, (dvoid **)&dhand,
                                 &htype, &in_out, &iter, &idx, &piece);
#endif
    if (status == OCI_SUCCESS)
    {
#ifdef ORACLE7
        status = (sword)upispi(c->hstptr, c->stmhp3, piece, piecebuf, nbytes);
#else
        c->out_ind = (ub2)0;
        c->rcode = (ub2)0;
        status = OCIStmtSetPieceInfo(dhand, htype, c->errhp,
                                     piecebuf, nbytes, piece,
                                     &(c->out_ind), &(c->rcode));
#endif
    }
    return(status);
}

#ifdef ORACLE7

/*
** Normal describe,
** promotion of array arguments only
*/
static int desc_normal(connection *c, char *pname,
                       int nargs, char *names[], ub4 counts[])
{
    sword        status;
    int          result;
    ub4          slen;
    ub4          cnt;
    char        *namep;
    int          i, j;
    int          pos;
    ub2          overload[DESCRIBE_MAX];
    ub2          position[DESCRIBE_MAX];
    ub2          level[DESCRIBE_MAX];
    char         args[DESCRIBE_MAX * ORA_MAXCOLNAME];
    ub2          alens[DESCRIBE_MAX];
    ub2          dtype[DESCRIBE_MAX];
    ub1          deflt[DESCRIBE_MAX];
    ub1          inout[DESCRIBE_MAX];
    ub4          len[DESCRIBE_MAX];
    sb2          precision[DESCRIBE_MAX];
    sb2          scale[DESCRIBE_MAX];
    ub1          radix[DESCRIBE_MAX];
    ub4          spare[DESCRIBE_MAX];

    slen = str_length(pname);

    result = 0;

    cnt = (ub4)DESCRIBE_MAX;

    status = (sword)upidpr(c->hstptr, pname, slen,
                           (ptr_t)0, (size_t)0, (ptr_t)0, (size_t)0,
                           overload, position, level, (text *)args, alens,
                           dtype, deflt, inout, len, precision, scale,
                           radix, spare, &cnt);
    if (status != 0) return(result);

    pos = 0;
    for (i = 0; i < cnt; ++i)
    {
        if (level[i] > 0) continue; /* Skip substructure expansions */

        /* If found at least one promotion, skip the rest of the overloads */
        if ((position[i] < pos) && (result)) break;
        pos = position[i];

        /*
        ** If this is an array type,
        ** look for a bound argument with a matching name that's bound
        ** as a scalar and promote it to an array bind type.
        **
        ** ### Note that this no longer matches the OCI8 case!
        */
        if (dtype[i] == PLSQL_TAB)
        {
            namep = args + (i * ORA_MAXCOLNAME);
            for (j = 0; j < nargs; ++j)
            {
                if ((!str_compare(names[j], namep, alens[i], 1)) &&
                    (namep[alens[i]] == '\0'))
                {
                    if (counts[j] == 0)
                    {
                        counts[j] = 1;
                        result = 1;
                    }
                    break;
                }
            }
        }
    }

    return(result);
}

/*
** Relaxed describe,
** promotes array arguments and eliminates unmatched parameters
*/
static int desc_relaxed(connection *c, char *pname,
                        int nargs, char *names[], ub4 counts[])
{
    sword        status;
    int          result;
    ub4          slen;
    ub4          cnt;
    char        *namep;
    int          i, j;
    int          pos;
    ub2          overload[DESCRIBE_MAX];
    ub2          position[DESCRIBE_MAX];
    ub2          level[DESCRIBE_MAX];
    char         args[DESCRIBE_MAX * ORA_MAXCOLNAME];
    ub2          alens[DESCRIBE_MAX];
    ub2          dtype[DESCRIBE_MAX];
    ub1          deflt[DESCRIBE_MAX];
    ub1          inout[DESCRIBE_MAX];
    ub4          len[DESCRIBE_MAX];
    sb2          precision[DESCRIBE_MAX];
    sb2          scale[DESCRIBE_MAX];
    ub1          radix[DESCRIBE_MAX];
    ub4          spare[DESCRIBE_MAX];

    slen = str_length(pname);

    result = 0;

    cnt = (ub4)DESCRIBE_MAX;

    status = (sword)upidpr(c->hstptr, pname, slen,
                           (ptr_t)0, (size_t)0, (ptr_t)0, (size_t)0,
                           overload, position, level, (text *)args, alens,
                           dtype, deflt, inout, len, precision, scale,
                           radix, spare, &cnt);
    if (status != 0) return(result);

    pos = 0;
    for (i = 0; i < DESCRIBE_MAX; ++i) spare[i] = LONG_MAXSZ;
    for (i = 0; i < cnt; ++i)
    {
        if (level[i] > 0) continue; /* Skip substructure expansions */

        /* If found at least one promotion, skip the rest of the overloads */
        if (position[i] < pos)
        {
            if (result) break;
            for (i = 0; i < DESCRIBE_MAX; ++i) spare[i] = LONG_MAXSZ;
        }
        pos = position[i];

        /*
        ** Look for a bound argument with a matching name
        */
        namep = args + (i * ORA_MAXCOLNAME);
        for (j = 0; j < nargs; ++j)
        {
            if ((!str_compare(names[j], namep, alens[i], 1)) &&
                (namep[alens[i]] == '\0'))
            {
                /*
                ** If this is an array type but the binding is scalar,
                ** promote it to an array bind type.
                */
                if ((dtype[i] == PLSQL_TAB) && (counts[j] == 0))
                {
                    counts[j] = 1;
                    result = 1;
                }
                if (j < DESCRIBE_MAX)
                {
                    /* Mark argument as found */
                    if ((dtype[i] != PLSQL_TAB) && (counts[j] > 0))
                        spare[j] = 0; /* Mark argument for demotion */
                    else
                        spare[j] = 1;
                }
            }
        }
    }

    j = (nargs > DESCRIBE_MAX) ? DESCRIBE_MAX : nargs;
    for (i = 0; i < j; ++i)
      if (spare[i] == 0)
      {
          counts[i] = spare[i];
          result = 1;
      }

    return(result);
}

/*
** Positional describe,
** promotion of array arguments only
*/
static int desc_positional(connection *c, char *pname,
                           int nargs, char *names[], ub4 counts[])
{
    sword        status;
    int          result;
    ub4          slen;
    ub4          cnt;
    char        *namep;
    int          i;
    int          pos;
    ub2          overload[DESCRIBE_MAX];
    ub2          position[DESCRIBE_MAX];
    ub2          level[DESCRIBE_MAX];
    char         args[DESCRIBE_MAX * ORA_MAXCOLNAME];
    ub2          alens[DESCRIBE_MAX];
    ub2          dtype[DESCRIBE_MAX];
    ub1          deflt[DESCRIBE_MAX];
    ub1          inout[DESCRIBE_MAX];
    ub4          len[DESCRIBE_MAX];
    sb2          precision[DESCRIBE_MAX];
    sb2          scale[DESCRIBE_MAX];
    ub1          radix[DESCRIBE_MAX];
    ub4          spare[DESCRIBE_MAX];

    slen = str_length(pname);

    result = 0;

    cnt = (ub4)DESCRIBE_MAX;

    status = (sword)upidpr(c->hstptr, pname, slen,
                           (ptr_t)0, (size_t)0, (ptr_t)0, (size_t)0,
                           overload, position, level, (text *)args, alens,
                           dtype, deflt, inout, len, precision, scale,
                           radix, spare, &cnt);
    if (status != 0) return(result);

    pos = 0;
    for (i = 0; i < cnt; ++i)
    {
        if (level[i] > 0) continue; /* Skip substructure expansions */

        /* If found at least one promotion, skip the rest of the overloads */
        if ((position[i] < pos) && (result)) break;
        pos = position[i];

        /*
        ** If this is an array type,
        ** look for a bound argument with a matching name that's bound
        ** as a scalar and promote it to an array bind type.
        */
        if ((dtype[i] == PLSQL_TAB) && (counts[i] == 0))
        {
            counts[i] = 1;
            result = 1;
        }
    }

    return(result);
}

#else /* OCI implementation */

/*
** Compare the call to PL/SQL arguments by name and type,
** attempt to adjust array-bound parameters.  Returns TRUE
** if any array-bound parameter is promoted by the process.
*/
static int check_normal(connection *c, dvoid *arg_list,
                        int nargs, char *names[], ub4 counts[])
{
    sword  status;
    ub4    pos;
    ub4    acount;
    int    argsfound = 0;
    int    i;
    char  *namep;
    ub4    namsz;
    dvoid *argp;
    ub2    dtype;
    int    result = 0;

    status = OCIAttrGet(arg_list, (ub4)OCI_DTYPE_PARAM, (dvoid *)&acount,
                        (ub4 *)0, (ub4)OCI_ATTR_NUM_PARAMS, c->errhp);
    if (status != OCI_SUCCESS) return(result);

    for (pos = 1; pos <= acount; ++pos)
    {
        status = OCIParamGet(arg_list, (ub4)OCI_DTYPE_PARAM,
                             c->errhp, (dvoid **)&argp, (ub4)pos);
        if (status != OCI_SUCCESS) break;

        status = OCIAttrGet(argp, (ub4)OCI_DTYPE_PARAM,
                            (dvoid *)&namep, (ub4 *)&namsz,
                            (ub4)OCI_ATTR_NAME, c->errhp);
        if (status != OCI_SUCCESS) break;

        status = OCIAttrGet(argp, (ub4)OCI_DTYPE_PARAM,
                            (dvoid *)&dtype, (ub4 *)0,
                            (ub4)OCI_ATTR_DATA_TYPE, c->errhp);
        if (status != OCI_SUCCESS) break;

        /*
        ** Look for a bound argument with a matching name that's bound
        ** as a scalar and promote it to an array bind type.
        */
        for (i = 0; i < nargs; ++i)
        {
            /* If the names match */
            if ((!str_compare(names[i], namep, namsz, 1)) &&
                (names[i][namsz] == '\0'))
            {
                ++argsfound;
                if (dtype != PLSQL_TAB)
                {
                    /* Array binding to scalar, signatures don't match */
                    if (counts[i] > 0) return(0);
                }
                else if (counts[i] == 0)
                {
                    counts[i] = 1;
                    result = 1;
                }
                break;
            }
        }
    }

    if (argsfound < nargs) result = 0; /* Some arguments not found */

    return(result);
}

/*
** Compare the call to PL/SQL arguments by name and type,
** attempt to adjust array-bound parameters.  Returns TRUE
** if any array-bound parameter is promoted by the process.
*/
static int check_relaxed(connection *c, dvoid *arg_list,
                         int nargs, char *names[], ub4 counts[])
{
    sword  status;
    ub4    pos;
    ub4    acount;
    int    i;
    char  *namep;
    ub4    namsz;
    dvoid *argp;
    ub2    dtype;
    int    result = 0;

    status = OCIAttrGet(arg_list, (ub4)OCI_DTYPE_PARAM, (dvoid *)&acount,
                        (ub4 *)0, (ub4)OCI_ATTR_NUM_PARAMS, c->errhp);
    if (status != OCI_SUCCESS) return(result);

    for (pos = 1; pos <= acount; ++pos)
    {
        status = OCIParamGet(arg_list, (ub4)OCI_DTYPE_PARAM,
                             c->errhp, (dvoid **)&argp, (ub4)pos);
        if (status != OCI_SUCCESS) break;

        status = OCIAttrGet(argp, (ub4)OCI_DTYPE_PARAM,
                            (dvoid *)&namep, (ub4 *)&namsz,
                            (ub4)OCI_ATTR_NAME, c->errhp);
        if (status != OCI_SUCCESS) break;

        status = OCIAttrGet(argp, (ub4)OCI_DTYPE_PARAM,
                            (dvoid *)&dtype, (ub4 *)0,
                            (ub4)OCI_ATTR_DATA_TYPE, c->errhp);
        if (status != OCI_SUCCESS) break;

        /*
        ** Look for a bound argument with a matching name
        */
        for (i = 0; i < nargs; ++i)
        {
            if ((!str_compare(names[i], namep, namsz, 1)) &&
                (names[i][namsz] == '\0'))
            {
                /*
                ** If this is an array type but the binding is scalar,
                ** promote it to an array bind type.
                */
                if ((dtype == PLSQL_TAB) && (counts[i] == 0))
                {
                    counts[i] = 1;
                    result = 1;
                }
                else if ((dtype != PLSQL_TAB) && (counts[i] > 0))
                {
                    counts[i] = 0; /* Mark argument for demotion */
                    result = 1;
                }
            }
        }
    }

    return(result);
}

/*
** Compare the call to PL/SQL arguments by position and type,
** attempt to adjust array-bound parameters.  Returns TRUE
** if any array-bound parameter is promoted by the process.
*/
static int check_positional(connection *c, dvoid *arg_list,
                            int nargs, char *names[], ub4 counts[])
{
    sword  status;
    ub4    pos;
    ub4    acount;
    char  *namep;
    ub4    namsz;
    dvoid *argp;
    ub2    dtype;
    int    result = 0;

    status = OCIAttrGet(arg_list, (ub4)OCI_DTYPE_PARAM, (dvoid *)&acount,
                        (ub4 *)0, (ub4)OCI_ATTR_NUM_PARAMS, c->errhp);
    if (status != OCI_SUCCESS) return(result);

    for (pos = 1; pos <= acount; ++pos)
    {
        status = OCIParamGet(arg_list, (ub4)OCI_DTYPE_PARAM,
                             c->errhp, (dvoid **)&argp, (ub4)pos);
        if (status != OCI_SUCCESS) break;

        status = OCIAttrGet(argp, (ub4)OCI_DTYPE_PARAM,
                            (dvoid *)&namep, (ub4 *)&namsz,
                            (ub4)OCI_ATTR_NAME, c->errhp);
        if (status != OCI_SUCCESS) break;

        status = OCIAttrGet(argp, (ub4)OCI_DTYPE_PARAM,
                            (dvoid *)&dtype, (ub4 *)0,
                            (ub4)OCI_ATTR_DATA_TYPE, c->errhp);
        if (status != OCI_SUCCESS) break;

        /*
        ** If this is an array type but the variable in this postion
        ** is a scalar,  promote it to an array bind type.
        */
        if ((dtype == PLSQL_TAB) && (counts[pos] == 0))
        {
            counts[pos] = 1;
            result = 1;
        }
    }

    return(result);
}

#endif

/*
** Describe PL/SQL procedure and attempt to adjust array parameters
** Returns TRUE for successful conversion, FALSE otherwise.
*/
int sql_describe(connection *c, char *pname, int descmode, char *schema,
                 int nargs, char *names[], ub4 counts[])
{
    int          result;
#ifdef ORACLE7

    if (descmode == DESC_MODE_STRICT)
        result = 0;
    if (descmode == DESC_MODE_NORMAL)
        result = desc_normal(c, pname, nargs, names, counts);
    if (descmode == DESC_MODE_POSITIONAL)
        result = desc_positional(c, pname, nargs, names, counts);
    if (descmode == DESC_MODE_RELAXED)
        result = desc_relaxed(c, pname, nargs, names, counts);

#else /* OCI implementation */

    sword        status;
    int          tmpresult;
    ub4          slen;
    ub1          ptype;
    ub4          cnt;
    char        *sptr;
    dvoid       *arg_list;
    dvoid       *argp;
    dvoid       *proc_list;
    ub2          pcount;
    OCIParam    *parmp;
    OCIDescribe *dschp;
    char        *namep;
    ub4          namsz;

    result = 0;
    if (descmode == DESC_MODE_STRICT) return(result);

    sptr = str_char(pname, '.', 1);
    if (sptr)
    {
        ptype = OCI_PTYPE_PKG;
        slen = (sptr - pname);
        ++sptr;
    }
    else
    {
        ptype = OCI_PTYPE_PROC;
        slen = str_length(pname);
    }

    /*
    ** Create the describe handle
    */
    status = OCIHandleAlloc(c->envhp, (dvoid **)&dschp,
                            (ub4)OCI_HTYPE_DESCRIBE, 0, (dvoid **)0);
    if (status != OCI_SUCCESS) return(result);

    status = OCIDescribeAny(c->svchp, c->errhp,
                            (dvoid *)pname, (ub4)slen, (ub1)OCI_OTYPE_NAME,
                            (ub1)OCI_DEFAULT, ptype, dschp);
    if (status != OCI_SUCCESS) goto descexit;

    status = OCIAttrGet(dschp, (ub4)OCI_HTYPE_DESCRIBE, (dvoid *)&parmp,
                        (ub4 *)0, (ub4)OCI_ATTR_PARAM, c->errhp);
    if (status != OCI_SUCCESS) goto descexit;

    if (ptype == OCI_PTYPE_PROC)
    {
        status = OCIAttrGet(parmp, (ub4)OCI_DTYPE_PARAM, (dvoid *)&arg_list,
                            (ub4 *)0, (ub4)OCI_ATTR_LIST_ARGUMENTS, c->errhp);
        if (status != OCI_SUCCESS) goto descexit;

        switch (descmode)
        {
        case DESC_MODE_NORMAL:
            result = check_normal(c, arg_list, nargs, names, counts);
            break;
        case DESC_MODE_RELAXED:
            result = check_relaxed(c, arg_list, nargs, names, counts);
            break;
        case DESC_MODE_POSITIONAL:
            result = check_positional(c, arg_list, nargs, names, counts);
            break;
        default:
            break;
        }
    }
    else
    {
        slen = str_length(sptr);
        status = OCIAttrGet(parmp, (ub4)OCI_DTYPE_PARAM, (dvoid *)&proc_list,
                            (ub4 *)0, (ub4)OCI_ATTR_LIST_SUBPROGRAMS,
                            c->errhp);
        if (status != OCI_SUCCESS) goto descexit;

        status = OCIAttrGet(proc_list, (ub4)OCI_DTYPE_PARAM, (dvoid *)&pcount,
                            (ub4 *)0, (ub4)OCI_ATTR_NUM_PARAMS, c->errhp);
        if (status != OCI_SUCCESS) goto descexit;

        for (cnt = 0; cnt < pcount; ++cnt)
        {
            status = OCIParamGet(proc_list, (ub4)OCI_DTYPE_PARAM,
                                 c->errhp, (dvoid **)&argp, (ub4)cnt);
            if (status != OCI_SUCCESS) break;

            status = OCIAttrGet(argp, (ub4)OCI_DTYPE_PARAM,
                                (dvoid *)&namep, (ub4 *)&namsz,
                                (ub4)OCI_ATTR_NAME, c->errhp);
            if (status != OCI_SUCCESS) break;

            if ((slen == namsz) && (!str_compare(namep, sptr, namsz, 1)))
            {
                status = OCIAttrGet(argp, (ub4)OCI_DTYPE_PARAM,
                                    (dvoid *)&arg_list, (ub4 *)0,
                                    (ub4)OCI_ATTR_LIST_ARGUMENTS, c->errhp);
                if (status != OCI_SUCCESS) break;

                switch (descmode)
                {
                case DESC_MODE_NORMAL:
                  tmpresult = check_normal(c,arg_list,nargs,names,counts);
                  break;
                case DESC_MODE_RELAXED:
                  tmpresult = check_relaxed(c,arg_list,nargs,names,counts);
                  break;
                case DESC_MODE_POSITIONAL:
                  tmpresult = check_positional(c,arg_list,nargs,names,counts);
                  break;
                default:
                  tmpresult = 0;
                  break;
                }
                if (tmpresult)
                {
                    result = 1;
                    break; /* Stop at first successful conversion */
                }
            }
        }
    }

descexit:
    status = OCIHandleFree((dvoid *)dschp, (ub4)OCI_HTYPE_DESCRIBE);

#endif /* OCI implementation */

    return(result);
}
