/*
** Copyright (c) 1999-2004 Oracle Corporation, All rights reserved.
**
** File:         owautil.c
** Description:  Utility functions
** Author:       Doug McMahon      dmcmahon@us.oracle.com
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
** String functions
**   str_concat       length-limited string concatenation
**   str_compare      length-limited case-controlled comparison
**   str_char         reversible search for character in string
**   str_substr       case-controlled search for substring in string
**   str_dup          duplicate string
**   str_chrcmp       case-aware comparison of two characters
**   str_prepend      prepend a string to another string
**   str_itoa         convert integer to string
**   str_itox         convert integer to hex string
**   str_atoi         convert decimal string to integer
**   str_atox         convert hex string to integer
**
** Misc. functions
**   util_get_method  get HTTP method name given numeric code
**   util_print_time  print formatted date/time
**   util_set_mime    determine mime type based on file extension
**   util_checksum    compute 32-bit checksum of input buffer
**   util_ipaddr      convert IP address/mask string to integer
**   util_decode64    decode a base-64 encoded string
**   util_make_rand   create linear additive feedback random number generator
**   util_gen_rand    generate a value from random number generator
**   util_seed_rand   seed the state of a random number generator
**   util_get_rand    get a random value from a pair of generators
**   util_init_rand   initialize a pair of generators from a key
**
** History
**
** 08/17/2000   D. McMahon      Created
** 05/31/2001   D. McMahon      Began revisioning with 1.3.9
** 09/07/2001   D. McMahon      Added TIMING diagnostic
** 11/26/2001   D. McMahon      Added str_chrcmp
** 08/16/2002   D. McMahon      Add method name/number conversions
** 05/26/2003   D. McMahon      Add random number generator code
** 01/19/2004   D. McMahon      Added str_dup
*/

#include <modowa.h>

/*
** Length-limited string concatenation, always null-terminates.
** Starts concatenation within the destination string at the
** point specified.  The maxlen protects the total size of the
** output buffer.  Returns the new total length of the string in
** dest.  With 0 for the starting offset, behaves like strncpy.
*/
int str_concat(char *dest, int dstart, const char *src, int maxlen)
{
    int i;

    if (maxlen < 0) maxlen = LONG_MAXSZ;

    if (!dest) return(-1);
    if (!src) src = "";

    dest += dstart;
    for (i = dstart; i < maxlen; ++i)
    {
        if (*src == '\0') break;
        *(dest++) = *(src++);
    }
    *dest = '\0';
    return(i);
}

/*
** Length-limited string comparison with case-control flag.
*/
int str_compare(const char *s1, const char *s2, int maxlen, int caseflag)
{
    int i;
    int n, m;

    if (maxlen < 0) maxlen = LONG_MAXSZ;

    if ((!s1) && (!s2)) return(0);
    if (!s1) return(-1);
    if (!s2) return(+1);
    for (i = 0; i < maxlen; ++i, ++s1, ++s2)
    {
        n = ((int)*s1 & 0xFF);
        m = ((int)*s2 & 0xFF);
        if (caseflag)
        {
            if ((n >= 'A') && (n <= 'Z')) n += ('a' - 'A');
            if ((m >= 'A') && (m <= 'Z')) m += ('a' - 'A');
        }
        n -= m;
        if (n != 0) return(n);
        if (*s1 == '\0') break;
    }
    return(0);
}

/*
** Search for character in string, either from the beginning or
** in reverse from the end.
*/
char *str_char(const char *sptr, int ch, int reverse_flag)
{
    char *s;

    if (reverse_flag)
    {
        s = (char *)sptr + str_length(sptr);
        while (s > sptr)
            if (*(--s) == ch) break;
    }
    else
    {
        for (s = (char *)sptr; *s != ch; ++s)
            if (*s == '\0') break;
    }
    if (*s != ch) s = (char *)0;

    return(s);
}

/*
** Search for substring within string, optionally without case sensitivity.
*/
char *str_substr(const char *sptr, const char *ssub, int caseflag)
{
    int i;

    if ((sptr) && (ssub))
    {
        for (i = str_length(ssub); *sptr != '\0'; ++sptr)
            if (!str_compare(sptr, ssub, i, caseflag))
                return((char *)sptr);
    }
    return((char *)0);
}

/*
** Duplicate string
*/
char *str_dup(const char *sptr)
{
    char *newstr;

    if (!sptr) return((char *)0);
    newstr = (char *)mem_alloc(str_length(sptr) + 1, 0);
    if (newstr) str_copy(newstr, sptr);
    return(newstr);
}

/*
** Case-aware comparison of two characters
*/
int str_chrcmp(char *a, char *b, int caseflag)
{
    int la = ((int)*a & 0xFF);
    int lb = ((int)*b & 0xFF);
    if (caseflag)
    {
        if ((la >= 'A') && (la <= 'Z')) la += ('a' - 'A');
        if ((lb >= 'A') && (lb <= 'Z')) lb += ('a' - 'A');
    }
    return(la - lb);
}

/*
** Prepend a string to another string; shift the base string to
** make room for the prefix, then copy it in.
*/
void str_prepend(char *base_str, char *prefix_str)
{
    int i, j;

    j = str_length(prefix_str);
    i = str_length(base_str) + 1;

    while (i > 0)
    {
        --i;
        base_str[j + i] = base_str[i];
    }

    mem_copy(base_str, prefix_str, j);
}

/*
** Convert integer to string
*/
int str_itoa(int i, char *s)
{
    int   j;
    int   n;
    char *sptr;
    char  buf[LONG_MAXSTRLEN];

    sptr = s;
    if (i <= 0)
    {
        j = -i;
        *(sptr++) = ((i == 0) ? '0' : '-');
    }
    else
    {
        j = i;
    }
    for (n = 0; j != 0; j = j/10) buf[n++] = '0' + (j % 10);
    for (j = n - 1; j >= 0; --j) *(sptr++) = buf[j];
    *sptr = '\0';
    return((int)(sptr - s));
}

/*
** Convert integer to hex string
*/
int str_itox(un_long i, char *s)
{
    char *hex = (char *)"0123456789ABCDEF";
    int   j;

    for (j = 8; j > 0; i >>= 4) s[--j] = hex[i & 0xF];
    s[8] = '\0';
    return(8);
}

/*
** Convert decimal string to integer
*/
int str_atoi(char *s)
{
    int i;
    for (i = 0; (*s >= '0') && (*s <= '9'); ++s)
        i = (i * 10) + (*s - '0');
    return(i);
}

/*
** Convert hex string to integer
*/
int str_atox(char *s)
{
    char *hexstr = (char *)"0123456789ABCDEFabcdef";
    char *hptr;
    int   i;
    int   val;

    i = 0;
    while (1)
    {
        hptr = str_char(hexstr, *s, 0);
        if (!hptr) break;
        val = (hptr - hexstr);
        if (val >= 16) val -= 6;
        i = (i * 16) + val;
    }
    return(i);
}

static char *method_list[] = {"GET","PUT","POST","DELETE","CONNECT","OPTIONS",
                              "TRACE","PATCH","PROPFIND","PROPPATCH","MKCOL",
                              "COPY","MOVE","LOCK","UNLOCK"};
                              /* ### What about "HEAD"? ### */

/*
** Convert Apache numeric method number to HTTP string
*/
char *util_get_method(int m)
{
    if ((m < 0) || (m > sizeof(method_list)/sizeof(*method_list))) return("");
    return(method_list[m]);
}

/*
** Print Unix time as YYYY/MM/DD HH24:MI:SS
**                 or YYYY/MM/DD HH24:MI:SS.FFFFFF
*/
static const int days_in_year[]  = {365,366,365,365};
static const int days_in_month[] = {31,28,31,30,31,30,31,31,30,31,30,31,
                                    31,29,31,30,31,30,31,31,30,31,30,31};

void util_print_time(un_long t, char *outbuf, un_long *ms)
{
    int i;
    int leap;
    int secs;
    int days;
    int year;
    int mon;
    int hr;
    int min;

    days = (int)(t/(24*60*60));
    secs = (int)(t - ((un_long)days * 24*60*60));
    year = days/(365*4 + 1);
    days -= (year * (365*4 + 1));
    year = 1970 + year * 4;
    for (i = 0; i < 4; ++i)
        if (days < days_in_year[i])
            break;
        else
            days -= days_in_year[i];
    leap = (i == 1) ? 12 : 0;
    year += i;
    for (i = 0; i < 12; ++i)
      if (days < days_in_month[i + leap])
          break;
      else
          days -= days_in_month[i + leap];
    mon = i + 1;
    ++days;
    hr = secs/(60*60);
    secs -= (hr * 60*60);
    min = secs/60;
    secs -= (min * 60);
    os_str_print(outbuf, "%4d/%2.2d/%2.2d %2.2d:%2.2d:%2.2d",
                 year, mon, days, hr, min, secs);
    if (ms) os_str_print(outbuf + 19, ".%6.6ld", *ms);
}

/*
** Mime type determination
** ### This is a kludge.  Apache should provide some means for
** ### getting a mime type based on a file extension.
*/

/*
** Sort-of order by frequency because underlying search is linear
*/
static const char mime_list[] =
"[text/html]"                       " html htm "
"[image/gif]"                       " gif "
"[image/jpeg]"                      " jpeg jpg jpe "
"[text/plain]"                      " asc txt "
"[text/css]"                        " css "
"[text/xml]"                        " xml "
"[application/x-javascript]"        " js "
"[image/png]"                       " png "
"[image/tiff]"                      " tiff tif "
"[image/bmp]"                       " bmp "
"[application/x-shockwave-flash]"   " swf "
"[application/msword]"              " doc "
"[application/vnd.ms-excel]"        " xls "
"[application/vnd.ms-powerpoint]"   " ppt "
"[application/pdf]"                 " pdf "
"[application/rtf]"                 " rtf "
"[application/x-gunzip]"            " gz "
"[application/zip]"                 " zip "
"[application/x-tar]"               " tar "
"[audio/x-wav]"                     " wav "
"[audio/mpeg]"                      " mpga mp2 mp3 "
"[audio/basic]"                     " au snd "
"[audio/midi]"                      " mid midi kar "
"[audio/x-aiff]"                    " aif aiff aifc "
"[audio/x-pn-realaudio]"            " ram rm "
"[audio/x-pn-realaudio-plugin]"     " rpm "
"[audio/x-realaudio]"               " ra "
"[video/mpeg]"                      " mpeg mpg mpe "
"[video/quicktime]"                 " qt mov "
"[text/richtext]"                   " rtx "
"[text/rtf]"                        " rtf "
"[text/sgml]"                       " sgml sgm "
"[text/tab-separated-values]"       " tsv "
"[text/vnd.wap.wml]"                " wml "
"[text/vnd.wap.wmlscript]"          " wmls "
"[text/x-setext]"                   " etx "
"[image/ief]"                       " ief "
"[image/vnd.wap.wbmp]"              " wbmp "
"[image/x-cmu-raster]"              " ras "
"[image/x-portable-anymap]"         " pnm "
"[image/x-portable-bitmap]"         " pbm "
"[image/x-portable-graymap]"        " pgm "
"[image/x-portable-pixmap]"         " ppm "
"[image/x-rgb]"                     " rgb "
"[image/x-xbitmap]"                 " xbm "
"[image/x-xpixmap]"                 " xpm "
"[image/x-xwindowdump]"             " xwd "
"[application/andrew-inset]"        " ez "
"[application/mac-binhex40]"        " hqx "
"[application/mac-compactpro]"      " cpt "
"[application/octet-stream]"        " bin dms lha lzh exe class so dll "
"[application/oda]"                 " oda "
"[application/postscript]"          " ai eps ps "
"[application/smil]"                " smi smil "
"[application/vnd.mif]"             " mif "
"[application/vnd.wap.wbxml]"       " wbxml "
"[application/vnd.wap.wmlc]"        " wmlc "
"[application/vnd.wap.wmlscriptc]"  " wmlsc "
"[application/x-bcpio]"             " bcpio "
"[application/x-cdlink]"            " vcd "
"[application/x-chess-pgn]"         " pgn "
"[application/x-cpio]"              " cpio "
"[application/x-csh]"               " csh "
"[application/x-director]"          " dcr dir dxr "
"[application/x-dvi]"               " dvi "
"[application/x-futuresplash]"      " spl "
"[application/x-gtar]"              " gtar "
"[application/x-hdf]"               " hdf "
"[application/x-koan]"              " skp skd skt skm "
"[application/x-latex]"             " latex "
"[application/x-netcdf]"            " nc cdf "
"[application/x-sh]"                " sh "
"[application/x-shar]"              " shar "
"[application/x-stuffit]"           " sit "
"[application/x-sv4cpio]"           " sv4cpio "
"[application/x-sv4crc]"            " sv4crc "
"[application/x-tcl]"               " tcl "
"[application/x-tex]"               " tex "
"[application/x-texinfo]"           " texinfo texi "
"[application/x-troff]"             " t tr roff "
"[application/x-troff-man]"         " man "
"[application/x-troff-me]"          " me "
"[application/x-troff-ms]"          " ms "
"[application/x-ustar]"             " ustar "
"[application/x-wais-source]"       " src "
"[chemical/x-pdb]"                  " pdb "
"[chemical/x-xyz]"                  " xyz "
"[model/iges]"                      " igs iges "
"[model/mesh]"                      " msh mesh silo "
"[model/vrml]"                      " wrl vrml "
"[video/x-msvideo]"                 " avi "
"[video/x-sgi-movie]"               " movie "
"[x-conference/x-cooltalk]"         " ice "
"\0";

/*
** Determine mime type based on file name extension
*/
void util_set_mime(char *fpath, char *pmimetype, int bin_flag)
{
    int   i;
    char *nptr, *sptr;
    char *ext;

#ifdef DOESNT_WORK
    request_rec   *subr;
    subr = ap_sub_req_lookup_uri(fpath, r);
    if (subr)
      if (subr->content_type)
      {
        str_copy(pmimetype, (char *)(subr->content_type));
        return;
      }
#endif

    if (*pmimetype) return;

    sptr = str_char(fpath, '.', 1);
    if (sptr)
    {
        ++sptr;
        ext = pmimetype;
        *ext = ' ';
        i = str_concat(ext + 1, 0, sptr, 64); /* ### ARBITRARY MAXIMUM ### */
        ext[++i] = ' ';
        ext[++i] = '\0';

        sptr = str_substr(mime_list, ext, 1);
        if (sptr)
        {
            while (*sptr != '[') --sptr;
            for (nptr = ++sptr; *nptr != ']'; ++nptr);
            str_concat(pmimetype, 0, sptr, nptr - sptr);
            return;
        }
    }

    str_copy(pmimetype, (bin_flag) ? "application/octet-stream": "text/plain");
}

/*
** 32-bit CRC, pre-calculated based on the polynomial:
**
**   x^32 + x^26 + x^23 + x^22 + x^16 + x^12 + x^11 +
**          x^10 + x^8  + x^7  + x^5  + x^4  + x^2  + x    + 1
**
*/
static const un_long util_crc32[] = {
    0x00000000,0x77073096,0xee0e612c,0x990951ba,0x076dc419,0x706af48f,
    0xe963a535,0x9e6495a3,0x0edb8832,0x79dcb8a4,0xe0d5e91e,0x97d2d988,
    0x09b64c2b,0x7eb17cbd,0xe7b82d07,0x90bf1d91,0x1db71064,0x6ab020f2,
    0xf3b97148,0x84be41de,0x1adad47d,0x6ddde4eb,0xf4d4b551,0x83d385c7,
    0x136c9856,0x646ba8c0,0xfd62f97a,0x8a65c9ec,0x14015c4f,0x63066cd9,
    0xfa0f3d63,0x8d080df5,0x3b6e20c8,0x4c69105e,0xd56041e4,0xa2677172,
    0x3c03e4d1,0x4b04d447,0xd20d85fd,0xa50ab56b,0x35b5a8fa,0x42b2986c,
    0xdbbbc9d6,0xacbcf940,0x32d86ce3,0x45df5c75,0xdcd60dcf,0xabd13d59,
    0x26d930ac,0x51de003a,0xc8d75180,0xbfd06116,0x21b4f4b5,0x56b3c423,
    0xcfba9599,0xb8bda50f,0x2802b89e,0x5f058808,0xc60cd9b2,0xb10be924,
    0x2f6f7c87,0x58684c11,0xc1611dab,0xb6662d3d,0x76dc4190,0x01db7106,
    0x98d220bc,0xefd5102a,0x71b18589,0x06b6b51f,0x9fbfe4a5,0xe8b8d433,
    0x7807c9a2,0x0f00f934,0x9609a88e,0xe10e9818,0x7f6a0dbb,0x086d3d2d,
    0x91646c97,0xe6635c01,0x6b6b51f4,0x1c6c6162,0x856530d8,0xf262004e,
    0x6c0695ed,0x1b01a57b,0x8208f4c1,0xf50fc457,0x65b0d9c6,0x12b7e950,
    0x8bbeb8ea,0xfcb9887c,0x62dd1ddf,0x15da2d49,0x8cd37cf3,0xfbd44c65,
    0x4db26158,0x3ab551ce,0xa3bc0074,0xd4bb30e2,0x4adfa541,0x3dd895d7,
    0xa4d1c46d,0xd3d6f4fb,0x4369e96a,0x346ed9fc,0xad678846,0xda60b8d0,
    0x44042d73,0x33031de5,0xaa0a4c5f,0xdd0d7cc9,0x5005713c,0x270241aa,
    0xbe0b1010,0xc90c2086,0x5768b525,0x206f85b3,0xb966d409,0xce61e49f,
    0x5edef90e,0x29d9c998,0xb0d09822,0xc7d7a8b4,0x59b33d17,0x2eb40d81,
    0xb7bd5c3b,0xc0ba6cad,0xedb88320,0x9abfb3b6,0x03b6e20c,0x74b1d29a,
    0xead54739,0x9dd277af,0x04db2615,0x73dc1683,0xe3630b12,0x94643b84,
    0x0d6d6a3e,0x7a6a5aa8,0xe40ecf0b,0x9309ff9d,0x0a00ae27,0x7d079eb1,
    0xf00f9344,0x8708a3d2,0x1e01f268,0x6906c2fe,0xf762575d,0x806567cb,
    0x196c3671,0x6e6b06e7,0xfed41b76,0x89d32be0,0x10da7a5a,0x67dd4acc,
    0xf9b9df6f,0x8ebeeff9,0x17b7be43,0x60b08ed5,0xd6d6a3e8,0xa1d1937e,
    0x38d8c2c4,0x4fdff252,0xd1bb67f1,0xa6bc5767,0x3fb506dd,0x48b2364b,
    0xd80d2bda,0xaf0a1b4c,0x36034af6,0x41047a60,0xdf60efc3,0xa867df55,
    0x316e8eef,0x4669be79,0xcb61b38c,0xbc66831a,0x256fd2a0,0x5268e236,
    0xcc0c7795,0xbb0b4703,0x220216b9,0x5505262f,0xc5ba3bbe,0xb2bd0b28,
    0x2bb45a92,0x5cb36a04,0xc2d7ffa7,0xb5d0cf31,0x2cd99e8b,0x5bdeae1d,
    0x9b64c2b0,0xec63f226,0x756aa39c,0x026d930a,0x9c0906a9,0xeb0e363f,
    0x72076785,0x05005713,0x95bf4a82,0xe2b87a14,0x7bb12bae,0x0cb61b38,
    0x92d28e9b,0xe5d5be0d,0x7cdcefb7,0x0bdbdf21,0x86d3d2d4,0xf1d4e242,
    0x68ddb3f8,0x1fda836e,0x81be16cd,0xf6b9265b,0x6fb077e1,0x18b74777,
    0x88085ae6,0xff0f6a70,0x66063bca,0x11010b5c,0x8f659eff,0xf862ae69,
    0x616bffd3,0x166ccf45,0xa00ae278,0xd70dd2ee,0x4e048354,0x3903b3c2,
    0xa7672661,0xd06016f7,0x4969474d,0x3e6e77db,0xaed16a4a,0xd9d65adc,
    0x40df0b66,0x37d83bf0,0xa9bcae53,0xdebb9ec5,0x47b2cf7f,0x30b5ffe9,
    0xbdbdf21c,0xcabac28a,0x53b39330,0x24b4a3a6,0xbad03605,0xcdd70693,
    0x54de5729,0x23d967bf,0xb3667a2e,0xc4614ab8,0x5d681b02,0x2a6f2b94,
    0xb40bbe37,0xc30c8ea1,0x5a05df1b,0x2d02ef8d};

/*
** Compute 32-bit CRC on input buffer
*/
un_long util_checksum(char *buffer, int buflen)
{
    int     i;
    un_long crc;

    crc = (~(un_long)0);

    for (i = 0; i < buflen; ++i)
        crc = util_crc32[(crc ^ buffer[i]) & 0xFF] ^ (crc >> 8);

    crc ^= (~(un_long)0);

    return(crc);
}

/*
** Convert IP address or mask from string to integer
*/
un_long util_ipaddr(char *ipstr)
{
    un_long ip;
    for (ip = 0; *ipstr != '\0'; ++ipstr)
    {
        if ((*ipstr >= '0') && (*ipstr <= '9'))
            ip = (ip & ~0xFF) | (((ip & 0xFF) * 10) + (*ipstr - '0'));
        else if (*ipstr == '.')
            ip <<= 8;
    }
    return(ip);
}

/*
** Round an integer to next highest multiple of another integer
*/
un_long util_round(un_long n, un_long r)
{
    un_long i;
    if (!r) return(n);
    i = n % r;
    if (i > 0) i = r - i;
    return(n + i);
}

/*
** Decode a base-64 encoded string
*/
int util_decode64(char *inbuf, char *outbuf)
{
    int   i, j;
    int   val;
    int   bits;
    char *bptr;
    char *b64;

    b64 = (char *)
          "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
    bits = 0;
    val = 0;
    j = 0;

    for (i = 0; inbuf[i] != '\0'; ++i)
    {
        bptr = str_char(b64, inbuf[i], 0);
        if (!bptr) break;
        val = (val << 6) | (bptr - b64);
        bits += 6;
        if (bits >= 8)
        {
            outbuf[j++] = (val >> (bits - 8)) & 0xFF;
            bits -= 8;
            val &= (0xFF >> (8 - bits));
        }
    }
    outbuf[j] = '\0';
    return(j);
}

/*
** Array of default polynomials indexed by sequence length.  The
** default polynomial for any length n is always (X[n] + X[k] + 0),
** where k is the constant from the table (i.e., these are all
** polynomials with three coefficients, one of which is n and the
** other obtained from the table).  In cases where no polynomial
** exists with two coefficients, the table contains a 0 -- in these
** cases, the next-lower non-zero value should be used.
*/
static int util_poly64[] =
       {0,0,1,1,1,2,1,1,0,4,3,2,0,0,0,1,0,3,7,0,3,2,1,5,0,3,0,0,3,2,0,3,0,
        13,0,2,11,0,0,4,0,3,0,0,0,0,0,5,0,9,0,0,3,0,0,24,0,7,19,0,1,0,0,1};

/*
** util_make_rand
**
** Make an additive feedback generator.  An additive feedback generator
** generates a sequence of values from:
**
**   X[i] = f(X[i - a], X[i - b], ..., X[i - m]) mod 2^n
**
** In this implementation, the generator works by keeping previous
** states up to a given cycle length, and adding back one tap.  The
** function f() can be non-linear, though the default combination
** function is addition.  n = 32 (so that the generator works on 32-bit
** machines).  Thus, the default sequence is of the form:
**
**   X[i] = (X[i - a] + X[i - b]) mod 2^32
**
** where "a" is the cycle length, and "b" is a feedback tap location.
** To maximize the cycle length for a given "a", "b" should be chosen
** so that together (a,b,0) is a primitive polynomial.
**
**      prand           Pointer to the structure to be initialized
**      length          Desired cycle length, up to 63
**      tap             Desired tap location, or 0 to use an internal default
*/
void util_make_rand(randstate *prand, int length, int tap)
{
    int cycle;

    if ((!prand) || (length < 2)) return; /* Fatal errors */

    cycle = length;

    if ((tap == 0) || (tap >= cycle))
    {
        /* Find a default from the table */
        cycle = sizeof(util_poly64)/sizeof(*util_poly64) - 1;
        if (cycle > length) cycle = length;
        while (util_poly64[cycle] == 0) --cycle;
        prand->tap = util_poly64[cycle];
    }
    else
        prand->tap = tap;

    prand->length = cycle;
    prand->ind = 0;

    /* Zero the state vector */
    while (cycle > 0) prand->state[--cycle] = 0;
}

/*
** util_gen_rand
**
** Additive feedback generator.  This generates a sequence of numbers:
**
**   X[i] = (X[i - a] + X[i - b]) mod 2^32
**
**      prand           Pointer to the state structure
**
**      returns:        Next sequence value
*/
un_long util_gen_rand(randstate *prand)
{
    int     length;
    un_long nxt, tap;
    un_long result;

    length = prand->length;

    nxt = prand->state[prand->ind];
    tap = prand->state[((prand->ind + length) - prand->tap) % length];

    result = nxt + tap;
    prand->carry = ((result < nxt) || (result < tap));

    prand->state[prand->ind] = result;
    prand->ind = (prand->ind + 1) % length;

    return(result);
}

/*
** util_seed_rand
**
** Seed the data for an additive feedback system.  The contents of the
** seed array are XORed to the state array, and then the generator is
** clocked through an equivalent number of steps.  Multiple calls can
** be made to seed successive values.
**
**      prand           Pointer to the state structure
**      srand           Pointer to seed data generator
*/
void util_seed_rand(randstate *prand, randstate *srand)
{
    int      i;

    for (i = 0; i < prand->length; ++i)
    {
        prand->state[prand->ind] ^= util_gen_rand(srand);
        util_gen_rand(prand);
    }
}

/*
** util_get_rand
**
** Generate random number from a pair of generators (a, b) according
** to the following algorithm:
**
**   1. If carry(a), clock(b)
**   2. If carry(b), clock(a)
**   3. Clock(a), set carry bit
**   4. Clock(b), set carry bit
**   5. Output (A XOR B)
**
**      arand           Pointer to the state structure for generator a
**      brand           Pointer to the state structure for generator b
**
**      returns:        Next random value
*/
un_long util_get_rand(randstate *arand, randstate *brand)
{
    int bcarry;

    if (!arand) return(util_gen_rand(brand));
    if (!brand) return(util_gen_rand(arand));

    bcarry = brand->carry;
    if (arand->carry) util_gen_rand(brand);
    if (bcarry) util_gen_rand(arand);

    return(util_gen_rand(arand) ^ util_gen_rand(brand));
}

/*
** util_init_rand
**
** Seed a pair of random number generators from a key value.
**
**      arand           Pointer to the state structure for generator a
**      brand           Pointer to the state structure for generator b
**      key             Seed value (array of bytes)
**      keylen          Number of bytes in the seed
*/
void util_init_rand(randstate *arand, randstate *brand, char *key, int keylen)
{
    int       i;
    int       keymax;
    randstate trand;

    keymax = (keylen < RAND_MAX_STATES) ? keylen : RAND_MAX_STATES;

    /* Spread whatever randomness exists across more bits */
    util_make_rand(&trand, keymax, 0);
    for (i = 0; i < keylen; ++i)
        trand.state[i % trand.length] ^= util_checksum(key, i + 1);
    for (i = 0; i < trand.length; ++i)
        util_gen_rand(&trand);

    util_make_rand(arand, 55, 24);
    util_make_rand(brand, 52, 19);

    util_seed_rand(arand, &trand);
    util_seed_rand(brand, &trand);
}
